// ==============================================================
//                 ORBITER MODULE: NTR_inlinetank
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2004 Martin Schweiger
//                   All rights reserved
//
// NTR_inlinetank.cpp
// Control module for NTR_inlinetank vessel class
//
// Notes:
// This is an example for a "minimal" vessel implementation which
// only overloads the clbkSetClassCaps method to define vessel
// capabilities and otherwise uses the default VESSEL class
// behaviour.
// ==============================================================
// does not launch from surface gracefully :)

// TODO: sun tracking and power output for other modules
// TODO: Isp change, integrate dV to shutdown
// DONE: PVA animation during checkout
// TODO: extending nozzle skirt during checkout
// TODO: visuals for gimballing animation
// DONE: ullage burn pre-startup
// TODO: tumbling trip?

#define STRICT
#define ORBITER_MODULE

#include "NTR_inlinetank.h"

using namespace std;

// todo: varying Isp
#define VMSG_NTROFFSET                  (VMSG_USER+7500)
#define VMSG_GET_PROPELLANT_TYPE        (VMSG_NTROFFSET+1)
#define VMSG_GET_FUEL_TYPE              (VMSG_NTROFFSET+2)
#define VMSG_GET_OXIDIZER_TYPE          (VMSG_NTROFFSET+3)
#define VMSG_GET_3RD_COMPONENT_TYPE     (VMSG_NTROFFSET+4)
#define VMSG_GET_PRESSURANT_TYPE        (VMSG_NTROFFSET+5)
#define VMSG_GET_POWER_OUTPUT_W         (VMSG_NTROFFSET+6)
#define VMSG_GET_TANK_DRAWDOWN_SEQUENCE (VMSG_NTROFFSET+7)


// avoid 1 and 0 defaults
enum PROPELLANT_TYPE
{
	UNKNOWN_PROPELLANT = -1,
	MONOPROPELLANT = 2,
	BIPROPELLANT,
	TRIPROPELLANT,
	BIPROPELLANT_COMBINED,
	TRIPROPELLANT_COMBINED
};

enum FUEL_TYPE
{
	UNKNOWN_FUEL = -1,
	HYDROGEN_CRYOGENIC = 2,
	HYDROGEN_SLUSH,
	METHANE_CRYOGENIC,
	PROPANE_CRYOGENIC,
	XENON_GASEOUS,
	HYDRAZINE,
	MONOMETHYLHYDRAZINE,
	UNSYMMETRIC_DIMETHYLHYDRAZINE,
	PHENYLHYDRAZINE,
	AEROZINE50,
	METHYLNITRATE,
	ETHYLNITRATE,
	ISOPROPYLNITRATE,
	ETHYLENEGLYCOLDINITRATE,
	TRINITROGLYCERINE,
	NITROMETHANE,
	NITROETHANE,
	NITROPROPANE,
	ETHYLENEOXIDE,
	HYDROGEN_PEROXIDE_MONOPROP,
	RP1,
	JP1,
	JP4,
	AMMONIA_CRYOGENIC,
	ANILINE,
	MONOETHYLAMINE,
	DIETHYLAMINE,
	TRIETHYLAMINE,
	DIETHYLENETRIAMINE,
	KEROSENE,
	HYDINE,
	ETHANOL,
	TONKA250
};

enum OXIDIZER_TYPE
{
	UNKNOWN_OXIDIZER = -1,
	OXYGEN_CRYOGENIC = 2,
	OZONE_CRYOGENIC,
	HYDROGEN_PEROXIDE,
	TETRANITROMETHANE,
	NITROGEN_TETROXIDE,
	NITRIC_ACID
};

enum PRESSURANT_TYPE
{
	UNKNOWN_PRESSURANT = -1,
	HELIUM = 2,
	NITROGEN_GASEOUS
};

enum TANK_TYPE
{
	UNKNOWN_TANK_TYPE = -1,
	INTEGRAL_TANK = 2,
	INLINE_TANK,
	DROP_TANK
};


// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	int i;
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,   -0.1,     0,   0.2,  0.25,   0.2,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		(CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
		(CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
		(CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	for (i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);
	return CL[i] + (aoa-AOA[i])*SCL[i];
}


double NTR_inlinetank::AnimationSpeeds[NUM_ANIMATIONS] = {
	0.1, // SOLAR1
	0.1  // SOLAR2
};

UINT NTR_inlinetank::Animations[NUM_ANIMATIONS] = { 
	NULL, //SOLAR1
	NULL  //SOLAR2
};

NTR_inlinetank::NTR_inlinetank (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
	//MaintainCommandedRates = false;
	//Controllers[RCS_R] = PIDControl("RCS Roll", RCS_DAP_Gain, 0.0, 0.0, -1.0, 1.0); //?
	//Controllers[RCS_P] = PIDControl("RCS Pitch", RCS_DAP_Gain, 0.0, 0.0, -1.0, 1.0); //?
	//Controllers[RCS_Y] = PIDControl("RCS Yaw", RCS_DAP_Gain, 0.0, 0.0, -1.0, 1.0); //?

	IsInLaunchMode = false;
	IsInCheckoutMode = false;
	Static_dP_dt_criterion_counter = 0;
	G_limit_criterion_counter = 0;
	old_STP = HUGE_VAL;
	DefineAnimations();
	for (int i = 0; i<NUM_ANIMATIONS; i++)
	{
		AnimationDeployments[i] = 1.0;
		AnimationStatus[i] = simple_anim_status_t::OPEN;
	};
//	CommandedRates = NULL_VECTOR;
	//JointGimbalAttitude = NULL_VECTOR;
//	UllageBurnTimeLeft = 0.0;
//	CurrentScreen = glass_screen_t::ENGINE_CONTROL;
	ShallTrackSun = false;
	ShallTrackPoint = false;
	TrackPointEulers = _V(0,0,0);
	for (int i = 0; i<3; i++)
	{
		ActiveDAPChannels[i] = true;
	};
	// test! works OK!
	//this->OptimizeTorque();
//	optProblem = new nlopt::opt(nlopt::algorithm::LN_COBYLA,NUM_REACTORS*2);
	visual = NULL;
}

NTR_inlinetank::~NTR_inlinetank ()
{
}

void NTR_inlinetank::clbkVisualCreated(VISHANDLE vis, int refcount)
{
	VESSEL2::clbkVisualCreated(vis, refcount);
	visual = vis;
}

void NTR_inlinetank::clbkVisualDestroyed(VISHANDLE vis, int refcount)
{
	visual = NULL;
	VESSEL2::clbkVisualDestroyed(vis, refcount);
}

void NTR_inlinetank::DefineAnimations (void)
{
	// 1. Solar array 1
	static UINT Solar1Grp[3] = {MESHGRP_ARRAY1,MESHGRP_ARRAY1_2, MESHGRP_SARJ1};
	static MGROUP_ROTATE MGR_Solar1 (MESH_MAIN,
		Solar1Grp, 3, 
		PVA1_Axis_Offset, 
		DIR_XY_225_Z_0, 
		(float)(PI*0.5));

	Animations[SOLAR1_ANIMATION] = CreateAnimation (1.0);
	AddAnimationComponent (Animations[SOLAR1_ANIMATION], 0, 1.0, &MGR_Solar1);

	// Solar array 2
	static UINT Solar2Grp[3] = {MESHGRP_ARRAY2,MESHGRP_ARRAY2_2, MESHGRP_SARJ2};
	static MGROUP_ROTATE MGR_Solar2 (MESH_MAIN,
		Solar2Grp, 3, 
		PVA2_Axis_Offset, 
		DIR_XY_045_Z_0, 
		(float)(PI*0.5));

	Animations[SOLAR2_ANIMATION] = CreateAnimation (1.0);
	AddAnimationComponent (Animations[SOLAR2_ANIMATION], 0.0, 1.0, &MGR_Solar2);

}

void NTR_inlinetank::SetThrusterGroupLevel_SingleStep(THGROUP_TYPE  thgt, double dlevel) const
{
	for (unsigned int i = 0; i<this->GetGroupThrusterCount(thgt); i++)
	{
		THRUSTER_HANDLE th = this->GetGroupThruster(thgt, i);
		if (th!=NULL)
		{
			this->SetThrusterLevel_SingleStep(th, dlevel);
		};
	};
}


void NTR_inlinetank::clbkPostStep (double simt, double simdt, double mjd)
{
}

//double NTR_inlinetank::GetStackMass()
//{
//	double M = 0.0;
//	if (stackInfo.size()==0) FillStackInfo(this, stackInfo);
//	for (unsigned int i = 0; i<stackInfo.size(); i++)
//	{
//		M += stackInfo[i].pV->GetMass();
//	};
//	return M;
//}
//
//bool NTR_inlinetank::GetStackGroundContact()
//{
//	if (stackInfo.size()==0) FillStackInfo(this, stackInfo);
//	for (unsigned int i = 0; i<stackInfo.size(); i++)
//	{
//		if (stackInfo[i].pV->GroundContact()) return true;
//	};
//	return false;
//}

int NTR_inlinetank::clbkGeneric(int msgid, int prm, void *context)
{
	switch(msgid)
	{
	case VMSG_GET_PROPELLANT_TYPE:
		switch(prm)
		{
		case 0: return PROPELLANT_TYPE::MONOPROPELLANT; // LH2
		case 1: return PROPELLANT_TYPE::BIPROPELLANT_COMBINED; // RCS
		default: return PROPELLANT_TYPE::UNKNOWN_PROPELLANT;
		};
	case VMSG_GET_FUEL_TYPE:
		switch(prm)
		{
		case 0: return FUEL_TYPE::HYDROGEN_CRYOGENIC; // LH2
		case 1: return FUEL_TYPE::MONOMETHYLHYDRAZINE; // RCS
		default: return FUEL_TYPE::UNKNOWN_FUEL;
		};
	case VMSG_GET_OXIDIZER_TYPE:
		switch(prm)
		{
		case 0: return OXIDIZER_TYPE::UNKNOWN_OXIDIZER;
		case 1: return OXIDIZER_TYPE::NITROGEN_TETROXIDE; // RCS
		default: return OXIDIZER_TYPE::UNKNOWN_OXIDIZER;
		};
	case VMSG_GET_TANK_DRAWDOWN_SEQUENCE:
		switch(prm)
		{
		case 0: return 3; //
		case 1: return -1; // RCS
		default: return -1;
		};
	case VMSG_GET_POWER_OUTPUT_W:
		return 0;
	default: return -1;
	};
}

// disable thrusters by "dumping" RCS fuel into main tanks? 

void NTR_inlinetank::clbkDockEvent  ( int  dock,  OBJHANDLE  mate)
{
	if (this->DockingStatus(dock) == 0) // turn on the beacons
	{
		for(int i=0; i<NUM_BEACONS; i++) beacons[i].active = true;
	}
	else
	{
		for(int i=0; i<NUM_BEACONS; i++) beacons[i].active = false;
	};
}

void NTR_inlinetank::clbkPostCreation  (   ) 
{
	old_STP = this->GetAtmPressure();
//	shallFillDrawdownSeq = true;
	//FillDrawdownSequence();
	//ComputeOptimumRCS(200.0);
	//if (GetStackGroundContact() || (this->GetAltitude()<300.0)) 
	//{
	//	IsInLaunchMode = true;
	//	for (int i = 0; i<NUM_BEACONS; i++) beacons[i].active = false;
	//};
}


//bool InputRADec (void *id, char *str, void *usrdata)
//{
//	NTR_inlinetank* ntr = (NTR_inlinetank*)(usrdata);
//	if (ntr==NULL) return true;
//	float RA, Dec;
//	sscanf_s(str, "%g %g", &RA, &Dec);
//	ntr->TrackPointEulers = BrianJ_RADecToEulers((double)RA, (double)Dec);
//	ntr->ShallTrackPoint = true;
//	ntr->ShallTrackSun = false;
//	return true;
//}

bool InputConfigParam (void *id, char *str, void *usrdata)
{
	NTR_inlinetank* ntr = (NTR_inlinetank*)(usrdata);
	if (ntr==NULL) return true;
	int ctrlr = -1;
	char * nextChar = NULL;
	bool ModifyController = false;
	if (_strnicmp(str, "DEADBAND", 8)==0)
	{
		return true;
	}
	else if (_strnicmp(str, "RCS", 3)==0)
	{
		ModifyController = true;
		if (_strnicmp(str+4, "PITCH", 5)==0)
		{
			ctrlr = RCS_P;
			nextChar = str+10;
		}
		else if (_strnicmp(str+4, "YAW", 3)==0)
		{
			ctrlr = RCS_Y;
			nextChar = str+8;
		}
		else if (_strnicmp(str+4, "ROLL", 4)==0)
		{
			ctrlr = RCS_R;
			nextChar = str+9;
		}
		else
		{
			return true;
		};
	}
	else
	{
		return true;
	};

	if (ModifyController && ctrlr>=0 && ctrlr<NUM_CONTROLLERS)
	{
		if (_strnicmp(nextChar, "PROP", 4)==0)
		{
			//ntr->Controllers[ctrlr].proportionalGain = atof(nextChar+4);
		}
		else if (_strnicmp(nextChar, "INT", 3)==0)
		{
			//ntr->Controllers[ctrlr].integralGain = atof(nextChar+3);
		}
		else if (_strnicmp(nextChar, "DERIV", 5)==0)
		{
			//ntr->Controllers[ctrlr].integralGain = atof(nextChar+5);
		}
		else
		{
			return true;
		};
		//ntr->Controllers[ctrlr].Reset();
		return true;
	};
	return true;
}

void NTR_inlinetank::clbkSaveState  ( FILEHANDLE  scn   )  
{
	VESSEL3::clbkSaveState(scn);
	if (IsInLaunchMode)     oapiWriteLine(scn, "  LAUNCH_MODE");
	if (IsInCheckoutMode)   oapiWriteLine(scn, "  CHECKOUT_MODE");
//	if (MaintainCommandedRates)   oapiWriteLine(scn, "  RATE_DAP");
	char line[1024];
	for (int i = 0; i<NUM_ANIMATIONS; i++)
	{
		sprintf_s(line, 1023, "  ANIMATION %d %d %g", i, (int) AnimationStatus[i], AnimationDeployments[i]);
		oapiWriteLine(scn, line);
	};
}

void NTR_inlinetank::clbkLoadStateEx (FILEHANDLE scn, void *status)
{
	char *lineChars;
	int num_anim, anim_state;
	float anim_deployment;
//	StageState = reactor_state_t::Offline;
//	MaintainCommandedRates = false;
	while (oapiReadScenario_nextline (scn, lineChars)) {
		if (_strnicmp(lineChars, "LAUNCH_MODE", 11)==0)
		{
			this->IsInLaunchMode = true;
			this->IsInCheckoutMode = false;
		}
		else if (_strnicmp(lineChars, "CHECKOUT_MODE", 13)==0)
		{
			this->IsInLaunchMode = false;
			this->IsInCheckoutMode = true;
		}
		else if (_strnicmp(lineChars, "RATE_DAP", 8)==0)
		{
			//this->MaintainCommandedRates = true;
		}
		else if (_strnicmp(lineChars, "ANIMATION", 9)==0)
		{
			sscanf_s(lineChars+9, "%d %d %g", &num_anim, &anim_state, &anim_deployment);
			if (num_anim>=0 && num_anim<NUM_ANIMATIONS)
			{
				if (anim_state>=0 && anim_state<simple_anim_status_t::LAST)
				{
					if (anim_deployment>=0.0 && anim_deployment<=1.0)
					{
						AnimationStatus[num_anim] = (simple_anim_status_t) anim_state;
						AnimationDeployments[num_anim] = (double)anim_deployment;
					};
				};
			};
		}
		// todo: save gimbal state!
		else {          // anything not recognised is passed on to Orbiter
			ParseScenarioLineEx (lineChars, status);
		};
	};
}

// TODO: 1EO and 2EO optimum (zero-torque) gimbal setups

int NTR_inlinetank::clbkConsumeBufferedKey  ( DWORD  key,  bool  down,  char *  kstate )
{
	return 0;
	//if (IsInLaunchMode || IsInCheckoutMode) return 0; // no key processing if a payload or very soon after separation
	//if (KEYMOD_CONTROL(kstate))
	//{
	//	if (down && (key== OAPI_KEY_Y))
	//	{
	//		CurrentScreen = (CurrentScreen == ENGINE_CONTROL ? DAP_CONFIG : ENGINE_CONTROL);
	//		return 1;
	//	}
	//	else if (down && MaintainCommandedRates)
	//	{
	//		if (key==OAPI_KEY_1)
	//		{
	//			// toggle pitch DAP channel
	//			ActiveDAPChannels[PITCH] = (ActiveDAPChannels[PITCH]?false:true);
	//			return 1;
	//		}
	//		else if (key==OAPI_KEY_2)
	//		{
	//			// toggle yaw DAP channel
	//			ActiveDAPChannels[YAW] = (ActiveDAPChannels[YAW]?false:true);
	//			return 1;
	//		}
	//		else if (key==OAPI_KEY_3)
	//		{
	//			// toggle roll DAP channel
	//			ActiveDAPChannels[ROLL] = (ActiveDAPChannels[ROLL]?false:true);
	//			return 1;
	//		};
	//		return 0;
	//	}
	//	else return 0;// pass-through
	//};
	//if (KEYMOD_ALT(kstate))
	//{
	//	if ((key == OAPI_KEY_M) 
	//		&& StageState == reactor_state_t::Offline 
	//		&& down)
	//	{
	//		oapiOpenInputBox("Enter commanded dV (in m/s)",
	//			SetCommanded_dV,
	//			0,20, 
	//			(void*)this);
	//		return 1;
	//	}
	//	else if ((key == OAPI_KEY_C) && down)
	//	{
	//		oapiOpenInputBox("Enter configuration parameter",
	//			InputConfigParam,
	//			0,90, 
	//			(void*)this);
	//		return 1;
	//	}
	//	else if ((key == OAPI_KEY_S) && down)
	//	{
	//		ShallTrackSun = (ShallTrackSun ? false: true);
	//		if (ShallTrackSun) ShallTrackPoint = false;
	//		return 1;
	//	}
	//	else if ((key == OAPI_KEY_P) && down)
	//	{
	//		if (ShallTrackPoint)
	//		{
	//			ShallTrackPoint = false;
	//		}
	//		else
	//		{
	//			oapiOpenInputBox("Enter right ascension and declination of aimpoint",
	//				InputRADec,
	//				0,90, 
	//				(void*)this);
	//		};
	//		return 1;
	//	}
	//	else if (ARE_PUMPS_WORKING(StageState) && down)
	//	{
	//		if (key == OAPI_KEY_1) 
	//		{
	//			GimbalSelect[0] = (GimbalSelect[0] ? false : true);
	//			return 1;
	//		}
	//		else if (key == OAPI_KEY_2)
	//		{
	//			GimbalSelect[1] = (GimbalSelect[1] ? false : true);
	//			return 1;
	//		}
	//		else if (key == OAPI_KEY_3)
	//		{
	//			GimbalSelect[2] = (GimbalSelect[2] ? false : true);
	//			return 1;
	//		} 
	//		else if (key == OAPI_KEY_8)
	//		{
	//			for (int i = 0; i<NUM_REACTORS; i++)
	//			{
	//				GimbalSelect[i] = false;
	//			};
	//			return 1;
	//		}
	//		else if (key == OAPI_KEY_0)
	//		{
	//			for (int i = 0; i<NUM_REACTORS; i++)
	//			{
	//				GimbalSelect[i] = false;
	//				GimbalNTR(i, 0.0, 0.0);
	//			};
	//			return 1;
	//		}
	//		else if (key==OAPI_KEY_4)
	//		{
	//			GimbalRollNTR(0.01,-90);
	//			return 1;
	//		}
	//		else if (key==OAPI_KEY_5)
	//		{
	//			GimbalRollNTR(0.01,+90);
	//			return 1;
	//		};
	//	};
	//	// Alt-1 Alt-2 Alt-3 - select/deselect engine for gimballing 
	//	// Alt-8 deselect all 
	//	// Alt-0 deselect all and return all engines to default settings.
	//	if (key == OAPI_KEY_L && down)
	//	{
	//		MaintainCommandedRates = (MaintainCommandedRates?false:true);
	//		if (MaintainCommandedRates)
	//		{
	//			//re-do the optimization
	//			FillDrawdownSequence_and_OptimizeRCS();
	//		};
	//		return 1;
	//	};
	//	return 0;
	//};
	//// test solar1
	//if (key == OAPI_KEY_K && down && !KEYMOD_CONTROL(kstate)) 
	//{
	//	if (AnimationStatus[SOLAR1_ANIMATION]==simple_anim_status_t::CLOSED) 
	//	{
	//		AnimationStatus[SOLAR1_ANIMATION]=simple_anim_status_t::OPENING;
	//	}
	//	else if (AnimationStatus[SOLAR1_ANIMATION]==simple_anim_status_t::OPEN)
	//	{
	//		AnimationStatus[SOLAR1_ANIMATION]=simple_anim_status_t::CLOSING;
	//	};
	//	if (AnimationStatus[SOLAR2_ANIMATION]==simple_anim_status_t::CLOSED) 
	//	{
	//		AnimationStatus[SOLAR2_ANIMATION]=simple_anim_status_t::OPENING;
	//	}
	//	else if (AnimationStatus[SOLAR2_ANIMATION]==simple_anim_status_t::OPEN)
	//	{
	//		AnimationStatus[SOLAR2_ANIMATION]=simple_anim_status_t::CLOSING;
	//	};
	//	return 1;
	//};
	//if (MaintainCommandedRates && down)
	//{
	//	if (key == OAPI_KEY_DOWN)
	//	{
	//		CommandedRates.data[PITCH] += 0.001;
	//		return 1;
	//	}
	//	else if (key == OAPI_KEY_UP)
	//	{
	//		CommandedRates.data[PITCH] -= 0.001;
	//		return 1;
	//	}
	//	else if (key == OAPI_KEY_LEFT)
	//	{
	//		CommandedRates.data[YAW] += 0.001;
	//		return 1;
	//	}
	//	else if (key == OAPI_KEY_RIGHT)
	//	{
	//		CommandedRates.data[YAW] -= 0.001;
	//		return 1;
	//	}
	//	else if (key == OAPI_KEY_INSERT)
	//	{
	//		CommandedRates.data[ROLL] -= 0.001;
	//		return 1;
	//	}
	//	else if (key == OAPI_KEY_DELETE)
	//	{
	//		CommandedRates.data[ROLL] += 0.001;
	//		return 1;
	//	};
	//};

	//if (ARE_PUMPS_WORKING(StageState) && down)
	//{
	//	if (!MaintainCommandedRates)
	//	{
	//		double dPitch = 0.0;
	//		double dYaw = 0.0;
	//		if (key == OAPI_KEY_DOWN)
	//		{
	//			dPitch = 0.1;
	//		}
	//		else if (key == OAPI_KEY_UP)
	//		{
	//			dPitch = -0.1;
	//		}
	//		else if (key == OAPI_KEY_LEFT)
	//		{
	//			dYaw = -0.1;
	//		}
	//		else if (key == OAPI_KEY_RIGHT)
	//		{
	//			dYaw = 0.1;
	//		}
	//		else return 0;
	//		for (int i = 0; i<NUM_REACTORS; i++)
	//		{
	//			if (GimbalSelect[i])
	//			{
	//				GimbalNTR(i, NTRGimbal[i][0]+dPitch, NTRGimbal[i][1]+dYaw);
	//			};
	//		};
	//		return 1;//no pass-through!
	//	};
	//};
	//return 0;
}


//double GetThrust()
//{
//	return 0.0;
//}

//time. stage. dT

// null rates, counter roll
// are the gimbal limits enough to prevent 1EO tumbling?

// att error - rate error phase plane. 

void NTR_inlinetank::SelfDestruct(const char * reason)
{
	char line[1024];
	sprintf_s(line, 1023, "%s self-destructed because %s", this->GetName(), reason);
	oapiWriteLog(line);
	oapiDeleteVessel(this->GetHandle());
}

void NTR_inlinetank::clbkPreStep (double simt, double simdt, double mjd)
{
	for (int i = 0; i<NUM_ANIMATIONS; i++)
	{
		if (AnimationStatus[i] == simple_anim_status_t::CLOSING || AnimationStatus[i] == simple_anim_status_t::OPENING)
		{
			double da = simdt*AnimationSpeeds[i];
			if (AnimationStatus[i]==simple_anim_status_t::CLOSING)
			{
				if (AnimationDeployments[i]>0.0)
				{
					AnimationDeployments[i] = max(0.0, AnimationDeployments[i]-da);
				}
				else
				{
					AnimationStatus[i]=simple_anim_status_t::CLOSED;
				};
			}
			else
			{
				if (AnimationDeployments[i]<1.0)
				{
					AnimationDeployments[i] = min(1.0, AnimationDeployments[i]+da);
				}
				else
				{
					AnimationStatus[i]=simple_anim_status_t::OPEN;
				};
			};
			SetAnimation(Animations[i], AnimationDeployments[i]);
		};
	};

	if (IsInLaunchMode)
	{
		// special processing for launches.
		// TODO: possibly add unlocking?
		// all thrusters are disabled.
		AnimationStatus[SOLAR1_ANIMATION] = simple_anim_status_t::CLOSED; // wrong!
		AnimationStatus[SOLAR2_ANIMATION] = simple_anim_status_t::CLOSED;
		AnimationDeployments[SOLAR1_ANIMATION] = 0.0;
		AnimationDeployments[SOLAR2_ANIMATION] = 0.0;

		for (unsigned int i =0; i< this->GetThrusterCount(); i++)
		{
			if (this->GetThrusterLevel(this->GetThrusterHandleByIndex(i))>0.0) this->SetThrusterLevel(this->GetThrusterHandleByIndex(i),0.0);
			// not sure it works!
		};

		// self-destruct criteria
		VECTOR3 forces, gravity;
		this->GetForceVector(forces);
		this->GetWeightVector(gravity);
		forces = forces-gravity;
		double ownMass = this->GetMass();
		VECTOR3 acc = forces/ownMass;
		double ownG = length(acc)/G;
		if (ownG> LAUNCH_G_LIMIT) 
		{
			G_limit_criterion_counter++;
		}
		else
		{
			G_limit_criterion_counter = 0;
		};
		if (this->G_limit_criterion_counter>=LAUNCH_G_STEP_LIMIT)	SelfDestruct("g limit exceeded");
		double STP = this->GetAtmPressure();
		double dP_dt = (STP-old_STP)/simdt;
		old_STP = STP;
		if (dP_dt>LAUNCH_STATIC_dP_dt_THRESHOLD) 
		{
			Static_dP_dt_criterion_counter++;
		}
		else
		{
			Static_dP_dt_criterion_counter=0;
		};
		if (this->Static_dP_dt_criterion_counter>=LAUNCH_STATIC_dP_dt_STEP_LIMIT) SelfDestruct("static pressure increased");

		//check for transition to checkout
		OBJHANDLE hEarth = oapiGetGbodyByName("Earth");
		if (hEarth==NULL) 
		{
			oapiWriteLog("No Earth in the system");
			throw "No Earth detected";
		};
		double EarthR = oapiGetSize(hEarth);
		double PeA, PeD, ApD, ApA;
		if (this->GetPeDist(PeD)!=hEarth)
		{
			oapiWriteLog("Not an Earth orbit");
			throw "Not an Earth orbit";
		};
		if (this->GetApDist(ApD)!=hEarth)
		{
			oapiWriteLog("Not an Earth orbit");
			throw "Not an Earth orbit";
		};
		PeA = (PeD - EarthR)/1000.0;
		ApA = (ApD - EarthR)/1000.0;
		if ((PeA>=SAFE_PeA) && (ApA<=SAFE_ApA)) 
		{
			IsInLaunchMode = false;
			IsInCheckoutMode = true;
		};
		return;
	};

	if (IsInCheckoutMode)
	{
		// placeholder code
		if (AnimationStatus[SOLAR1_ANIMATION] == simple_anim_status_t::OPEN
			&& AnimationStatus[SOLAR2_ANIMATION] == simple_anim_status_t::OPEN)
		{
			IsInCheckoutMode = false;
			CurrentScreen = ENGINE_CONTROL;
			return;
		};
		// deploy PVAs
		AnimationStatus[SOLAR1_ANIMATION] = simple_anim_status_t::OPENING;
		AnimationStatus[SOLAR2_ANIMATION] = simple_anim_status_t::OPENING;
		for (int i = 0; i<NUM_BEACONS; i++) beacons[i].active = true;
		return;
	};

	//Sanity check: if we are at 1 kPa DNP mark, disintegrate. KINDA FOOLISH. How do I take off from earth?
	// first off, check for ground contact at start? 
	if (this->GetDynPressure()>=MAX_DYN_PRESSURE)
	{
		SelfDestruct("dynamic pressure limit exceeded");
		return;
	};

	//todo: tank temperature? n mli for boiloff calcs?? 

	//TODO: how do we notify the IMFD of the vessel's true capabilities?

	// do the PMI and CoG change

	// test colors:
	if (visual!=NULL)
	{
		MESHHANDLE hMesh = GetMesh(visual, MESH_MAIN);
		if (hMesh!=NULL)
		{
			// TODO: interpolate between the initial and the heated color (read the color at Visual creation!)
			//MATERIAL* pMat = oapiMeshMaterial(hMesh, MAT_EngineMat);
			//if (pMat!=NULL && currentThrust>0.0)
			//{
			//	pMat->emissive.r = (float)(currentThrust/NTR_MAXTHRUST);
			//	pMat->emissive.b = 1.0f-pMat->emissive.r;
			//	pMat->emissive.g = (pMat->emissive.r + pMat->emissive.b)/2.0f;
			//};
		};
	};

	double LH2Mass = this->GetPropellantMass(hpr_LH2);
	double fuelMass = LH2Mass + this->GetPropellantMass(hpr_RCS);
	double maxMass = PB_LH2MASS + PB_RCSFUELMASS;
	double massRatio = fuelMass/maxMass;
	//interpolate PMI
	this->SetPMI(PB_PMI_FULL*massRatio+PB_PMI_EMPTY*(1-massRatio));
	//double currentCOG_elev = this->GetCOG_elev();
	//double tgtCOG_elev = fullCOG_elev*massRatio+emptyCOG_elev*(1-massRatio);
	//sprintf_s(oapiDebugString(), 255, "COG: %g m Tgt: %g m", currentCOG_elev, tgtCOG_elev);
	//this->ShiftCG(_V(0,0,(tgtCOG_elev-currentCOG_elev)));
	// how do I shift CoG?

	VECTOR3 angVel;
	GetAngularVel(angVel);
	VECTOR3 CurrentRates=angVel*DEG;

	/*if (ShallTrackSun)
	{
		MaintainCommandedRates = true;
		VECTOR3 SunwardPYR, SunwardEulers;
		OBJHANDLE hSun = oapiGetGbodyByName("Sun");
		if (hSun!=NULL)
		{
			SunwardPYR = CalcTargetPYR(this, hSun, SunwardEulers);
			CommandedRates.data[AXIS::PITCH] = RateFromAttError(SunwardPYR.data[AXIS::PITCH]*DEG, DB_FINE);
			CommandedRates.data[AXIS::YAW] = RateFromAttError(SunwardPYR.data[AXIS::YAW]*DEG, DB_FINE);
			CommandedRates.data[AXIS::ROLL] = RateFromAttError(CurrentRates.data[AXIS::ROLL], DB_MEDIUM);
		};
	}
	else if (ShallTrackPoint)
	{
		MaintainCommandedRates = true;
		VECTOR3 TrackPointPYR;
		TrackPointPYR = CalcPitchYawRollAngles(this, _V(0,0,1), TrackPointEulers);
		CommandedRates.data[AXIS::PITCH] = RateFromAttError(TrackPointPYR.data[AXIS::PITCH]*DEG, DB_FINE);
		CommandedRates.data[AXIS::YAW] = RateFromAttError(TrackPointPYR.data[AXIS::YAW]*DEG, DB_FINE);
		CommandedRates.data[AXIS::ROLL] = RateFromAttError(CurrentRates.data[AXIS::ROLL], DB_MEDIUM);
	};*/

	//if (MaintainCommandedRates)
	//{
	//	// control pitch and yaw only - actually, can control roll by symmetric rotation

	//	double pitchDelta = CommandedRates.data[AXIS::PITCH]-CurrentRates.data[AXIS::PITCH];
	//	double yawDelta = CommandedRates.data[AXIS::YAW]-CurrentRates.data[AXIS::YAW];
	//	double rollDelta = CommandedRates.data[AXIS::ROLL]-CurrentRates.data[AXIS::ROLL];

	//	double dRoll = Controllers[RCS_R].Step(rollDelta, simdt);
	//	double dPitch, dYaw;
	//	if (ActiveDAPChannels[ROLL])
	//	{
	//		if (dRoll>0.0)
	//		{
	//			this->SetThrusterGroupLevel_SingleStep(THGROUP_ATT_BANKRIGHT, dRoll); // can be extended to support RCS strings!
	//		}
	//		else
	//		{
	//			this->SetThrusterGroupLevel_SingleStep(THGROUP_ATT_BANKLEFT, -dRoll);
	//		};
	//	};
	//	if (ARE_PUMPS_WORKING(StageState))
	//	{
	//		dPitch = (ActiveDAPChannels[PITCH] ? Controllers[TVC_P].Step(pitchDelta, simdt) : 0.0);
	//		dYaw = (ActiveDAPChannels[YAW] ? Controllers[TVC_Y].Step(yawDelta, simdt) : 0.0);
	//		for (int i = 0; i<NUM_REACTORS; i++) GimbalNTR(i, dPitch, dYaw);
	//	}
	//	else // RCS-based DAP
	//	{
	//		dPitch = Controllers[RCS_P].Step(pitchDelta, simdt);
	//		dYaw = Controllers[RCS_Y].Step(yawDelta, simdt);
	//		if (ActiveDAPChannels[PITCH])
	//		{
	//			if (dPitch>0.0)
	//			{
	//				this->SetThrusterGroupLevel_SingleStep(THGROUP_ATT_PITCHUP, dPitch);
	//			}
	//			else
	//			{
	//				this->SetThrusterGroupLevel_SingleStep(THGROUP_ATT_PITCHDOWN, -dPitch);
	//			};
	//		};
	//		if (ActiveDAPChannels[YAW])
	//		{
	//			if (dYaw>0.0)
	//			{
	//				this->SetThrusterGroupLevel_SingleStep(THGROUP_ATT_YAWLEFT, dYaw);
	//			}
	//			else
	//			{
	//				this->SetThrusterGroupLevel_SingleStep(THGROUP_ATT_YAWRIGHT, -dYaw);
	//			};
	//		};
	//	};
	//};

	// TODO: add RCS string + TVC

	//if (ARE_PUMPS_WORKING(StageState)) 
	//{
	//	if (LH2Mass<=0.0)
	//	{
	//		StageState = reactor_state_t::Meltdown;
	//		currentThrust = 0.0;
	//		reactorPeriod_s = -ShutdownPeriod_s;
	//	}
	//	else if (StageState == reactor_state_t::Shutdown)
	//	{
	//		reactorPeriod_s = -ShutdownPeriod_s;
	//		currentThrust = currentThrust*exp(simdt/reactorPeriod_s);
	//		currentThrust = max(0.0,min(NTR_MAXTHRUST, currentThrust));
	//		if (currentThrust<=ZeroThrustLevel)
	//		{
	//			currentThrust = 0.0;
	//			StageState = reactor_state_t::Offline;
	//			this->SetThrusterGroupLevel(THGROUP_MAIN, 0.0);
	//			XePoisoning = 1.0; // no way you can work with it!
	//			commanded_dV = 0.0; //re-set dV
	//		};
	//	}
	//	else if (StageState == reactor_state_t::Startup)
	//	{
	//		reactorPeriod_s = StartupPeriod_s;
	//		if (currentThrust==0.0) {currentThrust = ZeroThrustLevel;} else currentThrust = currentThrust*exp(simdt/reactorPeriod_s);
	//		currentThrust = max(0.0,min(NTR_MAXTHRUST, currentThrust));
	//		if (currentThrust>=NTR_MAXTHRUST)
	//		{
	//			StageState = reactor_state_t::Nominal;
	//		};
	//	}
	//	else if (StageState == reactor_state_t::Nominal)
	//	{
	//		currentThrust = NTR_MAXTHRUST;
	//		reactorPeriod_s = numeric_limits<double>::infinity();
	//	}
	//	else
	//	{
	//		currentThrust = 0.0;
	//	};
	//	for (int i = 0; i<NUM_REACTORS; i++)
	//	{
	//		SetThrusterLevel_SingleStep(th_NTR[i], currentThrust/NTR_MAXTHRUST);
	//	};
	//	//if (MaintainCommandedRates)
	//	//{
	//	//	// control pitch and yaw only - actually, can control roll by symmetric rotation
	//	//	VECTOR3 angVel;
	//	//	GetAngularVel(angVel);
	//	//	VECTOR3 CurrentRates=angVel*DEG;

	//	//	double pitchDelta = CommandedRates.data[AXIS::PITCH]-angVel.data[AXIS::PITCH];
	//	//	double yawDelta = CommandedRates.data[AXIS::YAW]-angVel.data[AXIS::YAW];
	//	//	double rollDelta = CommandedRates.data[AXIS::ROLL]-angVel.data[AXIS::ROLL];

	//	//	double dRoll = RCSControlR.Step(rollDelta, simdt);
	//	//	double dPitch=TVCControlP.Step(pitchDelta, simdt);
	//	//	double dYaw=TVCControlY.Step(yawDelta, simdt);
	//	//	if (dRoll>0.0)
	//	//	{
	//	//		this->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_BANKRIGHT, dRoll);
	//	//	}
	//	//	else
	//	//	{
	//	//		this->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_BANKLEFT, dRoll);
	//	//	};
	//	//	for (int i = 0; i<NUM_REACTORS; i++) GimbalNTR(i, dPitch, dYaw);
	//	//};
	//	//if (NullTorque)
	//	//{
	//	////how do we do TVC? absolutely necessary for 1EO cases.
	//	// how do we null gravity-gradient torque?
	//	//vector<stack_thruster_t> thrusters;
	//	//vector<stack_fuel_tank_t> tanks;
	//	//VECTOR3 stack_COG, stack_PMI;
	//	//TransformStackThrusterInfoIntoRootFrame(stackInfo, 
	//	//										thrusters, 
	//	//										tanks, 
	//	//										stack_COG, 
	//	//										stack_PMI);

	//	//VECTOR3 torque = _V(0,0,0);
	//	//VECTOR3 torque2 = _V(0,0,0);
	//	//unsigned int th_indices[NUM_REACTORS];
	//	//unsigned int k = 0;
	//	//for (unsigned int i = 0; i<thrusters.size(); i++)
	//	//{
	//	//	if (thrusters[i].pV == this)
	//	//	{
	//	//		for (unsigned int j = 0; j<NUM_REACTORS; j++)
	//	//		{
	//	//			if (this->GetThrusterHandleByIndex(thrusters[i].thrusterID)==th_NTR[j])
	//	//			{
	//	//				th_indices[k] = i;
	//	//				k++;
	//	//				break;
	//	//			};
	//	//		};
	//	//	};
	//	//};
	//	//for (unsigned int i = 0; i<NUM_REACTORS; i++)
	//	//{
	//	//	stack_thruster_t& st = thrusters.at(th_indices[i]);
	//	//    torque = torque-crossp(st.root_pos-stack_COG, currentThrust*st.root_dir);
	//	//};
	//	// linear uncoupled regulator:
	//	//	VECTOR3 currAngVel;
	//	//	this->GetAngularVel(currAngVel);
	//	//	if (fabs(currAngVel.y)>1e-10)
	//	//	{
	//	//		VECTOR3 ALICE_Y = _V(0.0, -0.01*currAngVel.y, sqrt(1.0-0.01*currAngVel.y*0.01*currAngVel.y));
	//	//		//normalise(ALICE_Y);
	//	//		this->SetThrusterDir(th_NTR[NTR_Alice], ALICE_Y);
	//	//	};
	//	//	for (unsigned int i = 0; i<NUM_REACTORS; i++)
	//	//	{
	//	//		VECTOR3 dir, pos;
	//	//		this->GetThrusterDir(th_NTR[i], dir);
	//	//		this->GetThrusterRef(th_NTR[i], pos);
	//	//	    torque2 = torque2-crossp(pos, currentThrust*dir);
	//	//	};
	//	//	//sprintf_s(oapiDebugString(), 255, "engine-induced torque x: %g y: %g z: %g", torque.x, torque.y, torque.z);
	//	//	sprintf_s(oapiDebugString(), 255, "engine-induced torque x: %g y: %g z: %g", torque2.x, torque2.y, torque2.z);

	//	//}
	//	//else
	//	//{
	//	//	for (int i = 0; i<NUM_REACTORS; i++)
	//	//	this->SetThrusterDir(th_NTR[i], _V(0,0,1));
	//	//};

	//}
	//else
	//{
	//	for (int i = 0; i<NUM_REACTORS; i++)
	//	{
	//		SetThrusterLevel_SingleStep(th_NTR[i], 0.0);
	//	};
	//	if (XePoisoning>0.0) XePoisoning = XePoisoning*exp(-HALFLIFE_CONSTANT*simdt/Xe135_HalfLife_s);
	//};
}


//VECTOR3 cTorque(const double Thrust, const VECTOR3 POS, const VECTOR3 COG, const double Pitch, const double Yaw)
//{
//	VECTOR3 R;
//	VECTOR3 dir = _V(0, 0, 1);
//	dir = RotateVectorX(dir, -Pitch);
//	dir = RotateVectorY(dir, -Yaw);
//	VECTOR3 shoulder = POS - COG;
//	VECTOR3 thV = Thrust * dir;
//	R = crossp(shoulder, thV);
//	return R;
//}


//double totalTorque(unsigned n, const double *x,
//				   double *gradient, /* NULL if not needed */
//				   void *func_data)
//{
//	gimbal_problem_data_t * gpdt = (gimbal_problem_data_t*) func_data;
//	if (n!=gpdt->numThrusters*2)
//	{
//		oapiWriteLog("n!=numThrusters*2");
//		throw "n!=numThrusters*2";
//	};
//	double ff;
//	VECTOR3 Torq = _V(0, 0, 0);
//	for (int i=0; i<gpdt->numThrusters; i++)
//	{
//		Torq = Torq + cTorque(gpdt->Thrust[i], gpdt->Pos[i], gpdt->COG, (*(x+2 * i)), (*(x+2 * i + 1)));
//	};
//	ff = Torq.x * Torq.x + Torq.y * Torq.y + Torq.z * Torq.z;
//	gradient = NULL;
//	return ff;
//}



//WORKS!!!
//void NTR_inlinetank::OptimizeTorque(const vector<double> & thrust, vector<double> &pitch, vector<double> &yaw)
//{
//	// TODO: make optimization problem a private member of the class, to avoid allocation/deallocation
//	if (thrust.size()!=NUM_REACTORS)
//	{
//		oapiWriteLog("thrust vector sized incorrectly");
//		throw "thrust vector sized incorrectly";
//	};
//	vector<double> angles;
//	angles.resize(NUM_REACTORS*2);
//	double residTorqueMag;
//	gimbal_problem_data_t gpdt;
//	gpdt.numThrusters = NUM_REACTORS;
//	gpdt.Pos.resize(NUM_REACTORS);
//	//// need COG data from the stack!
//	//stackInfo
//	//th_NTR[1]
//	gpdt.Pos[0] = V_Alice;
//	gpdt.Pos[1] = V_Betty;
//	gpdt.Pos[2] = V_Clara;
//	//// TODO: make generic?
//	gpdt.Thrust.resize(NUM_REACTORS);
//	for (int i=0; i<NUM_REACTORS; i++) gpdt.Thrust[i] = thrust[i];
//	gpdt.COG = _V(0, 0, 2);
//	for (int i = 0; i<NUM_REACTORS*2; i++) angles[i] = 0.0;
//	optProblem->set_lower_bounds(-NTR_PITCH_GIMBAL_LIMIT);
//	optProblem->set_upper_bounds(NTR_PITCH_GIMBAL_LIMIT);
//	optProblem->set_min_objective(totalTorque, &gpdt);
//	nlopt::result res = optProblem->optimize(angles, residTorqueMag);
//	if (res==nlopt::result::SUCCESS || res==nlopt::result::FTOL_REACHED || res==nlopt::result::XTOL_REACHED )
//	{
//		for (int i = 0; i<NUM_REACTORS; i++)
//		{
//			pitch[i] = angles[2*i];
//			yaw[i] = angles[2*i+1];
//		};
//	//	char lx[1024];
//	//	sprintf_s(lx, 1023, "opt %g - res %d: %g %g %g %g %g %g", residTorqueMag, res, angles[0],angles[1],angles[2],angles[3],angles[4],angles[5]);
//	//	oapiWriteLog(lx);
//	}
//	else // fallback?
//	{
//		for (int i = 0; i<NUM_REACTORS; i++)
//		{
//			pitch[i] = 0.0;
//			yaw[i] = 0.0;
//		};
//	};
//
//	//else
//	//{
//	//	char lx[1024];
//	//	sprintf_s(lx, 1023, "res: %d", res);
//	//	oapiWriteLog(lx);
//	//}
//}

//bool NTR_inlinetank::GimbalRollNTR(double R, double theta_deg)
//{
//	// how do we enforce gimbal limits?
//	if (abs(R)<sin(NTR_PITCH_GIMBAL_LIMIT*RAD))
//	{
//
//		VECTOR3 dirs[3];
//		VECTOR3 pyrs[3];
//		for (int i = 0; i<3; i++)
//		{
//			dirs[i].x = R*sin((i*120.0+theta_deg)*RAD);
//			dirs[i].y = R*cos((i*120.0+theta_deg)*RAD);
//			//dirs[i].x = R*cos((i*120.0+theta_deg)*RAD);
//			//dirs[i].y = R*sin((i*120.0+theta_deg)*RAD);
//			dirs[i].z = sqrt(1.0-dirs[i].x*dirs[i].x-dirs[i].y*dirs[i].y);
//			pyrs[i] = GetPitchYawRoll(dirs[i], NULL_VECTOR); // smth happening? maybe we start from the wrong heading?
//			NTRGimbal[i][0]=pyrs[i].data[AXIS::PITCH];
//			NTRGimbal[i][1]=pyrs[i].data[AXIS::YAW];
//
//			//VECTOR3 dir = _V(0,0,1);
//
//			//dir=RotateVectorX(dir, -pitch); //positive OMS gimbal directs thrust downwards
//			//dir=RotateVectorY(dir, -yaw); //positive yaw gimbal directs thrust to right
//			SetThrusterDir(th_NTR[i], dirs[i]);
//		};
//
//		//char cbuf[255];
//		//sprintf_s(cbuf, 255, "NTR DIR: %d %f %f %f", engine, dir.x, dir.y, dir.z);
//		//oapiWriteLog(cbuf);
//		return true;
//	}
//	else return false;
//}


//// TODO: visual heating up of the engine nozzles by changing in-place the material definition.
//
//double NTR_inlinetank::RateFromAttError(const double AttError, const double Deadband)
//{
//	// need current rate as an input!
//	if (fabs(AttError)<Deadband-0.3*DB_OFFSET) return 0.0; // overshoot possible
//	if (fabs(AttError)>20.0*Deadband) return (AttError>0.0?-RATE_SLOW:RATE_SLOW);
//	if (fabs(AttError)>5.0*Deadband) return (AttError>0.0?-RATE_SLOWEST:RATE_SLOWEST);
//	double slope = (-RATE_SLOWEST+0.001)/(4.0*Deadband);
//	const double SUPERSLOW = 0.005;
//	double intercept = -RATE_SLOWEST*0.25+SUPERSLOW*1.2;
//	return intercept+slope*AttError; //(AttError>0.0? -intercept-slope*AttError:intercept+slope*AttError);
//}

// engine status display
bool NTR_inlinetank::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);


	oapiSetDefNavDisplay(1);
	oapiSetDefRCSDisplay(1);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
		roxl = (lw*10);
		royl = (int) (-88*d);
	}


	// Time to burn. sync.

	if (IsInLaunchMode)
	{
		CurrentScreen = glass_screen_t::LAUNCH;
	}
	else if (IsInCheckoutMode)
	{
		CurrentScreen = glass_screen_t::CHECKOUT;
	};


	switch (CurrentScreen)
	{
	case LAUNCH:
		{
			HUDTextOut_CLR(skp, 10, lineY[0], GREY, "Reactors are safed. Vehicle is in LAUNCH MODE.");
			HUDTextOut_CLR(skp, 10, lineY[1], GREY, "Self-destruct criteria: static dP/dt = + 0.1kPa/second for %d steps OR acc>7g for %d steps", LAUNCH_STATIC_dP_dt_STEP_LIMIT, LAUNCH_G_STEP_LIMIT);
			HUDTextOut_CLR(skp, 10, lineY[2], GREY, "Safe parking orbit: PeA>=%d km ApA<=%d km", (int)SAFE_PeA, (int)SAFE_ApA);
			return true;
		};
	case CHECKOUT:
		{
			HUDTextOut_CLR(skp, 10, lineY[0], GREY, "Reactors are safed. Vehicle is in CHECKOUT MODE.");
			return true;
		};
	case ENGINE_CONTROL:
		{
			HUDTextOut_CLR(skp, 10, lineY[0], GOLD, "[Ctrl-K]-shutdown [Alt-M]-required dV [Alt-123]-toggle engine TVC [Alt-8]-deselect all TVC [Alt-0]-reset TVC");

//			if (StageState == reactor_state_t::UllageBurn) HUDTextOut_CLR(skp, 10, lineY[1], GOLD, "Ullage burn");
//			if (StageState == reactor_state_t::Startup)    HUDTextOut_CLR(skp, 10, lineY[1], GOLD, "Startup initiated");
//			if (StageState == reactor_state_t::Shutdown) 
			//{
			//	if (FuelLevelTripActivated)
			//	{
			//		HUDTextOut_CLR(skp, 10, lineY[1], RED, "Shutdown initiated (FUEL LOW)");
			//	}
			//	else
			//	{
			//		HUDTextOut_CLR(skp, 10, lineY[1], GOLD, "Shutdown initiated");
			//	};
			//};
			//if (StageState == reactor_state_t::Meltdown)
			//{
			//	HUDTextOut_CLR(skp, 10, lineY[3], RED, "[ALICE] MELTDOWN"); // what to do with the idiot?
			//	HUDTextOut_CLR(skp, 10, lineY[4], RED, "[BETTY] MELTDOWN");
			//	HUDTextOut_CLR(skp, 10, lineY[5], RED, "[CLARA] MELTDOWN");
			//}
			//else if (StageState != reactor_state_t::Offline && StageState != reactor_state_t::UllageBurn)
			//{
			//	// 1 degree is too much for TVC adjustment!
			//	// tpa can be ok or cross-feeding the neighbour or down:

			//	// ok XfA XfB XfC OUT
			//	double flowRate = this->GetPropellantFlowrate(this->GetPropellantHandleByIndex(0))/3.0;
			//	HUDTextOut_CLR(skp, 10, lineY[2], GOLD, "Reactor  Thrust:  LH2 flow:   Period: TPA1 TPA2 Pitch   Yaw   TVC");
			//	HUDTextOut_CLR(skp, 10, lineY[3], GOLD, "[ALICE] %5d N   %3.1f kg/s   %3.0f s    ok   ok   %+1.1f   %+1.1f    %d", (int)currentThrust, flowRate, reactorPeriod_s, NTRGimbal[NTR_Alice][0], NTRGimbal[NTR_Alice][1], (int)(GimbalSelect[NTR_Alice]));
			//	HUDTextOut_CLR(skp, 10, lineY[4], GOLD, "[BETTY] %5d N   %3.1f kg/s   %3.0f s    ok   ok   %+1.1f   %+1.1f    %d", (int)currentThrust, flowRate, reactorPeriod_s, NTRGimbal[NTR_Betty][0], NTRGimbal[NTR_Betty][1], (int)(GimbalSelect[NTR_Betty]));
			//	HUDTextOut_CLR(skp, 10, lineY[5], GOLD, "[CLARA] %5d N   %3.1f kg/s   %3.0f s    ok   ok   %+1.1f   %+1.1f    %d", (int)currentThrust, flowRate, reactorPeriod_s, NTRGimbal[NTR_Clara][0], NTRGimbal[NTR_Clara][1], (int)(GimbalSelect[NTR_Clara]));
			//	double currentMass = this->GetStackMass();
			//	dV_sinceBurnStart = PB_NTR_ISP*log(massAtBurnStart/currentMass );
			//	// what is the final mass assuming we stop burn now?
			//	if (currentThrust>0.0)
			//	{
			//		double timeToShutdown = -log(ZeroThrustLevel/currentThrust)*ShutdownPeriod_s;
			//		double fuelMassToShutdown = (3.0*ShutdownPeriod_s*currentThrust/PB_NTR_ISP)*(1.0-exp(-timeToShutdown/ShutdownPeriod_s));
			//		double dV_atShutdown = PB_NTR_ISP*log(massAtBurnStart/(currentMass-fuelMassToShutdown));

			//		if (StageState == reactor_state_t::Shutdown)
			//		{
			//			HUDTextOut_CLR(skp, 10, lineY[7], GOLD, "dV since burn start: %10.2f final dV: %10.2f", dV_sinceBurnStart, dV_atShutdown);
			//		}
			//		else
			//		{
			//			HUDTextOut_CLR(skp, 10, lineY[7], GOLD, "dV since burn start: %10.2f final dV if shutdown now: %10.2f", dV_sinceBurnStart, dV_atShutdown);
			//		};
			//		VECTOR3 GlobalVelocity;
			//		this->GetGlobalVel(GlobalVelocity);
			//		HUDTextOut_CLR(skp, 10, lineY[8], GREEN, "VX: %10.2f VY: %10.2f VZ: %10.2f", GlobalVelocity.x, GlobalVelocity.y, GlobalVelocity.z);
			//	}
			//	else
			//	{			
			//		HUDTextOut_CLR(skp, 10, lineY[7], GOLD, "dV since burn start: %10.2f", dV_sinceBurnStart);
			//	};
			//}
			//else
			//{
			//	COLORREF XeColor = (XePoisoning<=MaxAllowedXe135Poisoning?GREEN:RED);
			//	HUDTextOut_CLR(skp, 10, lineY[3], XeColor, "[ALICE] Xe-135 %d", (int)(XePoisoning*100.0)); // concentration of xenon? or simply time to next burn?
			//	HUDTextOut_CLR(skp, 10, lineY[4], XeColor, "[BETTY] Xe-135 %d", (int)(XePoisoning*100.0));
			//	HUDTextOut_CLR(skp, 10, lineY[5], XeColor, "[CLARA] Xe-135 %d", (int)(XePoisoning*100.0));

			//};
			//VECTOR3 Torque;
			//this->GetTorqueVector(Torque);
			//HUDTextOut_CLR(skp, 10, lineY[14], GREEN, "Torque X: %+10.9f Y: %+10.9f Z: %+10.9f", Torque.x, Torque.y, Torque.z);
			//char pitchChannel[2],yawChannel[2], rollChannel[2];
			//pitchChannel[1] = yawChannel[1] = rollChannel[1] = '\0';
			//pitchChannel[0] = (ActiveDAPChannels[PITCH]? '*' : '~');
			//yawChannel[0] = (ActiveDAPChannels[YAW]? '*' : '~');
			//rollChannel[0] = (ActiveDAPChannels[ROLL]? '*' : '~');
			//if (MaintainCommandedRates)
			//{
			//	HUDTextOut_CLR(skp, 10, lineY[15], GREEN, "[Alt-L] - rate DAP ON [P%s %+2.3f/s Y%s %+2.3f/s R%s %+2.3f/s] [Ctrl-123]-toggle channels", pitchChannel, CommandedRates.data[PITCH], yawChannel, CommandedRates.data[YAW], rollChannel, CommandedRates.data[ROLL]);
			//}
			//else
			//{
			//	HUDTextOut_CLR(skp, 10, lineY[15], GREEN, "[Alt-L] - ATT FREE; DAP is off ... [Ctrl-Y] - switch to DAP CONFIG");
			//};


			//VECTOR3 SunwardPYR, SunwardEulers;
			//OBJHANDLE hSun = oapiGetGbodyByName("Sun");
			//if (hSun!=NULL)
			//{
			//	SunwardPYR = CalcTargetPYR(this, hSun, SunwardEulers);
			//	if (ShallTrackSun)
			//	{
			//		HUDTextOut_CLR(skp, 10, lineY[16], GOLD, "Sunward attitude errors P:%+3.2f Y:%+3.2f [Alt-S]-no Sun track {TRACKING SUN} [Alt-P]-track point", SunwardPYR.data[PITCH]*DEG, SunwardPYR.data[YAW]*DEG);
			//	}
			//	else if (ShallTrackPoint)
			//	{
			//		VECTOR3 TrackPYR;
			//		TrackPYR = CalcPitchYawRollAngles(this, _V(0,0,1), TrackPointEulers);

			//		HUDTextOut_CLR(skp, 10, lineY[16], GOLD, "Designated point errors P:%+3.2f Y:%+3.2f [Alt-S]-Sun track {POINT TRACKING} [Alt-P]-no track point", TrackPYR.data[PITCH]*DEG, TrackPYR.data[YAW]*DEG);
			//	}
			//	else
			//	{
			//		HUDTextOut_CLR(skp, 10, lineY[16], GOLD, "Sunward attitude errors P:%+3.2f Y:%+3.2f [Alt-S]-Sun track [Alt-P]-track point", SunwardPYR.data[PITCH]*DEG, SunwardPYR.data[YAW]*DEG);
			//	};

			//	// implement automatic sunward tracking with deadbands!
			//};

			//int hrs = (int)floor(NTR_Life_s/3600.0);
			//int mins = (int)floor((NTR_Life_s-3600.0*hrs)/60.0);
			//int seconds = (int)(NTR_Life_s-3600.0*hrs-60.0*mins);
			//HUDTextOut_CLR(skp, 10, lineY[6], RED, "Remaining engine lifetime: %d h %d min %d s", hrs, mins, seconds);

			//if (commanded_dV>0.0)
			//{
			//	// TODO: time to planned cutoff!
			//	HUDTextOut_CLR(skp, 10, lineY[9], GOLD, "Commanded dV: %10.2f", commanded_dV);
			//};

			char solarStatus[5];
			solarStatus[4] = '\0';
			solarStatus[0] = '1';
			switch (AnimationStatus[0])
			{
			case simple_anim_status_t::CLOSED: solarStatus[1] = 'X'; break;
			case simple_anim_status_t::CLOSING: solarStatus[1] = 'x'; break;
			case simple_anim_status_t::OPEN: solarStatus[1] = 'O'; break;
			case simple_anim_status_t::OPENING: solarStatus[1] = 'o'; break;
			default: solarStatus[1] = '?'; break;
			};
			solarStatus[2] = '2';
			switch (AnimationStatus[1])
			{
			case simple_anim_status_t::CLOSED: solarStatus[3] = 'X'; break;
			case simple_anim_status_t::CLOSING: solarStatus[3] = 'x'; break;
			case simple_anim_status_t::OPEN: solarStatus[3] = 'O'; break;
			case simple_anim_status_t::OPENING: solarStatus[3] = 'o'; break;
			default: solarStatus[3] = '?'; break;
			};

			HUDTextOut_CLR(skp, 10, lineY[10], GOLD, "Solar array: %s Aux.battery: [TBI]", solarStatus);
			HUDTextOut_CLR(skp, 10, lineY[11], GOLD, "Shear at dock: [TBI] Vibration spectrum: [TBI]");
			HUDTextOut_CLR(skp, 10, lineY[12], GOLD, "FWD gamma: [TBI] beta: [TBI] neutron flux: [TBI]");
			return true;
		};
	case DAP_CONFIG:
		{
			HUDTextOut_CLR(skp, 10, lineY[0], GOLD, "DAP CONFIG [Ctrl-Y]-switch to ENGINE CONTROL [Alt-C]-command entry");
			//HUDTextOut_CLR(skp, 10, lineY[1], GOLD, "Sample command to change gain: TVC Pitch Prop 100.0");

			//HUDTextOut_CLR(skp, 10, lineY[3], GOLD, "Controller Prop  Int   Deriv MinOut MaxOut");
			//for (int i = 0; i<NUM_CONTROLLERS; i++)
			//{
			//	HUDTextOut_CLR(skp, 10, lineY[i+4], GOLD, "%-9s %-+4.1f %-+2.2f %-+2.2f %-+2.3f %-+2.3f", Controllers[i].name, Controllers[i].proportionalGain, Controllers[i].integralGain, Controllers[i].derivativeGain, Controllers[i].minOutput, Controllers[i].maxOutput);
			//};

			return true;
		};
		//comms and telemetry
		// docking
	default: return true;
	};

	return true;

}

// suntracker from MBR-ITS:
//void p3p4::SolpSun (void) {				// auto sun tracking
//
//	double adj = oapiGetTimeAcceleration(); // dead-band adjustment factor
//
//
//	double solp_angle;
//	double Sun_angle;
//	char cbuf[7]= "Sun";
//	OBJHANDLE hRef = oapiGetObjectByName(cbuf);
//	VECTOR3 rsunloc;
//	VECTOR3 sunglobal;
//	oapiGetGlobalPos (hRef, &sunglobal);	// get the global pos of Sun
//	Global2Local(sunglobal,rsunloc);		// convert to vessel frame coords
//
//	/////////////// z axis rotation ////////////
//
//	solp_angle = (PI*((P4rot_proc*2)+ 0.0));	// current SARJ rotation angle
//	Sun_angle = atan2(-rsunloc.y, -rsunloc.x);	// current angle of sun around SARJ rotation axis(- adjusted 180deg)
//
//	if (Sun_angle < 0) Sun_angle = Sun_angle + (2*PI);		// make sure angle is always between
//	if (Sun_angle > (2*PI)) Sun_angle = Sun_angle - (2*PI);	// 0 to 360 deg
//
//
//	{
//	if ((solp_angle - Sun_angle) > -0.001*adj && (solp_angle - Sun_angle) < 0.001*adj) // deadband
//	{
//		P4_rot_sw = 0;					// turn off manual rot
//		P4rot_proc = Sun_angle/(2*PI);	// lock on to sun
//		SetAnimation (P4rot_anim, P4rot_proc);	// apply to animation
//	}
//	else if ((solp_angle - Sun_angle) > 0) // logic for deciding which way to rotate to find sun
//	{
//		if (solp_angle - Sun_angle < PI) P4_rot_sw = 2;
//		if (solp_angle - Sun_angle > PI) P4_rot_sw = 1;
//	}
//
//	else if ((solp_angle - Sun_angle) < 0)
//	{
//		if (solp_angle - Sun_angle > -PI) P4_rot_sw = 1;
//		if (solp_angle - Sun_angle < -PI) P4_rot_sw = 2;
//	}
//	}
//
//		double Sun_angleb;
//		// rotate frame of reference by amount equal to required SARJ rotation, and get x coord
//		double rotsunlocx = ((cos(Sun_angle))*rsunloc.x)+((sin(Sun_angle))*rsunloc.y);
//		Sun_angleb = atan2(rotsunlocx, rsunloc.z);  // angle of sun around BGA axis
//		// ensure angle is between 0-360
//		if (Sun_angleb < 0) Sun_angleb = Sun_angleb + (2*PI);
//		if (Sun_angleb > (2*PI)) Sun_angleb = Sun_angleb - (2*PI);
//
//		/////////////// AB y axis rotation ////////////
//
//{
//		double Solp_angleb;
//
//		// current BGA rotation angle
//		Solp_angleb = 2*PI*(ABrot_proc);
//
//
//	{
//	if ((Solp_angleb - Sun_angleb) > -0.001*adj && (Solp_angleb - Sun_angleb) < 0.001*adj) //deadband
//	{
//		rot_sw = 0;								// turn off manual rotation
//		ABrot_proc = Sun_angleb/(2*PI);			// lock on to sun
//		SetAnimation (ABrot_anim, ABrot_proc);	// apply to animation
//	}
//
//	// logic for deciding which way to turn to find sun
//	else if ((Solp_angleb - Sun_angleb) > 0)
//	{
//		if (Solp_angleb - Sun_angleb < PI) rot_sw = 2;
//		if (Solp_angleb - Sun_angleb > PI) rot_sw = 1;
//	}
//
//	else if ((Solp_angleb - Sun_angleb) < 0)
//	{
//		if (Solp_angleb - Sun_angleb > -PI) rot_sw = 1;
//		if (Solp_angleb - Sun_angleb < -PI) rot_sw = 2;
//	}
//	}
//}
//
//		/////////////// CD y axis rotation ////////////
//
//{
//		double Solp_anglec;
//
//
//		// current BGA rotation angle
//		Solp_anglec = 2*PI*(CDrot_proc);
//
//
//	{
//	if ((Solp_anglec - Sun_angleb) > -0.0005*adj && (Solp_anglec - Sun_angleb) < 0.0005*adj) //deadband
//	{
//		rot_sw2 = 0;								// turn off manual rotation
//		CDrot_proc = Sun_angleb/(2*PI);			// lock on to sun
//		SetAnimation (CDrot_anim, CDrot_proc);	// apply to animation
//	}
//
//	// logic for deciding which way to turn to find sun
//	else if ((Solp_anglec - Sun_angleb) > 0)
//	{
//		if (Solp_anglec - Sun_angleb < PI) rot_sw2 = 2;
//		if (Solp_anglec - Sun_angleb > PI) rot_sw2 = 1;
//	}
//
//	else if ((Solp_anglec - Sun_angleb) < 0)
//	{
//		if (Solp_anglec - Sun_angleb > -PI) rot_sw2 = 1;
//		if (Solp_anglec - Sun_angleb < -PI) rot_sw2 = 2;
//	}
//	}
//}
//
//}
//


// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void NTR_inlinetank::clbkSetClassCaps (FILEHANDLE cfg)
{
	//THRUSTER_HANDLE th_main;
	THRUSTER_HANDLE th_rcs[NUM_RCS_THRUSTERS], th_group[MAX_NUM_THRUSTERS_PER_GROUP]; 

	// physical vessel parameters
	SetSize (PB_SIZE);
	SetEmptyMass (PB_EMPTYMASS);
	SetPMI (PB_PMI_FULL);
	SetCrossSections (PB_CS);
	SetRotDrag (PB_RD);
	SetTouchdownPoints (PB_TDP[0], PB_TDP[1], PB_TDP[2]);
	SetClipRadius(NTR_CLIPRADIUS);
	EnableTransponder(true);
	SetTransponderChannel(NTR_Transponder);

	// docking port definitions
	DOCKHANDLE hDock1 = CreateDock (PB_DOCK1_POS, PB_DOCK1_DIR, PB_DOCK1_ROT);
	DOCKHANDLE hDock2 = CreateDock (PB_DOCK2_POS, PB_DOCK2_DIR, PB_DOCK2_ROT);
	EnableIDS(hDock1, true);
	EnableIDS(hDock2, true);
	SetIDSChannel(hDock1, NTR_IDS_Channel1);
	SetIDSChannel(hDock2, NTR_IDS_Channel2);


	// airfoil definitions
	CreateAirfoil3 (LIFT_VERTICAL,   PB_COP, vlift, NULL, PB_VLIFT_C, PB_VLIFT_S, PB_VLIFT_A);
	CreateAirfoil3 (LIFT_HORIZONTAL, PB_COP, hlift, NULL, PB_HLIFT_C, PB_HLIFT_S, PB_HLIFT_A);

	// control surface animations
	//UINT anim_Laileron = CreateAnimation (0.5);
	//UINT anim_Raileron = CreateAnimation (0.5);
	//UINT anim_elevator = CreateAnimation (0.5);
	//AddAnimationComponent (anim_Laileron, 0, 1, &trans_Laileron);
	//AddAnimationComponent (anim_Raileron, 0, 1, &trans_Raileron);
	//AddAnimationComponent (anim_elevator, 0, 1, &trans_Lelevator);
	//AddAnimationComponent (anim_elevator, 0, 1, &trans_Relevator);

	// aerodynamic control surface defintions
	//CreateControlSurface (AIRCTRL_ELEVATOR, 1.5, 0.7, _V( 0,0,-2.5), AIRCTRL_AXIS_XPOS, anim_elevator);
	//CreateControlSurface (AIRCTRL_AILERON, 1.5, 0.25, _V( 1,0,-2.5), AIRCTRL_AXIS_XPOS, anim_Laileron);
	//CreateControlSurface (AIRCTRL_AILERON, 1.5, 0.25, _V(-1,0,-2.5), AIRCTRL_AXIS_XNEG, anim_Raileron);

	// propellant resources
	hpr_LH2 = CreatePropellantResource (PB_LH2MASS);
	hpr_RCS = CreatePropellantResource (PB_RCSFUELMASS);

	// separate RCS and hydrogen tanks.

	//// main _dummy_ engine
	//th_main = CreateThruster (_V(0,0,-4.35), _V(0,0,1), 0, hpr_LH2, 1.0);
	////th_main = CreateThruster (_V(0,0,-4.35), _V(0,0,1), PB_MAXMAINTH, hpr, PB_ISP);
	//CreateThrusterGroup (&th_main, 1, THGROUP_MAIN);

	//th_NTR[NTR_Alice] = CreateThruster (V_Alice, _V(0,0,1), NTR_MAXTHRUST, hpr_LH2, PB_NTR_ISP);
	//th_NTR[NTR_Betty] = CreateThruster (V_Betty, _V(0,0,1), NTR_MAXTHRUST, hpr_LH2, PB_NTR_ISP);
	//th_NTR[NTR_Clara] = CreateThruster (V_Clara, _V(0,0,1), NTR_MAXTHRUST, hpr_LH2, PB_NTR_ISP);
	//CreateThrusterGroup (&th_NTR[0], 3, THGROUP_USER);

	//PARTICLESTREAMSPEC contrail_main = {
	//0, 5.0, 16, 200, 0.15, 1.0, 5, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
	//PARTICLESTREAMSPEC::LVL_PSQRT, 0, 2,
	//PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	//};
	//PARTICLESTREAMSPEC exhaust_main = {
	//	0, 2.0, 20, 200, 0.05, 0.1, NTR_EXHAUST_LENGTH, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
	//	PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
	//	PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	//};
	//AddExhaust (th_NTR[NTR_Alice], NTR_EXHAUST_LENGTH, 1, V_Alice, _V(0,0,-1));
	//AddExhaust (th_NTR[NTR_Betty], NTR_EXHAUST_LENGTH, 1, V_Betty, _V(0,0,-1));
	//AddExhaust (th_NTR[NTR_Clara], NTR_EXHAUST_LENGTH, 1, V_Clara, _V(0,0,-1));
	////AddExhaustStream (th_NTR[NTR_Alice], _V(0,0.3,-10), &contrail_main);
	//AddExhaustStream (th_NTR[NTR_Alice], V_Alice, &exhaust_main);
	//AddExhaustStream (th_NTR[NTR_Betty], V_Betty, &exhaust_main);
	//AddExhaustStream (th_NTR[NTR_Clara], V_Clara, &exhaust_main);

	// hover engine
	//th_hover = CreateThruster (_V(0,-1.5,0), _V(0,1,0), PB_MAXHOVERTH, hpr, PB_ISP);
	//CreateThrusterGroup (&th_hover, 1, THGROUP_HOVER);
	//AddExhaust (th_hover, 8, 1, _V(0,-1.5,1), _V(0,-1,0));
	//AddExhaust (th_hover, 8, 1, _V(0,-1.5,-1), _V(0,-1,0));

	//PARTICLESTREAMSPEC contrail_hover = {
	//	0, 5.0, 8, 200, 0.15, 1.0, 5, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
	//	PARTICLESTREAMSPEC::LVL_PSQRT, 0, 2,
	//	PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	//};
	//PARTICLESTREAMSPEC exhaust_hover = {
	//	0, 2.0, 10, 200, 0.05, 0.05, 8, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
	//	PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
	//	PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	//};

	//AddExhaustStream (th_hover, _V(0,-3, 1), &contrail_hover);
	//AddExhaustStream (th_hover, _V(0,-3,-1), &contrail_hover);
	//AddExhaustStream (th_hover, _V(0,-2, 1), &exhaust_hover);
	//AddExhaustStream (th_hover, _V(0,-2,-1), &exhaust_hover);

	// RCS engines : FIXME
	for (int i = 0; i<NUM_RCS_THRUSTERS; i++)
	{
		th_rcs[ i] = CreateThruster (V_RCS_POS[i], V_RCS_DIR[i], PB_MAXRCSTH, hpr_RCS, PB_ISP); //+Y
		AddExhaust (th_rcs[i], RCS_EXHAUST_LENGTH, 0.2, V_RCS_POS[i], -V_RCS_DIR[i]);
	};

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[2];
	th_group[2] = th_rcs[5];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[3];
	th_group[2] = th_rcs[4];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[4];
	th_group[2] = th_rcs[3];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[5];
	th_group[2] = th_rcs[2];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[4];
	th_group[2] = th_rcs[2];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_UP);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[5];
	th_group[2] = th_rcs[3];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_DOWN);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_LEFT);

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_RIGHT);

	// fix: 4 +z needed, located on aft quads
	// 2 -z on forward quads

	th_group[0] = th_rcs[12];
	th_group[1] = th_rcs[13];
	th_group[2] = th_rcs[14];
	th_group[3] = th_rcs[15];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_FORWARD);

	th_group[0] = th_rcs[16];
	th_group[1] = th_rcs[17];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BACK);

	// camera parameters
	SetCameraOffset (PB_DOCK1_POS);

	// associate a mesh for the visual
	MainMeshID = AddMesh ("NTR_inlinetank");
	//SetMeshVisibilityMode(meshID, MESHVIS_ALWAYS);
	SetMeshVisibilityMode(MainMeshID, MESHVIS_EXTERNAL);

	

	// enable only in space!
	for (int i = 0; i<NUM_BEACONS; i++)
	{
		beacons[i].shape = BEACONSHAPE_COMPACT;
		beacons[i].size = 0.2;
		beacons[i].falloff = 0.0;
		beacons[i].active = true; // in space
		beacons[i].pos = &(beaconPositions[i]);
		switch (i)
		{
		case BEACON_PLUSX:
		case BEACON2_PLUSX:
			beacons[i].period = 0.0;
			beacons[i].duration = 1.0;
			beacons[i].tofs = 0.0;
			beacons[i].col = &beaconColourGreen;
			break;
		case BEACON_MINUSX:
		case BEACON2_MINUSX:
			beacons[i].period = 0.0;
			beacons[i].duration = 1.0;
			beacons[i].tofs = 0.0;
			beacons[i].col = &beaconColourRed;
			break;
		case BEACON_PLUSY:
		case BEACON2_PLUSY:
			beacons[i].col = &beaconColourBlue;
			beacons[i].period = 2.0; //period
			beacons[i].duration = 1.0; //duration
			beacons[i].tofs = 1.0; //time offset
			break;
		case BEACON_MINUSY:
		case BEACON2_MINUSY:
			beacons[i].col = &beaconColourBlue;
			beacons[i].period = 2.0; //period
			beacons[i].duration = 1.0; //duration
			beacons[i].tofs = 0.0; //time offset
			break;
		};
	};

	for (int i = 0; i < NUM_BEACONS; i++)
	{
		AddBeacon(&beacons[i]);
	};
}

// ==============================================================
// Airfoil lift/drag functions
// ==============================================================

void NTR_inlinetank::vlift (VESSEL *v, double aoa, double M, double Re,
					  void *context, double *cl, double *cm, double *cd)
{
	static const double clp[] = {  // lift coefficient from -pi to pi in 10deg steps
		-0.1,-0.5,-0.4,-0.1,0,0,0,0,0,0,0,0,0,0,-0.2,-0.6,-0.6,-0.4,0.2,0.5,0.9,0.8,0.2,0,0,0,0,0,0,0,0,0,0.1,0.4,0.5,0.3,-0.1,-0.5
	};
	static const double aoa_step = 10.0*RAD;
	double a, fidx, saoa = sin(aoa);
	a = modf((aoa+PI)/aoa_step, &fidx);
	int idx = (int)(fidx+0.5);
	*cl = clp[idx]*(1.0-a) + clp[idx+1]*a;     // linear interpolation
	*cm = 0.0; //-0.03*sin(aoa-0.1);
	*cd = 0.03 + 0.4*saoa*saoa;                // profile drag
	*cd += oapiGetInducedDrag (*cl, 1.0, 0.5); // induced drag
	*cd += oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);  // wave drag
}

void NTR_inlinetank::hlift (VESSEL *v, double aoa, double M, double Re,
					  void *context, double *cl, double *cm, double *cd)
{
	static const double clp[] = {  // lift coefficient from -pi to pi in 45deg steps
		0,0.4,0,-0.4,0,0.4,0,-0.4,0,0.4
	};
	static const double aoa_step = 45.0*RAD;
	double a, fidx;
	a = modf((aoa+PI)/aoa_step, &fidx);
	int idx = (int)(fidx+0.5);
	*cl = clp[idx]*(1.0-a) + clp[idx+1]*a;     // linear interpolation
	*cm = 0.0;
	*cd = 0.03;
	*cd += oapiGetInducedDrag (*cl, 1.5, 0.6); // induced drag
	*cd += oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);  // wave drag
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new NTR_inlinetank (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (NTR_inlinetank*)vessel;
}
