// ==============================================================
//                 ORBITER MODULE: NTR_test
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2004 Martin Schweiger
//                   All rights reserved
//
// NTR_test.cpp
// Control module for NTR_test vessel class
//
// Notes:
// This is an example for a "minimal" vessel implementation which
// only overloads the clbkSetClassCaps method to define vessel
// capabilities and otherwise uses the default VESSEL class
// behaviour.
// ==============================================================
// FIXED: does not launch from surface gracefully :)

// TODO: sun tracking and power output for other modules
// TODO: Isp change, integrate dV to shutdown
// DONE: PVA animation during checkout
// TODO: extending nozzle skirt during checkout
// TODO: visuals for gimballing animation
// DONE: ullage burn pre-startup
// TODO: tumbling trip?
// TODO: fairing
// TODO: stow panels by default (reported by: PeterRoss) almost done?
// TODO: reduce RCS fuel waste en route to the tracked attitude!
// TODO: att control endgame is not optimized.
// TODO: add more "debugging" output to Orbiter.log
// TODO: add LVLH pointing mode, stack of maneuvers. Point antenna by rolling
// TODO: add RCS propulsive maneuvers - multi-thruster and with +Z reorientation

// can we use gimballing to achieve required inertial velocity at MECO? without an integrator - no.

#define STRICT
#define ORBITER_MODULE

#include "NTR_test.h"

using namespace std;

// TODO: varying Isp
#define VMSG_NTROFFSET                  (VMSG_USER+7500)
#define VMSG_GET_PROPELLANT_TYPE        (VMSG_NTROFFSET+1)
#define VMSG_GET_FUEL_TYPE              (VMSG_NTROFFSET+2)
#define VMSG_GET_OXIDIZER_TYPE          (VMSG_NTROFFSET+3)
#define VMSG_GET_3RD_COMPONENT_TYPE     (VMSG_NTROFFSET+4)
#define VMSG_GET_PRESSURANT_TYPE        (VMSG_NTROFFSET+5)
#define VMSG_GET_POWER_OUTPUT_W         (VMSG_NTROFFSET+6)
#define VMSG_GET_TANK_DRAWDOWN_SEQUENCE (VMSG_NTROFFSET+7)
#define VMSG_SEND_COMMAND               (VMSG_NTROFFSET+8)


// avoid 1 and 0 defaults
enum PROPELLANT_TYPE
{
	UNKNOWN_PROPELLANT = -1,
	MONOPROPELLANT = 2,
	BIPROPELLANT,
	TRIPROPELLANT,
	BIPROPELLANT_COMBINED,
	TRIPROPELLANT_COMBINED
};

enum FUEL_TYPE
{
	UNKNOWN_FUEL = -1,
	HYDROGEN_CRYOGENIC = 2,
	HYDROGEN_SLUSH,
	METHANE_CRYOGENIC,
	PROPANE_CRYOGENIC,
	XENON_GASEOUS,
	HYDRAZINE,
	MONOMETHYLHYDRAZINE,
	UNSYMMETRIC_DIMETHYLHYDRAZINE,
	PHENYLHYDRAZINE,
	AEROZINE50,
	METHYLNITRATE,
	ETHYLNITRATE,
	ISOPROPYLNITRATE,
	ETHYLENEGLYCOLDINITRATE,
	TRINITROGLYCERINE,
	NITROMETHANE,
	NITROETHANE,
	NITROPROPANE,
	ETHYLENEOXIDE,
	HYDROGEN_PEROXIDE_MONOPROP,
	RP1,
	JP1,
	JP4,
	AMMONIA_CRYOGENIC,
	ANILINE,
	MONOETHYLAMINE,
	DIETHYLAMINE,
	TRIETHYLAMINE,
	DIETHYLENETRIAMINE,
	KEROSENE,
	HYDINE,
	ETHANOL,
	TONKA250
};

enum OXIDIZER_TYPE
{
	UNKNOWN_OXIDIZER = -1,
	OXYGEN_CRYOGENIC = 2,
	OZONE_CRYOGENIC,
	HYDROGEN_PEROXIDE,
	TETRANITROMETHANE,
	NITROGEN_TETROXIDE,
	NITRIC_ACID
};

enum PRESSURANT_TYPE
{
	UNKNOWN_PRESSURANT = -1,
	HELIUM = 2,
	NITROGEN_GASEOUS
};

enum TANK_TYPE
{
	UNKNOWN_TANK_TYPE = -1,
	INTEGRAL_TANK = 2,
	INLINE_TANK,
	DROP_TANK
};

enum NTR_FAILURE_CODE {
	UNKNOWN_FAILURE = -1,
	NTR_OK = 0,
	ONE_REACTOR_OUT,
	TWO_REACTORS_OUT,
	THREE_REACTORS_OUT,
	ONE_TPA_OUT,
	TWO_TPA_OUT,
	MULTIPLE_TPA_OUT,
	ONE_REACTOR_MELTDOWN,
	MULTIPLE_REACTORS_MELTDOWN,
	TANK_LEAK,
	MAIN_BUS_A_UNDERVOLT,
	MAIN_BUS_B_UNDERVOLT,
	HARDDOCK_FAILURE,
	ONE_GIMBAL_OUT,
	MULTIPLE_GIMBALS_OUT,
	IMU_OVERHEAT,
	COMMS_LOSS
};

// dry power-up should be included (e.g. to 700 kW) Isp rampup should be modelled directly, from the ground up (integrate the eqn!)

// NERVA BOL Gamma flux in radial direction, gammas/cm^2/sec
// grp, lower(MeV), upper(MeV), flux, mass attenuation coeff(cm^2/g) [Al] [W]
//1	 7.5-30		1.2e+12		2.39e-02	4.57e-02
//2	 7-7.5		2.1e+12		2.51		4.37
//3  6-7		6.6e+12		2.60		4.27
//4	 5-6		1.9e+13		2.74		4.14
//5  4-5		4.8e+13		2.97		4.05
//6	 3-4		1.7e+14		3.32		4.03
//7	 2.6-3		1.3e+14		3.69		4.11
//8	 2.2-2.6	2.0e+14		4.00		4.24
//9	 1.8-2.2	3.3e+14		4.32		4.37
//10 1.35-1.8	5.5e+14		4.89		4.82
//11 0.9-1.35	1.0e+15		5.84		6.01
//12 0.4-0.9	2.4e+15		7.53		9.71
//13 0-0.4		3.6e+15		1.20e-01	7.38e-01

//BOL neutron flux in radial direction neutrons/cm^2/sec
//1 E<0.4eV			8.4E+13
//2 0.4eV<E<1MeV	3.7E+14
//3 E>1MeV			1.1E+14

// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	int i;
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,   -0.1,     0,   0.2,  0.25,   0.2,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		(CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
		(CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
		(CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	for (i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);
	return CL[i] + (aoa-AOA[i])*SCL[i];
}

// TODO: safe separation of the reactors from the tank in case of failed launch, broken arrow!
// Interesting. Can we set up the vehicle so that the NTRs point upwards at launch and are protected 
// from too strong impact at sea by the fairing shroud at failed launch? Prolly not...


// needs intermediate testing? e.g. steady state vibrations at something less than nominal power level


// manual TVC first, only then automatic.
// treat states separately for each engine, what to do about stage state?
// boiloff!


double NTR_test::AnimationSpeeds[NUM_ANIMATIONS] = {
	0.1, // SOLAR1
	0.1,  // SOLAR2
	0.1 // SBAND
};

UINT NTR_test::Animations[NUM_ANIMATIONS] = { 
	NULL, //SOLAR1
	NULL,  //SOLAR2
	NULL //SBAND
};

//MGROUP_ROTATE NTR_test::trans_Laileron (0, &GRP_LWING, 1, LWING_REF, LWING_AXIS, AILERON_RANGE);
//MGROUP_ROTATE NTR_test::trans_Raileron (0, &GRP_RWING, 1, RWING_REF, RWING_AXIS, AILERON_RANGE);
//MGROUP_ROTATE NTR_test::trans_Lelevator (0, &GRP_LWING, 1, LWING_REF, LWING_AXIS, -ELEVATOR_RANGE);
//MGROUP_ROTATE NTR_test::trans_Relevator (0, &GRP_RWING, 1, RWING_REF, RWING_AXIS, ELEVATOR_RANGE);

NTR_test::NTR_test (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
	MissionTime_s = 0.0;
	TotalHeatLoad = 0.0;
	Boiloff_per_day = 0.0;
	estimatedBurnTime_s = 0.0;
	timeEngineBurnStarted = numeric_limits<double>::infinity();
	T_ExternalSurface = 273.15;//FWIW
	T_LH2_Tank = LH2_triple_point;
	CryocoolerPowerInput = CryocoolerMinimumPower;
	BatteryStatus = BatteryStatus_t::CHARGED;
	BatteryCapacity = BATTERY_INITIAL_CAPACITY;
	BatteryCharge = BATTERY_INITIAL_CAPACITY;
	BatteryActivationThreshold = BATTERY_ACTIVATION_THRESHOLD;
	NumBatteryCycles = 0;
	solarSystem = new SolSystem();
	StageState = reactor_state_t::Offline;
	currentThrust = 0.0;
	fullCOG_elev = this->GetCOG_elev();
	emptyCOG_elev = 2.0; //placeholder
	dV_sinceBurnStart = 0.0;
	XePoisoning = 0.0;
	commanded_dV = 0.0;
	for (int i = 0; i<NUM_REACTORS; i++)
	{
		NTRGimbal[i][0] = 0.0;
		NTRGimbal[i][1] = 0.0;
		GimbalSelect[i] = false;
	};
	//NullTorque = false;
	MaintainCommandedRates = false;
	Controllers[TVC_P] = PIDControl("TVC Pitch", TVC_DAP_Gain, 0.0, 0.0, -NTR_PITCH_GIMBAL_LIMIT, NTR_PITCH_GIMBAL_LIMIT);
	Controllers[TVC_Y] = PIDControl("TVC Yaw", TVC_DAP_Gain, 0.0, 0.0, -NTR_YAW_GIMBAL_LIMIT, NTR_YAW_GIMBAL_LIMIT);
	// RCS- based roll control!
	Controllers[RCS_R] = PIDControl("RCS Roll", RCS_DAP_Gain, 0.0, 0.0, -1.0, 1.0); //?
	Controllers[RCS_P] = PIDControl("RCS Pitch", RCS_DAP_Gain, 0.0, 0.0, -1.0, 1.0); //?
	Controllers[RCS_Y] = PIDControl("RCS Yaw", RCS_DAP_Gain, 0.0, 0.0, -1.0, 1.0); //?

	NTR_Life_s = NTR_MAX_LIFE;
	FuelLevelTripActivated = false;
	IsInLaunchMode = false;
	IsInCheckoutMode = false;
	Static_dP_dt_criterion_counter = 0;
	G_limit_criterion_counter = 0;
	old_STP = HUGE_VAL;
	DefineAnimations();
	for (int i = 0; i<NUM_ANIMATIONS; i++)
	{
		AnimationDeployments[i] = 1.0;
		AnimationStatus[i] = simple_anim_status_t::OPEN;
	};
	CommandedRates = NULL_VECTOR;
	//JointGimbalAttitude = NULL_VECTOR;
	UllageBurnTimeLeft = 0.0;
	CurrentScreen = glass_screen_t::ENGINE_CONTROL;
	ShallTrackSun = false;
	ShallTrackPoint = false;
	TrackPointEulers = _V(0,0,0);
	for (int i = 0; i<3; i++)
	{
		ActiveDAPChannels[i] = true;
	};
	// test! works OK!
	//this->OptimizeTorque();
	optProblem = new nlopt::opt(nlopt::algorithm::LN_COBYLA,NUM_REACTORS*2);
	AutoRCSEnabled = false; // by default!
	ShallRecomputeRCSStrings = true;
	AutoRCSTorqueOrThrust = 1000.0; //??
	Att_Deadbands = _V(DB_FINE, DB_FINE, DB_FINE); //???
	Rate_Deadbands = _V(RATE_DB_DEFAULT, RATE_DB_DEFAULT, RATE_DB_DEFAULT); //???
	DAP_SteadyRate = RATE_SLOW;
	DAP_SlowRate = RATE_SLOWEST;
	visual = NULL;
	SystemHealthMap[NTR_Component_t::POWER_CONVERSION_AND_REGULATION_UNIT] = true;
	// multistage workaround
}

NTR_test::~NTR_test ()
{
	delete optProblem;
	delete solarSystem;
}

void NTR_test::clbkVisualCreated(VISHANDLE vis, int refcount)
{
	VESSEL2::clbkVisualCreated(vis, refcount);
	visual = vis;
}

void NTR_test::clbkVisualDestroyed(VISHANDLE vis, int refcount)
{
	visual = NULL;
	VESSEL2::clbkVisualDestroyed(vis, refcount);
}

void NTR_test::DefineAnimations (void)
{
	// 1. Solar array 1
	static UINT Solar1Grp[3] = {MESHGRP_Array1_default,MESHGRP_Array1_SolarTexture, MESHGRP_SARJ1_SARJ_black};
	static MGROUP_ROTATE MGR_Solar1 (MESH_MAIN,
		Solar1Grp, 3, 
		PVA1_Axis_Offset, 
		DIR_XY_225_Z_0, 
		(float)(PI*0.5));

	Animations[SOLAR1_ANIMATION] = CreateAnimation (1.0);
	AddAnimationComponent (Animations[SOLAR1_ANIMATION], 0, 1.0, &MGR_Solar1);

	// Solar array 2
	static UINT Solar2Grp[3] = {MESHGRP_Array2_default,MESHGRP_Array2_SolarTexture, MESHGRP_SARJ2_SARJ_black};
	static MGROUP_ROTATE MGR_Solar2 (MESH_MAIN,
		Solar2Grp, 3, 
		PVA2_Axis_Offset, 
		DIR_XY_045_Z_0, 
		(float)(PI*0.5));

	Animations[SOLAR2_ANIMATION] = CreateAnimation (1.0);
	AddAnimationComponent (Animations[SOLAR2_ANIMATION], 0.0, 1.0, &MGR_Solar2);

	static UINT SBandGrp[1] = {MESHGRP_SBand1_default};
	static MGROUP_ROTATE MGR_SBand (MESH_MAIN,
		SBandGrp, 1, 
		SBand_Axis_Offset, 
		DIR_XY_135_Z_0, 
		(float)(-PI*0.5));

	Animations[SBAND_ANIMATION] = CreateAnimation (1.0);
	AddAnimationComponent (Animations[SBAND_ANIMATION], 0, 1.0, &MGR_SBand);


}

void NTR_test::SetRCSstringLevel_SingleStep(const stack_RCS_string_t& s, const double dlevel) const
{
	for (unsigned int j = 0; j<s.size(); j++)
	{
		OBJHANDLE hV = oapiGetObjectByName(const_cast<char*>(s.at(j).vesselName.c_str()));
		if (hV==NULL) continue;
		VESSEL* v = oapiGetVesselInterface(hV);
		if (v==NULL) continue;
		THRUSTER_HANDLE th = v->GetThrusterHandleByIndex(s.at(j).thrusterID);
		if (th==NULL) continue;
		v->SetThrusterLevel_SingleStep(th, dlevel*s.at(j).thrustRatio);
	};
}

void NTR_test::SetThrusterGroupLevel_SingleStep(THGROUP_TYPE  thgt, double dlevel) const
{
	// RCS string!
	if (DockingStatus(0)==1 && (stack_RCS.size()>0) && AutoRCSEnabled)
	{
		bool foundString = false;
		stack_RCS_string_t s;
		switch (thgt)
		{
		case THGROUP_TYPE::THGROUP_ATT_BANKLEFT:
			{
				stack_RCS_config_t::const_iterator citer = stack_RCS.find(ROT_BANKLEFT_RCS);
				if (citer!=stack_RCS.end())
				{
					SetRCSstringLevel_SingleStep(citer->second, dlevel);
					return;
				};
				break;
			};
		case THGROUP_TYPE::THGROUP_ATT_BANKRIGHT:
			{
				stack_RCS_config_t::const_iterator citer = stack_RCS.find(ROT_BANKRIGHT_RCS);
				if (citer!=stack_RCS.end())
				{
					SetRCSstringLevel_SingleStep(citer->second, dlevel);
					return;
				};
				break;
			};
		case THGROUP_TYPE::THGROUP_ATT_PITCHDOWN:
			{
				stack_RCS_config_t::const_iterator citer = stack_RCS.find(ROT_PITCHDOWN_RCS);
				if (citer!=stack_RCS.end())
				{
					SetRCSstringLevel_SingleStep(citer->second, dlevel);
					return;
				};
				break;
			};
		case THGROUP_TYPE::THGROUP_ATT_PITCHUP:
			{
				stack_RCS_config_t::const_iterator citer = stack_RCS.find(ROT_PITCHUP_RCS);
				if (citer!=stack_RCS.end())
				{
					SetRCSstringLevel_SingleStep(citer->second, dlevel);
					return;
				};
				break;
			};
		case THGROUP_TYPE::THGROUP_ATT_YAWLEFT:
			{
				stack_RCS_config_t::const_iterator citer = stack_RCS.find(ROT_YAWLEFT_RCS);
				if (citer!=stack_RCS.end())
				{
					SetRCSstringLevel_SingleStep(citer->second, dlevel);
					return;
				};
				break;
			};
		case THGROUP_TYPE::THGROUP_ATT_YAWRIGHT:
			{
				stack_RCS_config_t::const_iterator citer = stack_RCS.find(ROT_YAWRIGHT_RCS);
				if (citer!=stack_RCS.end())
				{
					SetRCSstringLevel_SingleStep(citer->second, dlevel);
					return;
				};
				break;
			};
		};

	};
	for (unsigned int i = 0; i<this->GetGroupThrusterCount(thgt); i++)
	{
		THRUSTER_HANDLE th = this->GetGroupThruster(thgt, i);
		if (th!=NULL)
		{
			this->SetThrusterLevel_SingleStep(th, dlevel);
		};
	};
}

void NTR_test::ComputeOptimumRCS(double refThrust)
{
	// got axes mixed up
	// TODO: edit fuel costs and exclusion params

	TransformStackThrusterInfoIntoRootFrame(stackInfo, 
		stackThrusters, 
		stackTanks, 
		stackCOG, 
		stackPMI);

	for (unsigned int k = TRAN_FORWARD_RCS; k<=ROT_BANKRIGHT_RCS; k++)
	{

		stack_RCS_string_t RCS_string;
		for (unsigned int i = 0; i<stackThrusters.size(); i++)
		{
			if (stackThrusters[i].groupID.find(TRAN_FORWARD_MAIN) != stackThrusters[i].groupID.end() ||
				stackThrusters[i].groupID.find(TRAN_BACK_RETRO) != stackThrusters[i].groupID.end() ||
				stackThrusters[i].groupID.find(TRAN_UP_HOVER) != stackThrusters[i].groupID.end() ||
				stackThrusters[i].groupID.empty()) continue; // user thrusters get skipped as well.
			if (_strnicmp(stackThrusters[i].pV->GetClassNameA(),"NTR_inlinetank", 14)==0) continue; //skip over friendly tanks
			stack_RCS_item_t item;
			//RCS_string[i].
			item.index = i;
			item.thrustRatio = 0.1;
			item.thrusterID = stackThrusters[i].thrusterID;
			item.vesselName.assign(stackThrusters[i].pV->GetName());
			RCS_string.push_back(item);
		};
		// ignore side effects for rotational control!
		if (k>=ROT_RCS_FIRST)
		{
			ComputeOptimumRCSstring_LP(k,
				&stackThrusters, 
				&RCS_string, 
				stackCOG, refThrust, true); // should ignore side force!
		}
		else
		{
			ComputeOptimumRCSstring_LP(k,
				&stackThrusters, 
				&RCS_string, 
				stackCOG, refThrust, false); // for a standalone vessel should be the same as current arrangement
		};
		if (RCS_string.size()>0) stack_RCS[k] = RCS_string;
	};
};

void NTR_test::FillDrawdownSequence_and_OptimizeRCS()
{
	FillDrawdownSequence();
	if (stackInfo.size()>1) // add an inhibit check
	{
		// need recalculation at each step?
		if (ShallRecomputeRCSStrings) ComputeOptimumRCS(AutoRCSTorqueOrThrust); // TBI
	}
	else
	{
		stack_RCS.clear();
	};
}

void NTR_test::clbkPostStep (double simt, double simdt, double mjd)
{
	if (IsInLaunchMode) return; // for now
	if (IsInCheckoutMode) return; // for now

	MissionTime_s += simdt;

	OBJHANDLE hSun = GetCentralStar(solarSystem); //oapiGetGbodyByName("Sun");

	// TODO: relocate to PostStep!
	// from MBR-ITS
	VECTOR3 rsunloc;
	VECTOR3 sunglobal;
	oapiGetGlobalPos (hSun, &sunglobal);	// get the global pos of Sun
	Global2Local(sunglobal,rsunloc);		// convert to vessel frame coords
	double sunDist = length(rsunloc);
	double sunDist_AU = sunDist/AU;
	normalise(rsunloc);
	VECTOR3 normalab = _V(0, 0, 1);	// no rotation
	VECTOR3 normal_ab_ecl;
	GlobalRot(normalab, normal_ab_ecl);	// change to global reference for future MFD?
	double dot_p_ab_sun = dotp(normalab,rsunloc); // cos of angle beween sun->normal
	double SunAngle = acos(fabs(dot_p_ab_sun));
	if (dot_p_ab_sun < 0.0) dot_p_ab_sun = -SOLAR_BACK_FACE_PENALTY*dot_p_ab_sun; // more than 90deg from normal
	// back face generates 30% power

	// what if in flight? A POSSIBLE BUG LOCUS

	double solarIllumFactor = 1.0;
	OBJHANDLE hPlanetRef = this->GetGravityRef();
	if (hPlanetRef != hSun)
	{
		solarIllumFactor= IlluminationFactor(this->GetHandle(), this->GetGravityRef(), solarSystem);
	};
	double solarFlux = SOLAR_CONSTANT/(sunDist_AU*sunDist_AU);
	double insolation = solarIllumFactor*solarFlux;

	TotalHeatLoad = CylinderProjectionArea(SunAngle, NTR_cyl_length, NTR_cyl_diameter)*insolation*EXTERNAL_SURFACE_SOLAR_ABSORPTANCE;

	VECTOR3 planet_rel_pos;
	double ei_power;
	if (hPlanetRef != hSun)
	{
		//OBJHANDLE hPlanetRef = this->GetSurfaceRef();
		GetRelativePos(hPlanetRef, planet_rel_pos);
		double planetDist = length(planet_rel_pos);
		//double terminator_angle = asin(oapiGetSize(eRef)/earth_dist);

		double cos_planet_sc_sun_angle = (dotp(rsunloc, planet_rel_pos))/(sunDist*planetDist);
		//double earth_sc_sun_angle = acos(cos_earth_sc_sun_angle);

		//double sep_angle = (earth_sc_sun_angle+0.0045) - terminator_angle;
		//double power_adjust = sep_angle/0.009;

		//if (power_adjust > 1.0) power_adjust = 1;
		//else if (power_adjust < 0.0) power_adjust = 0;

		/////////// earth insolation ///////////////
		double PlanetAngle = 0.0;
		double cos_normal_planet = (dotp( -planet_rel_pos, normal_ab_ecl)/planetDist);

		PlanetAngle = acos(fabs(cos_normal_planet));

		if (cos_normal_planet < 0) cos_normal_planet = -SOLAR_BACK_FACE_PENALTY*cos_normal_planet;

		cos_planet_sc_sun_angle=-cos_planet_sc_sun_angle;
		if (cos_planet_sc_sun_angle < 0) cos_planet_sc_sun_angle = 0;

		char planetName[256];
		oapiGetObjectName(hPlanetRef, planetName, 255);
		double albedo = RoughPlanetaryAlbedo(planetName);//!!! get albedo model from planet name

		// the problem is, the planet is not a point source!
		// distance to atm or surface boundary? + 100 km
		double planetAlbedoRefSize = oapiGetSize(hPlanetRef)+100000.0;
		double distC = (planetDist/planetAlbedoRefSize);
		double planet_distance_coeff = min(1.0, 1.0/(distC*distC));

		ei_power = cos_normal_planet*cos_planet_sc_sun_angle*albedo*SOLAR_ARRAY_AREA*planet_distance_coeff*insolation*SOLAR_EFFICIENCY;
		double Planet_proj_area = CylinderProjectionArea(PlanetAngle, NTR_cyl_length, NTR_cyl_diameter);
		TotalHeatLoad += planet_distance_coeff*insolation*Planet_proj_area*albedo*EXTERNAL_SURFACE_SOLAR_ABSORPTANCE;
		TotalHeatLoad += solarFlux*Planet_proj_area*planet_distance_coeff*(1.0-albedo)*EXTERNAL_SURFACE_IR_EMISSIVITY;
	};

	//internal heat load!


	SolarArrayPowerOutput = SOLAR_ARRAY_AREA*insolation*SOLAR_EFFICIENCY*dot_p_ab_sun+ei_power; //WHERE D'YOU GET THE CONSTANT??
	if (AnimationStatus[0]!=simple_anim_status_t::OPEN || AnimationStatus[1]!=simple_anim_status_t::OPEN ) SolarArrayPowerOutput = 0.0;

	NetPowerOutput = SolarArrayPowerOutput*DISTRIBUTION_EFFICIENCY;
	// cycle through electric loads
	// cryocoolers/heaters are the most demanding - fixed + variable input

	// DONE: calculate eclipse length for the current orbit

	if (NetPowerOutput<BatteryActivationThreshold)
	{
		BatteryStatus = BatteryStatus_t::DISCHARGING;
		// ???
		double BatterySuppliedNetOutput = BatteryActivationThreshold-NetPowerOutput;
		double loadOnBattery = BatterySuppliedNetOutput/DISTRIBUTION_EFFICIENCY;
		if (BatteryCharge>0.0)
		{
			BatteryCharge -= loadOnBattery*(simdt/3600.0);
			BatteryCharge = max(BatteryCharge, 0.0);
			NetPowerOutput += BatterySuppliedNetOutput;
		};
	}
	else
	{
		if (BatteryCharge<BatteryCapacity)
		{
			if (BatteryStatus != BatteryStatus_t::CHARGING) NumBatteryCycles +=1;
			BatteryStatus = BatteryStatus_t::CHARGING;
			BatteryCharge += RECHARGE_RATE*(simdt/3600.0);
			BatteryCharge = min(BatteryCapacity, BatteryCharge);
			NetPowerOutput -= RECHARGE_RATE/CHARGER_EFFICIENCY;
			TotalHeatLoad += (1.0-CHARGER_EFFICIENCY)*RECHARGE_RATE; // this is thermal!
		}
		else
		{
			BatteryStatus = BatteryStatus_t::CHARGED;
		};
	};

	TotalHeatLoad += (1.0-DISTRIBUTION_EFFICIENCY)*NetPowerOutput; // wiring losses

	if (NetPowerOutput<CDH_Power) 
	{
		NetPowerOutput = 0.0; // shouldn't be here
		SystemHealthMap[NTR_Component_t::CDH_UNIT] = false;
	}
	else
	{
		NetPowerOutput -= CDH_Power; // if insufficient power for CDH, we're assuming that instead of death there are primary batteries in the unit...
		SystemHealthMap[NTR_Component_t::CDH_UNIT] = true;
	};

	// start taking away power for various systems - is rigid, will have to be re-coded?


	if (NetPowerOutput<CryocoolerMinimumPower)
	{
		CryocoolerPowerInput = 0.0; // we're stuck! Add some cooldown...
		SystemHealthMap[NTR_Component_t::CRYOCOOLER] = false;
	}
	else
	{
		SystemHealthMap[NTR_Component_t::CRYOCOOLER] = true;
		CryocoolerPowerInput=min(CryocoolerMaximumPower, NetPowerOutput);
		NetPowerOutput -= CryocoolerPowerInput;
		if (NetPowerOutput<0.0) NetPowerOutput = 0.0; // shouldn't be here
	};

	TotalHeatLoad += CryocoolerPowerInput;

	T_ExternalSurface = pow(TotalHeatLoad/(EXTERNAL_SURFACE_IR_EMISSIVITY*NTR_radiator_area*StefanBoltzmann),0.25);

	double Cryocooler_Heat_Rejection = CryocoolerPowerInput/CryocoolerSpecificPower;

	double boiloff_per_second = MLI_LH2_boiloff_per_second(LH2_tank_area, Num_MLI_Layers, T_ExternalSurface, T_LH2_Tank,
		MLI_layer_thickness, 
		MLI_eq_eff_conductivity, 
		Cryocooler_Heat_Rejection);

	Boiloff_per_day = boiloff_per_second*86400.0;

	SetPropellantMass(hpr_LH2, GetPropellantMass(hpr_LH2)-boiloff_per_second*simdt);

	double LH2_Mass = GetPropellantMass(hpr_LH2);
	double LH2_MaxMass = GetPropellantMaxMass(hpr_LH2);
	// TODO: add boiloff depletion!
	if (ARE_PUMPS_WORKING(StageState)) // only if pumps are working
	{
		// avoids DAP artefacts?
		if (oapiGetTimeAcceleration() > 10) oapiSetTimeAcceleration(10);

		// lifetime reduction proportional to thrust relative to max thrust.
		double lifetime_reduction_ratio = currentThrust/NTR_MAXTHRUST;
		NTR_Life_s -= simdt*lifetime_reduction_ratio;
		NTR_Life_s = max(NTR_Life_s, 0.0);
		if (NTR_Life_s<=0.0)
		{
			StageState = reactor_state_t::Meltdown;
		};

		// DOES NOT WORK!
		//SURFHANDLE texA;
		//double lscale, wscale;
		//VECTOR3 exApos, exAdir;
		//if (GetExhaustSpec  ( Exhaust_A,   &lscale,  &wscale, &exApos, &exAdir, &texA)==true)
		//{
		//	double exT = 2900.0*currentThrust/NTR_MAXTHRUST;
		//	double R,G,B;
		//	BlackBody2RGB(exT, R,G, B);
		//	oapi::Sketchpad* skp2= new oapi::Sketchpad(texA);
		//	DWORD ck = oapiGetColour((DWORD)(255.0*R),(DWORD)(255.0*G),(DWORD)(255.0*B));
		//	oapiColourFill(texA, ck);
		//	delete skp2;
		//	//oapiSetSurfaceColourKey  (texA, ck ); 
		//};

		// shutdown if at commanded dV:
		double currentMass = this->GetStackMass();
		dV_sinceBurnStart = PB_NTR_ISP*log(massAtBurnStart/currentMass );
		double timeToShutdown = -log(ZeroThrustLevel/currentThrust)*ShutdownPeriod_s;
		double fuelMassToShutdown = (3.0*ShutdownPeriod_s*currentThrust/PB_NTR_ISP)*(1.0-exp(-timeToShutdown/ShutdownPeriod_s));

		double dV_atShutdown = PB_NTR_ISP*log(massAtBurnStart/(currentMass-fuelMassToShutdown));
		if (StageState != reactor_state_t::Meltdown && commanded_dV>0.0 && dV_atShutdown>commanded_dV)
		{
			StageState = reactor_state_t::Shutdown;
		};

		if (StageState != reactor_state_t::Meltdown && LH2_Mass<(fuelMassToShutdown+1.0))
		{
			// fuel level trip!
			FuelLevelTripActivated = true; // will not be reset.
			StageState = reactor_state_t::Shutdown;
		};

		if (LH2_Mass<LH2_MaxMass)
		{
			double refuel = min(NTR_MAX_INTERTANK_FLOW, LH2_MaxMass - LH2_Mass);
			// there was a suggestion to make drawdown consecutive, parallel, and a mix of both.
			if (drawdownSequence.size()>1)
			{
				if (drawdownSequence[0].pVessel != this)
				{
					PROPELLANT_HANDLE hpr_TX = drawdownSequence[0].pVessel->GetPropellantHandleByIndex(drawdownSequence[0].tankID);
					double FMass = drawdownSequence[0].pVessel->GetPropellantMass(hpr_TX);
					if (FMass == 0.0)
					{
						// exclude from the sequence, miss a step.
						drawdownSequence.erase(drawdownSequence.begin());
					}
					else
					{
						if (refuel>=FMass)
						{
							drawdownSequence[0].pVessel->SetPropellantMass(hpr_TX, 0.0);
							this->SetPropellantMass(hpr_LH2, LH2_Mass+FMass);
							drawdownSequence.erase(drawdownSequence.begin());
						}
						else
						{
							drawdownSequence[0].pVessel->SetPropellantMass(hpr_TX, FMass-refuel);
							this->SetPropellantMass(hpr_LH2, LH2_Mass+refuel);
						};
					};
				};
			};
		};
	};
}

double NTR_test::GetStackMass()
{
	double M = 0.0;
	if (stackInfo.size()==0) FillStackInfo(this, stackInfo);
	for (unsigned int i = 0; i<stackInfo.size(); i++)
	{
		M += stackInfo[i].pV->GetMass();
	};
	return M;
}

bool NTR_test::GetStackGroundContact()
{
	if (stackInfo.size()==0) FillStackInfo(this, stackInfo);
	for (unsigned int i = 0; i<stackInfo.size(); i++)
	{
		if (stackInfo[i].pV->GroundContact()) return true;
	};
	return false;
}

void NTR_test::FillDrawdownSequence()
{
	FillStackInfo(const_cast<VESSEL*>((VESSEL*)(this)), stackInfo);
	// drawdown sequence?
	drawdownSequence.clear();
	for (unsigned int i = 0; i<stackInfo.size(); i++)
	{
		if (stackInfo[i].pV == this) continue;
		if (stackInfo[i].fuel_tanks.size() == 0) continue;
		if (stackInfo[i].fuel_tanks[0].mass == 0) continue;
		drawdown_t dS;
		dS.pVessel = stackInfo[i].pV;
		dS.tankID = 0;
		drawdownSequence.push_back(dS);
	};
	drawdown_t dOwn;
	dOwn.pVessel = this;
	dOwn.tankID = 0; //LH2
	drawdownSequence.push_back(dOwn);
}

double H2_VibrationDOF(const double T)
{
	if (T<=0.0) throw "T<=0!";
	double cvt_over_t = H2_CharacteristicVibrationT/T;
	return 2.0*cvt_over_t /(exp(cvt_over_t)-1.0);
}

double H2_Gamma(const double T) 
{
	if (T<=0.0) throw "T<=0!";
	double vibDOF = H2_VibrationDOF(T);
	double gamma=(5.0+H2_RotationDOF+vibDOF)/(3.0+H2_RotationDOF+vibDOF);
	return gamma;
}

double H2_Isp(const double To, const double Po, const double Pe)
{
	double gamma = H2_Gamma(To);
	return (1/G)*sqrt(pow(Pe/Po,(gamma-1.0)/gamma)*2.0*gamma*H2_R*To/(gamma-1.0)); // works
}


int NTR_test::clbkGeneric(int msgid, int prm, void *context)
{
	switch(msgid)
	{
	case VMSG_GET_PROPELLANT_TYPE:
		switch(prm)
		{
		case 0: return PROPELLANT_TYPE::MONOPROPELLANT; // LH2
		case 1: return PROPELLANT_TYPE::BIPROPELLANT_COMBINED; // RCS
		default: return PROPELLANT_TYPE::UNKNOWN_PROPELLANT;
		};
	case VMSG_GET_FUEL_TYPE:
		switch(prm)
		{
		case 0: return FUEL_TYPE::HYDROGEN_CRYOGENIC; // LH2
		case 1: return FUEL_TYPE::MONOMETHYLHYDRAZINE; // RCS
		default: return FUEL_TYPE::UNKNOWN_FUEL;
		};
	case VMSG_GET_OXIDIZER_TYPE:
		switch(prm)
		{
		case 0: return OXIDIZER_TYPE::UNKNOWN_OXIDIZER;
		case 1: return OXIDIZER_TYPE::NITROGEN_TETROXIDE; // RCS
		default: return OXIDIZER_TYPE::UNKNOWN_OXIDIZER;
		};
	case VMSG_GET_TANK_DRAWDOWN_SEQUENCE:
		switch(prm)
		{
		case 0: return 3; //
		case 1: return -1; // RCS
		default: return -1;
		};
	case VMSG_GET_POWER_OUTPUT_W:
		return 0;
	case VMSG_SEND_COMMAND:
		InputConfigParam(0, (char*) context, this);
		return 1;
	default: return -1;
	};
}

// TODO: line purges before startup??

// how do I set the required roll rate

void NTR_test::clbkDockEvent  ( int  dock,  OBJHANDLE  mate)
{
	// check for propellant to use from docked vessels.
	shallFillDrawdownSeq = true;
	//FillDrawdownSequence();
	// should extinguish lights after docking and turn them on after undocking
	if (this->DockingStatus(dock) == 0) // turn on the beacons
	{
		for(int i=0; i<NUM_BEACONS; i++) beacons[i].active = true;
	}
	else
	{
		for(int i=0; i<NUM_BEACONS; i++) beacons[i].active = false;
	};
}

void NTR_test::clbkPostCreation  (   ) 
{
	old_STP = this->GetAtmPressure();
	shallFillDrawdownSeq = true;
	//FillDrawdownSequence();
	//ComputeOptimumRCS(200.0);
	if (GetStackGroundContact() || (this->GetAltitude()<300.0)) 
	{
		IsInLaunchMode = true;
		for (int i = 0; i<NUM_BEACONS; i++) beacons[i].active = false;
	};
	oldNozzleB = oldNozzleG = oldNozzleR = 0.3f;//??
	/*for (double T = 500.0; T<3050.0; T+=100.0)
	{
		double rr, gg, bb;
		BlackBody2RGB(T, rr, gg, bb);
		char lns[256];
		sprintf_s(lns, 255, "%d K %g %g %g", (int)T, rr, gg, bb);
		oapiWriteLog(lns);
	};*/
	if (sign(this->GetFuelMass())<=0)
	{
		oapiWriteLog("Multistage bug workaround. Adding fuel. If this is not intended, please notify author.");

		this->SetPropellantMass(hpr_LH2, PB_LH2MASS);
		this->SetPropellantMass(hpr_RCS, PB_RCSFUELMASS);
		this->IsInCheckoutMode = false;
		this->IsInLaunchMode = true;
		this->StageState = reactor_state_t::Offline;
	};

}

//double burnTimeError(double x,
//					 double *func_data)
//{
//	double fuelMassToExpend = *func_data;
//	double Tx = x;
//	double TxZ = Tx/ZeroThrustLevel;
//	//double lTx = log(Tx/ZeroThrustLevel);
//	double rhs = (3.0/PB_NTR_ISP)*(
//		(ZeroThrustLevel*StartupPeriod_s*(TxZ-1.0))
//		+(Tx*ShutdownPeriod_s*(1.0-1.0/TxZ)));
//
//	//sprintf_s(oapiDebugString(), 255, "%g %g", rhs, fuelMassToExpend);
//	return (rhs-fuelMassToExpend)*(rhs-fuelMassToExpend);
//}

bool SetCommanded_dV (void *id, char *str, void *usrdata)
{
	NTR_test* ntr = (NTR_test*)(usrdata);
	if (ntr==NULL) return true;
	ntr->commanded_dV = min(max(atof(str),0.0),500000.0); // pretty large!
	// two cases: burn to max, or burn to something lower than max.
	double curMass = ntr->GetStackMass();
	double timeToMax = log(NTR_MAXTHRUST/ZeroThrustLevel)*StartupPeriod_s;
	double fuelMassToMax = (3.0*StartupPeriod_s*ZeroThrustLevel/PB_NTR_ISP)*(exp(timeToMax/StartupPeriod_s)-1.0);
	double timeToShutdownFromMax = log(NTR_MAXTHRUST/ZeroThrustLevel)*ShutdownPeriod_s;
	double fuelMassToShutdownFromMax = (3.0*ShutdownPeriod_s*NTR_MAXTHRUST/PB_NTR_ISP)*(1.0-exp(-timeToShutdownFromMax/ShutdownPeriod_s));
	//double dV_atMax = PB_NTR_ISP*log(curMass/(curMass-fuelMassToMax));
	double e_to_dV_over_isp = exp(ntr->commanded_dV/PB_NTR_ISP);
	double fuelMassToExpend = curMass*(e_to_dV_over_isp-1.0)/e_to_dV_over_isp;
	double deltaMf = fuelMassToExpend-(fuelMassToMax+fuelMassToShutdownFromMax);
	const double flowRateAtMax = 3.0*NTR_MAXTHRUST/PB_NTR_ISP;
	if (deltaMf>=0.0)
	{ // 50:20 59:53
		ntr->estimatedBurnTime_s = (deltaMf/flowRateAtMax)+timeToMax+timeToShutdownFromMax;
		return true;
	}
	// find the limiting thrust level!
	double Tx = ZeroThrustLevel+fuelMassToExpend*PB_NTR_ISP/(3.0*(StartupPeriod_s+ShutdownPeriod_s));
	// TODO: add ullage burn dV component and roll RCS mass losses
	double lnx = log(Tx/ZeroThrustLevel);
	ntr->estimatedBurnTime_s = lnx*(StartupPeriod_s+ShutdownPeriod_s);
	return true;
}

//TODO: venting streams?

//TODO: alternative ways of setting up target coords. 

bool InputRADec (void *id, char *str, void *usrdata)
{
	NTR_test* ntr = (NTR_test*)(usrdata);
	if (ntr==NULL) return true;
	float RA, Dec;
	sscanf_s(str, "%g %g", &RA, &Dec);
	ntr->TrackPointEulers = BrianJ_RADecToEulers((double)RA, (double)Dec);
	ntr->ShallTrackPoint = true;
	ntr->ShallTrackSun = false;
	return true;
}

bool InputQuaternion (void *id, char *str, void *usrdata)
{
	NTR_test* ntr = (NTR_test*)(usrdata);
	if (ntr==NULL) return true;
	float q0, q1, q2, q3;
	sscanf_s(str, "%g %g %g %g", &q0, &q1, &q2, &q3);
	quaternion_t q = _Q1(q0, q1, q2, q3);
	ntr->TrackPointEulers = QuaternionToEuler(q);
	ntr->ShallTrackPoint = true;
	ntr->ShallTrackSun = false;
	return true;
}

bool InputEulers (void *id, char *str, void *usrdata)
{
	NTR_test* ntr = (NTR_test*)(usrdata);
	if (ntr==NULL) return true;
	float p, y, r;
	sscanf_s(str, "%g %g %g", &p, &y, &r);
	VECTOR3 eulers = _V(p*DEG_TO_RAD, y*DEG_TO_RAD, r*DEG_TO_RAD);
	ntr->TrackPointEulers = eulers;
	ntr->ShallTrackPoint = true;
	ntr->ShallTrackSun = false;
	return true;
}

bool InputConfigParam (void *id, char *str, void *usrdata)
{
	NTR_test* ntr = (NTR_test*)(usrdata);
	if (ntr==NULL) return true;
	int ctrlr = -1;
	char * nextChar = NULL;
	bool ModifyController = false;
	if (_strnicmp(str, "DEADBAND", 8)==0)
	{
		return true;
	}
	else if (_strnicmp(str, "LOAD COMMANDS", 13)==0)
	{
		ntr->LoadCommands(str+14);
		return true;
	}
	else if (_strnicmp(str, "PDOT", 4)==0)
	{
		ntr->CommandedRates.data[AXIS::PITCH] = atof(str+4);
		return true;
	}
	else if (_strnicmp(str, "YDOT", 4)==0)
	{
		ntr->CommandedRates.data[AXIS::YAW] = atof(str+4);
		return true;
	}
	else if (_strnicmp(str, "RDOT", 4)==0)
	{
		ntr->CommandedRates.data[AXIS::ROLL] = atof(str+4);
		return true;
	}
	else if (_strnicmp(str, "BURN FOR DV", 11)==0)
	{
		SetCommanded_dV(id, str+11, ntr);
		ntr->SetThrusterGroupLevel_SingleStep(THGROUP_MAIN, 1.0);
		return true;
	}
	else if (_strnicmp(str, "SHUTDOWN NOW", 12)==0)
	{
		ntr->Shutdown();
		return true;
	}
	else if (_strnicmp(str, "SHUTDOWN AT DV", 14)==0)
	{
		if (ARE_PUMPS_WORKING(ntr->StageState)) ntr->commanded_dV = atof(str+14);
		return true;
	}
	else if (_strnicmp(str, "DISCHARGE THRESHOLD", 19)==0)
	{
		ntr->BatteryActivationThreshold = atof(str+19);
		if (ntr->BatteryActivationThreshold<=CDH_Power) ntr->BatteryActivationThreshold = CDH_Power;
		return true;
	}
	else if (_strnicmp(str, "RECOMPUTE RCS AUTO", 18)==0)
	{
		ntr->ShallRecomputeRCSStrings = true;
		return true;
	}
	else if (_strnicmp(str, "RECOMPUTE RCS MANUAL", 20)==0)
	{
		ntr->ShallRecomputeRCSStrings = false;
		return true;
	}
	else if (_strnicmp(str, "BEACONS ON", 10)==0)
	{
		for (int i = 0; i<NUM_BEACONS;i++) ntr->beacons[i].active = true;
		return true;
	}
	else if (_strnicmp(str, "BEACONS OFF", 10)==0)
	{
		for (int i = 0; i<NUM_BEACONS;i++) ntr->beacons[i].active = false;
		return true;
	}
	else if (_strnicmp(str, "AUTO", 4)==0)
	{
		if (_strnicmp(str+5, "RCS", 3)==0)
		{
			if (_strnicmp(str+9, "ENABLE", 6)==0)
			{
				ntr->AutoRCSEnabled = true;
				return true;
			}
			else if (_strnicmp(str+9, "INHIBIT", 7)==0)
			{
				ntr->AutoRCSEnabled = false;
				return true;
			}
			else if (_strnicmp(str+9, "LEVEL", 5)==0)
			{
				ntr->AutoRCSTorqueOrThrust = atof(str+15);
				if (ntr->AutoRCSTorqueOrThrust == 0.0) ntr->AutoRCSTorqueOrThrust = 1000.0;
				if (ntr->ShallRecomputeRCSStrings) ntr->ComputeOptimumRCS(ntr->AutoRCSTorqueOrThrust);
				return true;
			}
			else if (_strnicmp(str+9, "SET", 3)==0)
			{
				ntr->SetRCSString(str+13);
				return true;
			}

			else 
			{
				return true;
			};
		}
		else
		{
			return true;
		};
	}
	else if (_strnicmp(str, "TVC", 3)==0)
	{
		ModifyController = true;
		if (_strnicmp(str+4, "PITCH", 5)==0)
		{
			ctrlr = TVC_P;
			nextChar = str+10;
		}
		else if (_strnicmp(str+4, "YAW", 3)==0)
		{
			ctrlr = TVC_Y;
			nextChar = str+8;
		}
		else
		{
			return true;
		};
	}
	else if (_strnicmp(str, "RCS", 3)==0)
	{
		ModifyController = true;
		if (_strnicmp(str+4, "PITCH", 5)==0)
		{
			ctrlr = RCS_P;
			nextChar = str+10;
		}
		else if (_strnicmp(str+4, "YAW", 3)==0)
		{
			ctrlr = RCS_Y;
			nextChar = str+8;
		}
		else if (_strnicmp(str+4, "ROLL", 4)==0)
		{
			ctrlr = RCS_R;
			nextChar = str+9;
		}
		else
		{
			return true;
		};
	}
	else if (_strnicmp(str, "STEADY RATE", 11)==0)
	{
		ntr->DAP_SteadyRate = atof(str+11);
		if (ntr->DAP_SteadyRate<=0.0) ntr->DAP_SteadyRate = RATE_SLOW;
		return true;
	}
	else if (_strnicmp(str, "SLOW RATE", 9)==0)
	{
		ntr->DAP_SlowRate = atof(str+9);
		if (ntr->DAP_SlowRate<=0.0) ntr->DAP_SlowRate = RATE_SLOWEST;
		return true;
	}
	else if (_strnicmp(str, "TRACK QUAT", 10)==0)
	{
		// use Alt-P to drop tracking
		InputQuaternion(id, str+11, (void*)ntr);
		return true;
	}
	else if (_strnicmp(str, "TRACK EULER", 11)==0)
	{
		// use Alt-P to drop tracking
		InputEulers(id, str+12, (void*)ntr);
		return true;
	}
	else if (_strnicmp(str, "TRACK SUN ON", 12)==0)
	{
		ntr->TrackSun(true);
		return true;
	}
	else if (_strnicmp(str, "TRACK SUN OFF", 13)==0)
	{
		ntr->TrackSun(false);
		return true;
	}
	else if (_strnicmp(str, "ATT DB", 6)==0)
	{
		if (_strnicmp(str+7, "PITCH", 5)==0)
		{
			//13
			ntr->Att_Deadbands.data[AXIS::PITCH] = atof(str+13);
			if (ntr->Att_Deadbands.data[AXIS::PITCH]<=0.0) ntr->Att_Deadbands.data[AXIS::PITCH] = DB_FINE;
			return true;
		}
		else if (_strnicmp(str+7, "YAW", 3)==0)
		{
			//11
			ntr->Att_Deadbands.data[AXIS::YAW] = atof(str+11);
			if (ntr->Att_Deadbands.data[AXIS::YAW]<=0.0) ntr->Att_Deadbands.data[AXIS::YAW] = DB_FINE;
			return true;
		}
		else if (_strnicmp(str+7, "ROLL", 4)==0)
		{
			//12
			ntr->Att_Deadbands.data[AXIS::ROLL] = atof(str+12);
			if (ntr->Att_Deadbands.data[AXIS::ROLL]<=0.0) ntr->Att_Deadbands.data[AXIS::ROLL] = DB_FINE;
			return true;
		}
		else
		{
			return true;
		};
	}
	else if (_strnicmp(str, "RATE DB", 7)==0)
	{
		if (_strnicmp(str+8, "PITCH", 5)==0)
		{
			//14
			ntr->Rate_Deadbands.data[AXIS::PITCH] = atof(str+14);
			if (ntr->Rate_Deadbands.data[AXIS::PITCH]<=0.0) ntr->Rate_Deadbands.data[AXIS::PITCH] = RATE_DB_DEFAULT;
			return true;
		}
		else if (_strnicmp(str+8, "YAW", 3)==0)
		{
			//12
			ntr->Rate_Deadbands.data[AXIS::YAW] = atof(str+12);
			if (ntr->Rate_Deadbands.data[AXIS::YAW]<=0.0) ntr->Rate_Deadbands.data[AXIS::YAW] = RATE_DB_DEFAULT;
			return true;
		}
		else if (_strnicmp(str+8, "ROLL", 4)==0)
		{
			//13
			ntr->Rate_Deadbands.data[AXIS::ROLL] = atof(str+13);
			if (ntr->Rate_Deadbands.data[AXIS::ROLL]<=0.0) ntr->Rate_Deadbands.data[AXIS::ROLL] = RATE_DB_DEFAULT;
			return true;
		}
		else
		{
			return true;
		};
	}
	else
	{
		return true;
	};

	if (ModifyController && ctrlr>=0 && ctrlr<NUM_CONTROLLERS)
	{
		if (_strnicmp(nextChar, "PROP", 4)==0)
		{
			ntr->Controllers[ctrlr].proportionalGain = atof(nextChar+4);
		}
		else if (_strnicmp(nextChar, "INT", 3)==0)
		{
			ntr->Controllers[ctrlr].integralGain = atof(nextChar+3);
		}
		else if (_strnicmp(nextChar, "DERIV", 5)==0)
		{
			ntr->Controllers[ctrlr].derivativeGain = atof(nextChar+5);
		}
		else
		{
			return true;
		};
		ntr->Controllers[ctrlr].Reset();
		return true;
	};
	return true;
}

void NTR_test::clbkSaveState  ( FILEHANDLE  scn   )  
{
	VESSEL3::clbkSaveState(scn);
	if (IsInLaunchMode)     oapiWriteLine(scn, "  LAUNCH_MODE");
	if (IsInCheckoutMode)   oapiWriteLine(scn, "  CHECKOUT_MODE");
	if (MaintainCommandedRates)   oapiWriteLine(scn, "  RATE_DAP");
	if (AutoRCSEnabled)           oapiWriteLine(scn, "  AUTORCS ON");
	if (ShallRecomputeRCSStrings) 
	{
		oapiWriteLine(scn, "  RECOMPUTE RCS AUTO");
	}
	else
	{
		oapiWriteLine(scn, "  RECOMPUTE RCS MANUAL");
	};

	char line[1024];
	switch (StageState)
	{
	case reactor_state_t::Meltdown:		oapiWriteLine(scn, "  MELTDOWN");
	case reactor_state_t::UllageBurn:	
		{
			sprintf_s(line, 1023, "  ULLAGE_BURN %g", UllageBurnTimeLeft);
			oapiWriteLine(scn, line);
		};
	case reactor_state_t::Startup:		oapiWriteLine(scn, "  STARTUP");
	case reactor_state_t::Shutdown:		oapiWriteLine(scn, "  SHUTDOWN");
	case reactor_state_t::Nominal:		oapiWriteLine(scn, "  NOMINAL");
	};
	sprintf_s(line, 1023, "  THRUST %g", currentThrust);
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  XE135 %g", XePoisoning);
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  INITMASS %g", massAtBurnStart);
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  AUTORCS LEVEL %g", AutoRCSTorqueOrThrust);
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  DEADBANDS %g %g %g %g %g %g", Att_Deadbands.data[PITCH], 
		Att_Deadbands.data[YAW], 
		Att_Deadbands.data[ROLL], 
		Rate_Deadbands.data[PITCH],
		Rate_Deadbands.data[YAW], 
		Rate_Deadbands.data[ROLL]);
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  DAP CHANNELS - - -");
	if (ActiveDAPChannels[PITCH]) line[15] = '+';
	if (ActiveDAPChannels[YAW]) line[17] = '+';
	if (ActiveDAPChannels[ROLL]) line[19] = '+';
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  STEADY RATE %g", DAP_SteadyRate);
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  SLOW RATE %g", DAP_SlowRate);
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  COMMANDED_DV %g", commanded_dV);
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  NTR_LIFETIME %g", NTR_Life_s);
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  MISSION TIME %g", MissionTime_s);
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  BATTERY CHARGE %g", BatteryCharge);
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  BATTERY ACTIVATION %g", BatteryActivationThreshold);
	oapiWriteLine(scn, line);
	sprintf_s(line, 1023, "  BATTERY CYCLES %d", NumBatteryCycles);
	oapiWriteLine(scn, line);
	for (int i = 0; i<NUM_ANIMATIONS; i++)
	{
		sprintf_s(line, 1023, "  ANIMATION %d %d %g", i, (int) AnimationStatus[i], AnimationDeployments[i]);
		oapiWriteLine(scn, line);
	};
	int count = 1023;
	char rcs_s[1024];
	for (stack_RCS_config_t::iterator citer = stack_RCS.begin(); citer!=stack_RCS.end(); citer++)
	{
		RCS2string_s(citer->first, citer->second, rcs_s, count);
		sprintf_s(line, 1023, "  AUTORCS SET %s", rcs_s);
		oapiWriteLine(scn, line);
	};
	if (ShallTrackPoint) 
	{
		sprintf_s(line, 1023, "  TRACK POINT %g %g %g", TrackPointEulers.x, TrackPointEulers.y, TrackPointEulers.z);
		oapiWriteLine(scn, line);
	}
	else if (ShallTrackSun)
	{
		oapiWriteLine(scn, "  TRACK SUN");
	};
}

void NTR_test::clbkLoadStateEx (FILEHANDLE scn, void *status)
{
	char *lineChars;
	int num_anim, anim_state;
	float anim_deployment, db_pitch, db_yaw, db_roll, rdb_p, rdb_y, rdb_r, track_x, track_y, track_z;
	StageState = reactor_state_t::Offline;
	MaintainCommandedRates = false;
	AutoRCSEnabled = false;

	while (oapiReadScenario_nextline (scn, lineChars)) {
		if (_strnicmp(lineChars, "LAUNCH_MODE", 11)==0)
		{
			this->IsInLaunchMode = true;
			this->IsInCheckoutMode = false;
		}
		else if (_strnicmp(lineChars, "CHECKOUT_MODE", 13)==0)
		{
			this->IsInLaunchMode = false;
			this->IsInCheckoutMode = true;
		}
		else if (_strnicmp(lineChars, "MELTDOWN", 8)==0)
		{
			StageState = reactor_state_t::Meltdown;
		}
		else if (_strnicmp(lineChars, "ULLAGE_BURN", 11)==0)
		{
			StageState = reactor_state_t::UllageBurn;
			this->UllageBurnTimeLeft = min(max(0.0, atof(lineChars+11)), UllageBurnPeriod_s);
		}
		else if (_strnicmp(lineChars, "STARTUP", 7)==0)
		{
			StageState = reactor_state_t::Startup;
		}
		else if (_strnicmp(lineChars, "SHUTDOWN", 8)==0)
		{
			StageState = reactor_state_t::Shutdown;
		}
		else if (_strnicmp(lineChars, "NOMINAL", 7)==0)
		{
			StageState = reactor_state_t::Nominal;
		}
		else if (_strnicmp(lineChars, "RATE_DAP", 8)==0)
		{
			this->MaintainCommandedRates = true;
		}
		else if (_strnicmp(lineChars, "AUTORCS ON", 10)==0)
		{
			this->AutoRCSEnabled = true;
		}
		else if (_strnicmp(lineChars, "RECOMPUTE RCS AUTO", 18)==0)
		{
			this->ShallRecomputeRCSStrings = true;
		}
		else if (_strnicmp(lineChars, "RECOMPUTE RCS MANUAL", 20)==0)
		{
			this->ShallRecomputeRCSStrings = false;
		}
		else if (_strnicmp(lineChars, "MISSION TIME", 12)==0)
		{
			this->MissionTime_s = max(atof(lineChars+12),0.0);
		}
		else if (_strnicmp(lineChars, "BATTERY CHARGE", 14)==0)
		{
			this->BatteryCharge = min(max(atof(lineChars+14),0.0), BATTERY_INITIAL_CAPACITY);
		}
		else if (_strnicmp(lineChars, "BATTERY ACTIVATION", 18)==0)
		{
			this->BatteryActivationThreshold = max(atof(lineChars+18),0.0);
		}
		else if (_strnicmp(lineChars, "BATTERY CYCLES", 14)==0)
		{
			this->NumBatteryCycles = max(0, atoi(lineChars+14));
		}
		else if (_strnicmp(lineChars, "AUTORCS LEVEL", 13)==0)
		{
			this->AutoRCSTorqueOrThrust = atof(lineChars+13);
			if (AutoRCSTorqueOrThrust<=0.0) this->AutoRCSTorqueOrThrust = 1000.0;
		}
		else if (_strnicmp(lineChars, "AUTORCS SET", 11)==0)
		{
			SetRCSString(lineChars+12);
		}
		else if (_strnicmp(lineChars, "THRUST", 6)==0)
		{
			this->currentThrust = atof(lineChars+6);
		}
		else if (_strnicmp(lineChars, "XE135", 5)==0)
		{
			this->XePoisoning = atof(lineChars+5);
		}
		else if (_strnicmp(lineChars, "INITMASS", 8)==0)
		{
			this->massAtBurnStart = max(0.0, atof(lineChars+8));
		}
		else if (_strnicmp(lineChars, "COMMANDED_DV", 12)==0)
		{
			this->commanded_dV = max(0.0, atof(lineChars+12));
		}
		else if (_strnicmp(lineChars, "NTR_LIFETIME", 12)==0)
		{
			this->NTR_Life_s = max(0.0, atof(lineChars+12));
		}
		else if (_strnicmp(lineChars, "DEADBANDS", 9)==0)
		{
			sscanf_s(lineChars+9, "%g %g %g %g %g %g", &db_pitch, &db_yaw, &db_roll, &rdb_p, &rdb_y, &rdb_r);
			if (db_pitch>0.0 && db_yaw>0.0 && db_roll>0.0)
			{
				Att_Deadbands = _V(db_pitch, db_yaw, db_roll);
			};
			if (rdb_p>0.0 && rdb_y>0.0 && rdb_r>0.0)
			{
				Rate_Deadbands = _V(rdb_p, rdb_y, rdb_r);
			};
		}
		else if (_strnicmp(lineChars, "STEADY RATE", 11)==0)
		{
			DAP_SteadyRate = atof(lineChars+11);
			if (DAP_SteadyRate<=0.0) DAP_SteadyRate = RATE_SLOW;
		}
		else if (_strnicmp(lineChars, "SLOW RATE", 9)==0)
		{
			DAP_SlowRate = atof(lineChars+9);
			if (DAP_SlowRate<=0.0) DAP_SlowRate = RATE_SLOWEST;
		}
		else if (_strnicmp(lineChars, "TRACK SUN", 9)==0)
		{
			ShallTrackPoint = false;
			ShallTrackSun = true;
		}
		else if (_strnicmp(lineChars, "TRACK POINT", 11)==0)
		{
			sscanf_s(lineChars+11, "%g %g %g", &track_x, &track_y, &track_z);
			TrackPointEulers = _V(track_x, track_y, track_z);
			ShallTrackSun = false;
			ShallTrackPoint = true;
		}
		else if (_strnicmp(lineChars, "DAP CHANNELS", 12)==0)
		{
			this->ActiveDAPChannels[PITCH] = (_strnicmp(lineChars+13, "+", 1)==0 ? true : false);
			this->ActiveDAPChannels[YAW] = (_strnicmp(lineChars+15, "+", 1)==0 ? true : false);
			this->ActiveDAPChannels[ROLL] = (_strnicmp(lineChars+17, "+", 1)==0 ? true : false);
		}
		else if (_strnicmp(lineChars, "ANIMATION", 9)==0)
		{
			sscanf_s(lineChars+9, "%d %d %g", &num_anim, &anim_state, &anim_deployment);
			if (num_anim>=0 && num_anim<NUM_ANIMATIONS)
			{
				if (anim_state>=0 && anim_state<simple_anim_status_t::LAST)
				{
					if (anim_deployment>=0.0 && anim_deployment<=1.0)
					{
						AnimationStatus[num_anim] = (simple_anim_status_t) anim_state;
						AnimationDeployments[num_anim] = (double)anim_deployment;
					};
				};
			};
		}
		// TODO: save gimbal state!
		else {          // anything not recognised is passed on to Orbiter
			ParseScenarioLineEx (lineChars, status);
		};
	};
}

void NTR_test::Shutdown()
{
	if (StageState == reactor_state_t::Offline ||
		StageState == reactor_state_t::Meltdown) return;

	if (StageState==reactor_state_t::UllageBurn)
	{
		this->SetThrusterGroupLevel(THGROUP_MAIN,0.0);
		StageState = reactor_state_t::Offline;
		UllageBurnTimeLeft = 0.0;
	}
	else
	{
		StageState = reactor_state_t::Shutdown;
	};
}

void NTR_test::ToggleTrackSun()
{
	ShallTrackSun = (ShallTrackSun ? false: true);
	if (ShallTrackSun) ShallTrackPoint = false;
	CommandedRates = NULL_VECTOR;
}

void NTR_test::TrackSun(const bool ShallTrack)
{
	ShallTrackSun = ShallTrack;
	if (ShallTrackSun) ShallTrackPoint = false;
	CommandedRates = NULL_VECTOR;
}

// TODO: 1EO and 2EO optimum (zero-torque) gimbal setups

int NTR_test::clbkConsumeBufferedKey  ( DWORD  key,  bool  down,  char *  kstate )
{
	if (IsInLaunchMode || IsInCheckoutMode) return 0; // no key processing if a payload or very soon after separation
	if (KEYMOD_CONTROL(kstate))
	{
		if ((key == OAPI_KEY_K) && down) 
		{
			this->Shutdown();
			return 1;
		}
		//if ((key == OAPI_KEY_K) 
		//	&& StageState != reactor_state_t::Offline 
		//	&& StageState != reactor_state_t::Meltdown 
		//	&& down)
		//{
		//	if (StageState==reactor_state_t::UllageBurn)
		//	{
		//		this->SetThrusterGroupLevel(THGROUP_MAIN,0.0);
		//		StageState = reactor_state_t::Offline;
		//		UllageBurnTimeLeft = 0.0;
		//	}
		//	else
		//	{
		//		StageState = reactor_state_t::Shutdown;
		//	};
		//	return 1;
		//}
		else if (down && (key== OAPI_KEY_Y))
		{
			switch (CurrentScreen)
			{
			case ENGINE_CONTROL: 
				CurrentScreen = DAP_CONFIG;
				break;
			case DAP_CONFIG:
				CurrentScreen = POWER_CONTROL;
				break;
			case POWER_CONTROL:
				CurrentScreen = ENGINE_CONTROL;
				break;
			default:
				CurrentScreen = ENGINE_CONTROL;
				break;
			};
			//CurrentScreen = (CurrentScreen == ENGINE_CONTROL ? DAP_CONFIG : ENGINE_CONTROL);
			return 1;
		}
		else if (down && MaintainCommandedRates)
		{
			if (key==OAPI_KEY_1)
			{
				// toggle pitch DAP channel
				ActiveDAPChannels[PITCH] = (ActiveDAPChannels[PITCH]?false:true);
				return 1;
			}
			else if (key==OAPI_KEY_2)
			{
				// toggle yaw DAP channel
				ActiveDAPChannels[YAW] = (ActiveDAPChannels[YAW]?false:true);
				return 1;
			}
			else if (key==OAPI_KEY_3)
			{
				// toggle roll DAP channel
				ActiveDAPChannels[ROLL] = (ActiveDAPChannels[ROLL]?false:true);
				return 1;
			};
			return 0;
		}
		else return 0;// pass-through
	};
	if (KEYMOD_ALT(kstate))
	{
		if ((key == OAPI_KEY_M) 
			&& StageState == reactor_state_t::Offline 
			&& down)
		{
			oapiOpenInputBox("Enter commanded dV (in m/s)",
				SetCommanded_dV,
				0,20, 
				(void*)this);
			return 1;
		}
		else if ((key == OAPI_KEY_C) && down)
		{
			oapiOpenInputBox("Enter configuration parameter",
				InputConfigParam,
				0,90, 
				(void*)this);
			return 1;
		}
		else if ((key == OAPI_KEY_S) && down)
		{
			ToggleTrackSun();
			return 1;
		}
		else if ((key == OAPI_KEY_P) && down)
		{
			if (ShallTrackPoint)
			{
				ShallTrackPoint = false;
			}
			else
			{
				oapiOpenInputBox("Enter right ascension and declination of aimpoint",
					InputRADec,
					0,90, 
					(void*)this);
			};
			return 1;
		}
		else if (ARE_PUMPS_WORKING(StageState) && down)
		{
			if (key == OAPI_KEY_1) 
			{
				GimbalSelect[0] = (GimbalSelect[0] ? false : true);
				return 1;
			}
			else if (key == OAPI_KEY_2)
			{
				GimbalSelect[1] = (GimbalSelect[1] ? false : true);
				return 1;
			}
			else if (key == OAPI_KEY_3)
			{
				GimbalSelect[2] = (GimbalSelect[2] ? false : true);
				return 1;
			} 
			else if (key == OAPI_KEY_8)
			{
				for (int i = 0; i<NUM_REACTORS; i++)
				{
					GimbalSelect[i] = false;
				};
				return 1;
			}
			else if (key == OAPI_KEY_0)
			{
				for (int i = 0; i<NUM_REACTORS; i++)
				{
					GimbalSelect[i] = false;
					GimbalNTR(i, 0.0, 0.0);
				};
				return 1;
			}
			else if (key==OAPI_KEY_4)
			{
				GimbalRollNTR(0.01,-90);
				return 1;
			}
			else if (key==OAPI_KEY_5)
			{
				GimbalRollNTR(0.01,+90);
				return 1;
			};
		};
		// Alt-1 Alt-2 Alt-3 - select/deselect engine for gimballing 
		// Alt-8 deselect all 
		// Alt-0 deselect all and return all engines to default settings.
		if (key == OAPI_KEY_L && down)
		{
			MaintainCommandedRates = (MaintainCommandedRates?false:true);
			if (MaintainCommandedRates)
			{
				//re-do the optimization
				FillDrawdownSequence_and_OptimizeRCS();
			}
			else // null out rates!
			{
				CommandedRates = NULL_VECTOR;
			};
			return 1;
		};
		return 0;
	};
	// test solar1
	if (key == OAPI_KEY_K && down && !KEYMOD_CONTROL(kstate)) 
	{
		// SHORTEN and get rid of duplicates
		if (AnimationStatus[SOLAR1_ANIMATION]==simple_anim_status_t::CLOSED) 
		{
			AnimationStatus[SOLAR1_ANIMATION]=simple_anim_status_t::OPENING;
		}
		else if (AnimationStatus[SOLAR1_ANIMATION]==simple_anim_status_t::OPEN)
		{
			AnimationStatus[SOLAR1_ANIMATION]=simple_anim_status_t::CLOSING;
		};
		if (AnimationStatus[SOLAR2_ANIMATION]==simple_anim_status_t::CLOSED) 
		{
			AnimationStatus[SOLAR2_ANIMATION]=simple_anim_status_t::OPENING;
		}
		else if (AnimationStatus[SOLAR2_ANIMATION]==simple_anim_status_t::OPEN)
		{
			AnimationStatus[SOLAR2_ANIMATION]=simple_anim_status_t::CLOSING;
		};
		if (AnimationStatus[SBAND_ANIMATION]==simple_anim_status_t::CLOSED) 
		{
			AnimationStatus[SBAND_ANIMATION]=simple_anim_status_t::OPENING;
		}
		else if (AnimationStatus[SBAND_ANIMATION]==simple_anim_status_t::OPEN)
		{
			AnimationStatus[SBAND_ANIMATION]=simple_anim_status_t::CLOSING;
		};
		return 1;
	};
	if (MaintainCommandedRates && down)
	{ // think about multiple keypresses - Orbiter doesn't follow the guidelines.
		if (key == OAPI_KEY_DOWN)
		{
			CommandedRates.data[PITCH] += 0.001;
			return 1;
		}
		else if (key == OAPI_KEY_UP)
		{
			CommandedRates.data[PITCH] -= 0.001;
			return 1;
		}
		else if (key == OAPI_KEY_LEFT)
		{
			CommandedRates.data[YAW] += 0.001;
			return 1;
		}
		else if (key == OAPI_KEY_RIGHT)
		{
			CommandedRates.data[YAW] -= 0.001;
			return 1;
		}
		else if (key == OAPI_KEY_INSERT)
		{
			CommandedRates.data[ROLL] -= 0.001;
			return 1;
		}
		else if (key == OAPI_KEY_DELETE)
		{
			CommandedRates.data[ROLL] += 0.001;
			return 1;
		};
	};

	if (ARE_PUMPS_WORKING(StageState) && down)
	{
		if (!MaintainCommandedRates)
		{
			double dPitch = 0.0;
			double dYaw = 0.0;
			if (key == OAPI_KEY_DOWN)
			{
				dPitch = 0.1;
			}
			else if (key == OAPI_KEY_UP)
			{
				dPitch = -0.1;
			}
			else if (key == OAPI_KEY_LEFT)
			{
				dYaw = -0.1;
			}
			else if (key == OAPI_KEY_RIGHT)
			{
				dYaw = 0.1;
			}
			else return 0;
			for (int i = 0; i<NUM_REACTORS; i++)
			{
				if (GimbalSelect[i])
				{
					GimbalNTR(i, NTRGimbal[i][0]+dPitch, NTRGimbal[i][1]+dYaw);
				};
			};
			return 1;//no pass-through!
		};
	};
	return 0;
}


//double GetThrust()
//{
//	return 0.0;
//}

//time. stage. dT

// null rates, counter roll
// are the gimbal limits enough to prevent 1EO tumbling?

// att error - rate error phase plane. 

void NTR_test::SelfDestruct(const char * reason)
{
	char line[1024];
	sprintf_s(line, 1023, "%s self-destructed because %s", this->GetName(), reason);
	oapiWriteLog(line);
	oapiDeleteVessel(this->GetHandle());
}

void NTR_test::clbkPreStep (double simt, double simdt, double mjd)
{
	// process command queue
	if (!CommandList.empty())
	{
		command_t topCommand = CommandList.top();
		if (topCommand.first<=mjd) 
		{
			CommandList.pop();
			InputConfigParam(0, const_cast<char*>(topCommand.second.c_str()), this);
		};
	};
	for (int i = 0; i<NUM_ANIMATIONS; i++)
	{
		if (AnimationStatus[i] == simple_anim_status_t::CLOSING || AnimationStatus[i] == simple_anim_status_t::OPENING)
		{
			double da = simdt*AnimationSpeeds[i];
			if (AnimationStatus[i]==simple_anim_status_t::CLOSING)
			{
				if (AnimationDeployments[i]>0.0)
				{
					AnimationDeployments[i] = max(0.0, AnimationDeployments[i]-da);
				}
				else
				{
					AnimationStatus[i]=simple_anim_status_t::CLOSED;
				};
			}
			else
			{
				if (AnimationDeployments[i]<1.0)
				{
					AnimationDeployments[i] = min(1.0, AnimationDeployments[i]+da);
				}
				else
				{
					AnimationStatus[i]=simple_anim_status_t::OPEN;
				};
			};
			SetAnimation(Animations[i], AnimationDeployments[i]);
		};
	};

	if (IsInLaunchMode)
	{
		// special processing for launches.
		// TODO: possibly add unlocking?
		// all thrusters are disabled.
		AnimationStatus[SOLAR1_ANIMATION] = simple_anim_status_t::CLOSING; // wrong!
		AnimationStatus[SOLAR2_ANIMATION] = simple_anim_status_t::CLOSING;
		AnimationDeployments[SOLAR1_ANIMATION] = 0.0;
		AnimationDeployments[SOLAR2_ANIMATION] = 0.0;

		AnimationStatus[SBAND_ANIMATION] = simple_anim_status_t::CLOSING;
		AnimationDeployments[SBAND_ANIMATION] = 0.0;

		for (unsigned int i =0; i< this->GetThrusterCount(); i++)
		{
			if (this->GetThrusterLevel(this->GetThrusterHandleByIndex(i))>0.0) this->SetThrusterLevel(this->GetThrusterHandleByIndex(i),0.0);
			// not sure it works!
		};

		// self-destruct criteria
		VECTOR3 forces, gravity;
		this->GetForceVector(forces);
		this->GetWeightVector(gravity);
		forces = forces-gravity;
		double ownMass = this->GetMass();
		VECTOR3 acc = forces/ownMass;
		double ownG = length(acc)/G;
		if (ownG> LAUNCH_G_LIMIT) 
		{
			G_limit_criterion_counter++;
		}
		else
		{
			G_limit_criterion_counter = 0;
		};
		if (this->G_limit_criterion_counter>=LAUNCH_G_STEP_LIMIT)	SelfDestruct("g limit exceeded");
		double STP = this->GetAtmPressure();
		double dP_dt = (STP-old_STP)/simdt;
		old_STP = STP;
		if (dP_dt>LAUNCH_STATIC_dP_dt_THRESHOLD) 
		{
			Static_dP_dt_criterion_counter++;
		}
		else
		{
			Static_dP_dt_criterion_counter=0;
		};
		if (this->Static_dP_dt_criterion_counter>=LAUNCH_STATIC_dP_dt_STEP_LIMIT) SelfDestruct("static pressure increased");

		//check for transition to checkout
		OBJHANDLE hEarth = oapiGetGbodyByName("Earth");
		if (hEarth==NULL) 
		{
			oapiWriteLog("No Earth in the system");
			throw "No Earth detected";
		};
		double EarthR = oapiGetSize(hEarth);
		double PeA, PeD, ApD, ApA;
		if (this->GetPeDist(PeD)!=hEarth)
		{
			oapiWriteLog("Not an Earth orbit");
			throw "Not an Earth orbit";
		};
		if (this->GetApDist(ApD)!=hEarth)
		{
			oapiWriteLog("Not an Earth orbit");
			throw "Not an Earth orbit";
		};
		PeA = (PeD - EarthR)/1000.0;
		ApA = (ApD - EarthR)/1000.0;
		if ((PeA>=SAFE_PeA) && (ApA<=SAFE_ApA)) 
		{
			IsInLaunchMode = false;
			IsInCheckoutMode = true;
		};
		return;
	};

	if (IsInCheckoutMode)
	{
		// placeholder code
		if (AnimationStatus[SOLAR1_ANIMATION] == simple_anim_status_t::OPEN
			&& AnimationStatus[SOLAR2_ANIMATION] == simple_anim_status_t::OPEN)
		{
			IsInCheckoutMode = false;
			CurrentScreen = ENGINE_CONTROL;
			return;
		};
		// deploy PVAs
		AnimationStatus[SOLAR1_ANIMATION] = simple_anim_status_t::OPENING;
		AnimationStatus[SOLAR2_ANIMATION] = simple_anim_status_t::OPENING;
		AnimationStatus[SBAND_ANIMATION] = simple_anim_status_t::OPENING;
		for (int i = 0; i<NUM_BEACONS; i++) beacons[i].active = true;
		return;
	};

	if (shallFillDrawdownSeq)
	{
		shallFillDrawdownSeq = false;
		FillDrawdownSequence_and_OptimizeRCS();
	};
	//Sanity check: if we are at 1 kPa DNP mark, disintegrate. KINDA FOOLISH. How do I take off from earth?
	// first off, check for ground contact at start? 
	if (this->GetDynPressure()>=MAX_DYN_PRESSURE)
	{
		SelfDestruct("dynamic pressure limit exceeded");
		return;
	};

	//DONE: tank temperature? n mli for boiloff calcs?? 

	//TODO: how do we notify the IMFD of the vessel's true capabilities?

	// do the PMI and CoG change

	// test colors:
	if (visual!=NULL)
	{
		MESHHANDLE hMesh = GetMesh(visual, MESH_MAIN);
		if (hMesh!=NULL)
		{
			MESHGROUP* Nozzle_C = oapiMeshGroup(hMesh, MESHGRP_NTR_Charlo_NozzleInside);
			// TODO: interpolate between the initial and the heated color (read the color at Visual creation!)
			// turns out no material indexing needed?
			MATERIAL* pMat = oapiMeshMaterial(hMesh, Nozzle_C->MtrlIdx);
			if (pMat!=NULL)
			{
				if (currentThrust>0.0)
				{
					double exT = 2900.0*currentThrust/NTR_MAXTHRUST; // too simplistic!
					double R,G,B;
					BlackBody2RGB(exT, R,G, B);

					pMat->emissive.r = (float)R; //(float)(currentThrust/NTR_MAXTHRUST);
					pMat->emissive.b = (float)B; //1.0f-pMat->emissive.r;
					pMat->emissive.g = (float)G; //(pMat->emissive.r + pMat->emissive.b)/2.0f;
				}
				else //restore old materials
				{
					pMat->emissive.r = oldNozzleR;
					pMat->emissive.b = oldNozzleB;
					pMat->emissive.g = oldNozzleG;
				}
			};
		};
	};

	double LH2Mass = this->GetPropellantMass(hpr_LH2);
	double fuelMass = LH2Mass + this->GetPropellantMass(hpr_RCS);
	double maxMass = PB_LH2MASS + PB_RCSFUELMASS;
	double massRatio = fuelMass/maxMass;
	//interpolate PMI
	this->SetPMI(PB_PMI_FULL*massRatio+PB_PMI_EMPTY*(1.0-massRatio));
	double currentCOG_elev = this->GetCOG_elev();
	double tgtCOG_elev = fullCOG_elev*massRatio+emptyCOG_elev*(1.0-massRatio);
	//sprintf_s(oapiDebugString(), 255, "COG: %g m Tgt: %g m", currentCOG_elev, tgtCOG_elev);
	//this->ShiftCG(_V(0,0,(tgtCOG_elev-currentCOG_elev)));
	this->ShiftCentreOfMass(_V(0,0,(tgtCOG_elev-currentCOG_elev)));// DOES NOT WORK AT ALL
	//this->ShiftCentreOfMass(_V(0,0,tgtCOG_elev));//-currentCOG_elev)));
	//sprintf_s(oapiDebugString(), 255, "COG elev %g", currentCOG_elev);
	// how do I shift CoG?
	if (this->GetThrusterGroupLevel(THGROUP_MAIN)>0.0 
		&& (StageState == reactor_state_t::Offline)
		&& XePoisoning<=MaxAllowedXe135Poisoning)
	{
		StageState = reactor_state_t::UllageBurn;
		UllageBurnTimeLeft = UllageBurnPeriod_s;
		currentThrust = 0.0;
		dV_sinceBurnStart = 0.0;
		massAtBurnStart = this->GetStackMass();
	};
	if (StageState == reactor_state_t::UllageBurn)
	{
		UllageBurnTimeLeft = max(0.0, UllageBurnTimeLeft-simdt);

		//THGROUP_HANDLE thgh = this->GetThrusterGroupHandle(THGROUP_ATT_FORWARD);
		SetThrusterGroupLevel_SingleStep(THGROUP_ATT_FORWARD, 1.0);
		if (UllageBurnTimeLeft <= 0.0)
		{
			SetThrusterGroupLevel_SingleStep(THGROUP_ATT_FORWARD, 0.0);
			StageState = reactor_state_t::Startup;
		};
	};

	VECTOR3 angVel;
	GetAngularVel(angVel);
	VECTOR3 CurrentRates=angVel*DEG;
	VECTOR3 currAtt;
	this->GetGlobalOrientation(currAtt);
	OBJHANDLE hSun = GetCentralStar(solarSystem); //oapiGetGbodyByName("Sun");
	if (ShallTrackSun)
	{
		MaintainCommandedRates = true;
		VECTOR3 SunwardPYR, SunwardEulers;
		if (hSun!=NULL)
		{
			// ADD RATE DEADBANDS!? here?
			SunwardPYR = CalcTargetPYR(this, hSun, SunwardEulers);
			// test the case with 
			CommandedRates = RatesFromAttErrors(currAtt, SunwardEulers, SunwardPYR, Att_Deadbands, CurrentRates, true);
			//CommandedRates.data[AXIS::PITCH] = RateFromAttError(SunwardPYR.data[AXIS::PITCH]*DEG, Att_Deadbands.data[AXIS::PITCH]);
			//CommandedRates.data[AXIS::YAW] = RateFromAttError(SunwardPYR.data[AXIS::YAW]*DEG, Att_Deadbands.data[AXIS::YAW]);
			//CommandedRates.data[AXIS::ROLL] = RateFromAttError(CurrentRates.data[AXIS::ROLL], Att_Deadbands.data[AXIS::ROLL]);//sign?
		};
	}
	else if (ShallTrackPoint)
	{
		MaintainCommandedRates = true;
		VECTOR3 TrackPointPYR;
		//ADD RATE DEADBANDS = here?
		TrackPointPYR = CalcPitchYawRollAngles(this, _V(0,0,1), TrackPointEulers);
		CommandedRates = RatesFromAttErrors(currAtt, TrackPointEulers, TrackPointPYR, Att_Deadbands, CurrentRates, true); // ignore roll
		//CommandedRates.data[AXIS::PITCH] = RateFromAttError(TrackPointPYR.data[AXIS::PITCH]*DEG, Att_Deadbands.data[AXIS::PITCH]);
		//CommandedRates.data[AXIS::YAW] = RateFromAttError(TrackPointPYR.data[AXIS::YAW]*DEG, Att_Deadbands.data[AXIS::YAW]);
		//CommandedRates.data[AXIS::ROLL] = RateFromAttError(CurrentRates.data[AXIS::ROLL], Att_Deadbands.data[AXIS::ROLL]);//sign?
	};

	//TODO: LVLH pointing. What's the roll that minimizes 
	//angle between the antenna and the target ground site (TDRS, target spacecraft etc.)?
	// what's more, we do need to keep the stack steady when the target is nearly perpendicular

	if (MaintainCommandedRates)
	{
		// control pitch and yaw only - actually, can control roll by symmetric rotation
		if (oapiGetTimeAcceleration() > 10) oapiSetTimeAcceleration(10); // avoid bloody mess this way.

		double pitchDelta = CommandedRates.data[AXIS::PITCH]-CurrentRates.data[AXIS::PITCH];
		double yawDelta = CommandedRates.data[AXIS::YAW]-CurrentRates.data[AXIS::YAW];
		double rollDelta = CommandedRates.data[AXIS::ROLL]-CurrentRates.data[AXIS::ROLL];

		double dRoll = Controllers[RCS_R].Step(rollDelta, simdt);
		double dPitch, dYaw;
		if (ActiveDAPChannels[ROLL] && fabs(rollDelta)>Rate_Deadbands.data[AXIS::ROLL])
		{
			if (dRoll>0.0)
			{
				this->SetThrusterGroupLevel_SingleStep(THGROUP_ATT_BANKRIGHT, dRoll); // can be extended to support RCS strings!
			}
			else
			{
				this->SetThrusterGroupLevel_SingleStep(THGROUP_ATT_BANKLEFT, -dRoll);
			};
		};
		if (ARE_PUMPS_WORKING(StageState))
		{
			dPitch = (ActiveDAPChannels[PITCH] ? Controllers[TVC_P].Step(pitchDelta, simdt) : 0.0);
			dYaw = (ActiveDAPChannels[YAW] ? Controllers[TVC_Y].Step(yawDelta, simdt) : 0.0);
			// since TVC costs nothing to us, to the hell with rate deadbands, we're doing it to floating point accuracy! (TODO: to be reviewed)
			for (int i = 0; i<NUM_REACTORS; i++) GimbalNTR(i, dPitch, dYaw);
		}
		else // RCS-based DAP
		{
			dPitch = Controllers[RCS_P].Step(pitchDelta, simdt);
			dYaw = Controllers[RCS_Y].Step(yawDelta, simdt);
			if (ActiveDAPChannels[PITCH] && fabs(pitchDelta)>Rate_Deadbands.data[AXIS::PITCH])
			{
				if (dPitch>0.0)
				{
					this->SetThrusterGroupLevel_SingleStep(THGROUP_ATT_PITCHUP, dPitch);
				}
				else
				{
					this->SetThrusterGroupLevel_SingleStep(THGROUP_ATT_PITCHDOWN, -dPitch);
				};
			};
			if (ActiveDAPChannels[YAW] && fabs(pitchDelta)>Rate_Deadbands.data[AXIS::YAW])
			{
				if (dYaw>0.0)
				{
					this->SetThrusterGroupLevel_SingleStep(THGROUP_ATT_YAWLEFT, dYaw);
				}
				else
				{
					this->SetThrusterGroupLevel_SingleStep(THGROUP_ATT_YAWRIGHT, -dYaw);
				};
			};
		};
	};

	//// vibration
	//if (ARE_PUMPS_WORKING(StageState))
	//{
	//	if (_finite(timeEngineBurnStarted)==0)
	//	{
	//		timeEngineBurnStarted = simt;
	//	}
	//	else
	//	{
	//		double amplitude = 6.0;
	//		double freq = 22.11;//Hz
	//		double dZp = amplitude*sin(PI2*timeEngineBurnStarted/freq);
	//		//VECTOR3 camPos = PB_DOCK_POS;
	//		//camPos.z += dZp;
	//		//this->SetCameraOffset(camPos);
	//		sprintf_s(oapiDebugString(), 255, "dZp: %g", dZp);
	//		timeEngineBurnStarted = simt;
	//	};

	//}
	//else
	//{
	//	if (_finite(timeEngineBurnStarted)!=0)
	//	{
	//		timeEngineBurnStarted = numeric_limits<double>::infinity();
	//		//this->SetCameraOffset(PB_DOCK_POS);
	//	};
	//};

	// TODO: integrate burn and track final velocity vector - 

	// implicit integration scheme, perhaps... now where are we?

	// flight plan - to be displayed somewhere... AR&D...

	// TODO: add RCS string + TVC

	//TODO: link up power and cryo systems to the next tank(s)

	if (ARE_PUMPS_WORKING(StageState)) 
	{
		if (LH2Mass<=0.0)
		{
			StageState = reactor_state_t::Meltdown;
			currentThrust = 0.0;
			reactorPeriod_s = -ShutdownPeriod_s;
		}
		else if (StageState == reactor_state_t::Shutdown)
		{
			reactorPeriod_s = -ShutdownPeriod_s;
			currentThrust = currentThrust*exp(simdt/reactorPeriod_s);
			currentThrust = max(0.0,min(NTR_MAXTHRUST, currentThrust));
			if (currentThrust<=ZeroThrustLevel)
			{
				currentThrust = 0.0;
				StageState = reactor_state_t::Offline;
				this->SetThrusterGroupLevel(THGROUP_MAIN, 0.0);
				XePoisoning = 1.0; // no way you can work with it!
				commanded_dV = 0.0; //re-set dV
			};
		}
		else if (StageState == reactor_state_t::Startup)
		{
			reactorPeriod_s = StartupPeriod_s;
			if (currentThrust==0.0) {currentThrust = ZeroThrustLevel;} else currentThrust = currentThrust*exp(simdt/reactorPeriod_s);
			currentThrust = max(0.0,min(NTR_MAXTHRUST, currentThrust));
			if (currentThrust>=NTR_MAXTHRUST)
			{
				StageState = reactor_state_t::Nominal;
			};
		}
		else if (StageState == reactor_state_t::Nominal)
		{
			currentThrust = NTR_MAXTHRUST;
			reactorPeriod_s = numeric_limits<double>::infinity();
		}
		else
		{
			currentThrust = 0.0;
		};
		for (int i = 0; i<NUM_REACTORS; i++)
		{
			SetThrusterLevel_SingleStep(th_NTR[i], currentThrust/NTR_MAXTHRUST);
		};
		//if (MaintainCommandedRates)
		//{
		//	// control pitch and yaw only - actually, can control roll by symmetric rotation
		//	VECTOR3 angVel;
		//	GetAngularVel(angVel);
		//	VECTOR3 CurrentRates=angVel*DEG;

		//	double pitchDelta = CommandedRates.data[AXIS::PITCH]-angVel.data[AXIS::PITCH];
		//	double yawDelta = CommandedRates.data[AXIS::YAW]-angVel.data[AXIS::YAW];
		//	double rollDelta = CommandedRates.data[AXIS::ROLL]-angVel.data[AXIS::ROLL];

		//	double dRoll = RCSControlR.Step(rollDelta, simdt);
		//	double dPitch=TVCControlP.Step(pitchDelta, simdt);
		//	double dYaw=TVCControlY.Step(yawDelta, simdt);
		//	if (dRoll>0.0)
		//	{
		//		this->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_BANKRIGHT, dRoll);
		//	}
		//	else
		//	{
		//		this->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_BANKLEFT, dRoll);
		//	};
		//	for (int i = 0; i<NUM_REACTORS; i++) GimbalNTR(i, dPitch, dYaw);
		//};
		//if (NullTorque)
		//{
		////how do we do TVC? absolutely necessary for 1EO cases.
		// how do we null gravity-gradient torque?
		//vector<stack_thruster_t> thrusters;
		//vector<stack_fuel_tank_t> tanks;
		//VECTOR3 stack_COG, stack_PMI;
		//TransformStackThrusterInfoIntoRootFrame(stackInfo, 
		//										thrusters, 
		//										tanks, 
		//										stack_COG, 
		//										stack_PMI);

		//VECTOR3 torque = _V(0,0,0);
		//VECTOR3 torque2 = _V(0,0,0);
		//unsigned int th_indices[NUM_REACTORS];
		//unsigned int k = 0;
		//for (unsigned int i = 0; i<thrusters.size(); i++)
		//{
		//	if (thrusters[i].pV == this)
		//	{
		//		for (unsigned int j = 0; j<NUM_REACTORS; j++)
		//		{
		//			if (this->GetThrusterHandleByIndex(thrusters[i].thrusterID)==th_NTR[j])
		//			{
		//				th_indices[k] = i;
		//				k++;
		//				break;
		//			};
		//		};
		//	};
		//};
		//for (unsigned int i = 0; i<NUM_REACTORS; i++)
		//{
		//	stack_thruster_t& st = thrusters.at(th_indices[i]);
		//    torque = torque-crossp(st.root_pos-stack_COG, currentThrust*st.root_dir);
		//};
		// linear uncoupled regulator:
		//	VECTOR3 currAngVel;
		//	this->GetAngularVel(currAngVel);
		//	if (fabs(currAngVel.y)>1e-10)
		//	{
		//		VECTOR3 ALICE_Y = _V(0.0, -0.01*currAngVel.y, sqrt(1.0-0.01*currAngVel.y*0.01*currAngVel.y));
		//		//normalise(ALICE_Y);
		//		this->SetThrusterDir(th_NTR[NTR_Alice], ALICE_Y);
		//	};
		//	for (unsigned int i = 0; i<NUM_REACTORS; i++)
		//	{
		//		VECTOR3 dir, pos;
		//		this->GetThrusterDir(th_NTR[i], dir);
		//		this->GetThrusterRef(th_NTR[i], pos);
		//	    torque2 = torque2-crossp(pos, currentThrust*dir);
		//	};
		//	//sprintf_s(oapiDebugString(), 255, "engine-induced torque x: %g y: %g z: %g", torque.x, torque.y, torque.z);
		//	sprintf_s(oapiDebugString(), 255, "engine-induced torque x: %g y: %g z: %g", torque2.x, torque2.y, torque2.z);

		//}
		//else
		//{
		//	for (int i = 0; i<NUM_REACTORS; i++)
		//	this->SetThrusterDir(th_NTR[i], _V(0,0,1));
		//};

	}
	else
	{
		for (int i = 0; i<NUM_REACTORS; i++)
		{
			SetThrusterLevel_SingleStep(th_NTR[i], 0.0);
		};
		if (XePoisoning>0.0) XePoisoning = XePoisoning*exp(-HALFLIFE_CONSTANT*simdt/Xe135_HalfLife_s);
	};
}

// TODO: DAP FREE, DAP AUTO, DAP HOLD

//from SSU: with side effects!!
bool NTR_test::GimbalNTR(int engine, double pitch, double yaw)
{
	if(abs(pitch)<=NTR_PITCH_GIMBAL_LIMIT && abs(yaw)<=NTR_YAW_GIMBAL_LIMIT) {
		NTRGimbal[engine][0]=pitch;
		NTRGimbal[engine][1]=yaw;

		VECTOR3 dir = _V(0,0,1);

		dir=RotateVectorX(dir, -pitch); //positive OMS gimbal directs thrust downwards
		dir=RotateVectorY(dir, -yaw); //positive yaw gimbal directs thrust to right

		SetThrusterDir(th_NTR[engine], dir);

		//char cbuf[255];
		//sprintf_s(cbuf, 255, "NTR DIR: %d %f %f %f", engine, dir.x, dir.y, dir.z);
		//oapiWriteLog(cbuf);
		return true;
	}
	else
	{
		SetThrusterDir(th_NTR[engine], _V(0,0,1));
		return false; // safe?
	};
}

VECTOR3 cTorque(const double Thrust, const VECTOR3 POS, const VECTOR3 COG, const double Pitch, const double Yaw)
{
	VECTOR3 R;
	VECTOR3 dir = _V(0, 0, 1);
	dir = RotateVectorX(dir, -Pitch);
	dir = RotateVectorY(dir, -Yaw);
	VECTOR3 shoulder = POS - COG;
	VECTOR3 thV = Thrust * dir;
	R = crossp(shoulder, thV);
	return R;
}


double totalTorque(unsigned n, const double *x,
				   double *gradient, /* NULL if not needed */
				   void *func_data)
{
	gimbal_problem_data_t * gpdt = (gimbal_problem_data_t*) func_data;
	if (n!=gpdt->numThrusters*2)
	{
		oapiWriteLog("n!=numThrusters*2");
		throw "n!=numThrusters*2";
	};
	double ff;
	VECTOR3 Torq = _V(0, 0, 0);
	for (int i=0; i<gpdt->numThrusters; i++)
	{
		Torq = Torq + cTorque(gpdt->Thrust[i], gpdt->Pos[i], gpdt->COG, (*(x+2 * i)), (*(x+2 * i + 1)));
	};
	ff = Torq.x * Torq.x + Torq.y * Torq.y + Torq.z * Torq.z;
	gradient = NULL;
	return ff;
}


//WORKS!!!
void NTR_test::OptimizeTorque(const vector<double> & thrust, vector<double> &pitch, vector<double> &yaw)
{
	// TODO: make optimization problem a private member of the class, to avoid allocation/deallocation
	if (thrust.size()!=NUM_REACTORS)
	{
		oapiWriteLog("thrust vector sized incorrectly");
		throw "thrust vector sized incorrectly";
	};
	vector<double> angles;
	angles.resize(NUM_REACTORS*2);
	double residTorqueMag;
	gimbal_problem_data_t gpdt;
	gpdt.numThrusters = NUM_REACTORS;
	gpdt.Pos.resize(NUM_REACTORS);
	//// need COG data from the stack!
	//stackInfo
	//th_NTR[1]
	gpdt.Pos[0] = V_Alice;
	gpdt.Pos[1] = V_Betty;
	gpdt.Pos[2] = V_Clara;
	//// TODO: make generic?
	gpdt.Thrust.resize(NUM_REACTORS);
	for (int i=0; i<NUM_REACTORS; i++) gpdt.Thrust[i] = thrust[i];
	gpdt.COG = _V(0, 0, 2);
	for (int i = 0; i<NUM_REACTORS*2; i++) angles[i] = 0.0;
	optProblem->set_lower_bounds(-NTR_PITCH_GIMBAL_LIMIT);
	optProblem->set_upper_bounds(NTR_PITCH_GIMBAL_LIMIT);
	optProblem->set_min_objective(totalTorque, &gpdt);
	nlopt::result res = optProblem->optimize(angles, residTorqueMag);
	if (res==nlopt::result::SUCCESS || res==nlopt::result::FTOL_REACHED || res==nlopt::result::XTOL_REACHED )
	{
		for (int i = 0; i<NUM_REACTORS; i++)
		{
			pitch[i] = angles[2*i];
			yaw[i] = angles[2*i+1];
		};
		//	char lx[1024];
		//	sprintf_s(lx, 1023, "opt %g - res %d: %g %g %g %g %g %g", residTorqueMag, res, angles[0],angles[1],angles[2],angles[3],angles[4],angles[5]);
		//	oapiWriteLog(lx);
	}
	else // fallback?
	{
		for (int i = 0; i<NUM_REACTORS; i++)
		{
			pitch[i] = 0.0;
			yaw[i] = 0.0;
		};
	};

	//else
	//{
	//	char lx[1024];
	//	sprintf_s(lx, 1023, "res: %d", res);
	//	oapiWriteLog(lx);
	//}
}

bool NTR_test::GimbalRollNTR(double R, double theta_deg)
{
	// how do we enforce gimbal limits?
	if (abs(R)<sin(NTR_PITCH_GIMBAL_LIMIT*RAD))
	{

		VECTOR3 dirs[3];
		VECTOR3 pyrs[3];
		for (int i = 0; i<3; i++)
		{
			dirs[i].x = R*sin((i*120.0+theta_deg)*RAD);
			dirs[i].y = R*cos((i*120.0+theta_deg)*RAD);
			//dirs[i].x = R*cos((i*120.0+theta_deg)*RAD);
			//dirs[i].y = R*sin((i*120.0+theta_deg)*RAD);
			dirs[i].z = sqrt(1.0-dirs[i].x*dirs[i].x-dirs[i].y*dirs[i].y);
			pyrs[i] = GetPitchYawRoll(dirs[i], NULL_VECTOR); // smth happening? maybe we start from the wrong heading?
			NTRGimbal[i][0]=pyrs[i].data[AXIS::PITCH];
			NTRGimbal[i][1]=pyrs[i].data[AXIS::YAW];

			//VECTOR3 dir = _V(0,0,1);

			//dir=RotateVectorX(dir, -pitch); //positive OMS gimbal directs thrust downwards
			//dir=RotateVectorY(dir, -yaw); //positive yaw gimbal directs thrust to right
			SetThrusterDir(th_NTR[i], dirs[i]);
		};

		//char cbuf[255];
		//sprintf_s(cbuf, 255, "NTR DIR: %d %f %f %f", engine, dir.x, dir.y, dir.z);
		//oapiWriteLog(cbuf);
		return true;
	}
	else return false;
}


// TODO: visual heating up of the engine nozzles by changing in-place the material definition.

// commanded rates(output) - in degrees, attitude vectors in radians, deadbands in degrees, current rates in degrees
VECTOR3 NTR_test::RatesFromAttErrors(const VECTOR3 CurrentAttitude, const VECTOR3 DesiredAttitude, 
									 const VECTOR3 AttError,
									 const VECTOR3 AttDeadbands, const VECTOR3 CurrentRates, const bool IgnoreRoll)
{
	double RotAngle;
	VECTOR3 RotAxis;
	//take into account torques!
	// add bias!
	VECTOR3 DA = DesiredAttitude;
	//if (IgnoreRoll) DA.data[ROLL] = CurrentAttitude.data[ROLL]; // probably wrong
	FindAngleAndAxis_KALCMANU(CurrentAttitude, DesiredAttitude, RotAxis, RotAngle);
	VECTOR3 currentPlusZ, targetPlusZ;
	RotateVector(_V(0,0,1), CurrentAttitude, currentPlusZ);
	RotateVector(_V(0,0,1), DesiredAttitude, targetPlusZ);
	double dAng = angsep(currentPlusZ, targetPlusZ) * DEG; // degrees


	if (RotAngle*DEG<=5.0 || (IgnoreRoll && dAng<=5.0)) // add check for angsep if we're not interested in correct roll.
	{
		// manipulate gains?
		
		// return RateFromAttError
		RotAxis.data[PITCH] = RateFromAttError(AttError.data[PITCH]*DEG,Att_Deadbands.data[PITCH]);
		RotAxis.data[YAW] = RateFromAttError(AttError.data[YAW]*DEG,Att_Deadbands.data[YAW]);
		if (IgnoreRoll)
		{
			RotAxis.data[ROLL] = RateFromAttError(CurrentRates.data[ROLL],Att_Deadbands.data[ROLL]); //?
		}
		else
		{
			RotAxis.data[ROLL] = RateFromAttError(AttError.data[ROLL],Att_Deadbands.data[ROLL]); //?
		};
		return RotAxis;
	}
	else
	{
		// make speeds proportional to the rotaxis
		// maximum component is equal to steady rate!
		normalise(RotAxis);
		return -DAP_SteadyRate*RotAxis;
	};
}
//TODO: augment attitudes with the requirement to point antenna in a certain direction by rolling.
// TODO: list of tracking stations.
// WHAT TO DO WITH RCS pressurant??


// TODO: hold vs maneuver! rate delta/att delta Phase plane - TBD
double NTR_test::RateFromAttError(const double AttError, const double Deadband, const double RateError) //, const int Axis)
{
	//if (fabs(AttError)<NullStartAngle(
	// THIS IS PATENTLY WRONG! FIXME FIXME FIXME
	// need current rate as an input! won't work for massive underequipped stacks like core+2 tanks!
	if (fabs(AttError)<Deadband-0.3*DB_OFFSET) return 0.0; // overshoot possible
	if (fabs(AttError)>20.0*Deadband) return (AttError>0.0?-DAP_SteadyRate:DAP_SteadyRate);
	if (fabs(AttError)>5.0*Deadband) return (AttError>0.0?-DAP_SlowRate:DAP_SlowRate);
	double slope = (-DAP_SlowRate+0.001)/(4.0*Deadband);
	double SUPERSLOW = DAP_SlowRate/10.0; //0.005;
	double intercept = -DAP_SlowRate*0.25+SUPERSLOW*1.2;
	return intercept+slope*AttError; //(AttError>0.0? -intercept-slope*AttError:intercept+slope*AttError);
}

// engine status display
bool NTR_test::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);


	oapiSetDefNavDisplay(1);
	oapiSetDefRCSDisplay(1);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
		roxl = (lw*10);
		royl = (int) (-88*d);
	}


	// Time to burn. sync.

	if (IsInLaunchMode)
	{
		CurrentScreen = glass_screen_t::LAUNCH;
	}
	else if (IsInCheckoutMode)
	{
		CurrentScreen = glass_screen_t::CHECKOUT;
	};


	switch (CurrentScreen)
	{
	case LAUNCH:
		{
			HUDTextOut_CLR(skp, 10, lineY[0], GREY, "Reactors are safed. Vehicle is in LAUNCH MODE.");
			HUDTextOut_CLR(skp, 10, lineY[1], GREY, "Self-destruct criteria: static dP/dt = + 0.1kPa/second for %d steps OR acc>7g for %d steps", LAUNCH_STATIC_dP_dt_STEP_LIMIT, LAUNCH_G_STEP_LIMIT);
			HUDTextOut_CLR(skp, 10, lineY[2], GREY, "Safe parking orbit: PeA>=%d km ApA<=%d km", (int)SAFE_PeA, (int)SAFE_ApA);
			return true;
		};
	case CHECKOUT:
		{
			HUDTextOut_CLR(skp, 10, lineY[0], GREY, "Reactors are safed. Vehicle is in CHECKOUT MODE.");
			return true;
		};
	case ENGINE_CONTROL:
		{
			HUDTextOut_CLR(skp, 10, lineY[0], GOLD, "ENGINE CONTROL [Ctrl-Y]-switch to DAP CONFIG [Ctrl-K]-shutdown [Alt-M]-required dV");
			//"[Alt-123]-toggle engine TVC [Alt-8]-deselect all TVC [Alt-0]-reset TVC"

			if (StageState == reactor_state_t::UllageBurn) HUDTextOut_CLR(skp, 10, lineY[1], GOLD, "Ullage burn");
			if (StageState == reactor_state_t::Startup)    HUDTextOut_CLR(skp, 10, lineY[1], GOLD, "Startup initiated");
			if (StageState == reactor_state_t::Shutdown) 
			{
				if (FuelLevelTripActivated)
				{
					HUDTextOut_CLR(skp, 10, lineY[1], RED, "Shutdown initiated (FUEL LOW)");
				}
				else
				{
					HUDTextOut_CLR(skp, 10, lineY[1], GOLD, "Shutdown initiated");
				};
			};
			if (StageState == reactor_state_t::Meltdown)
			{
				HUDTextOut_CLR(skp, 10, lineY[3], RED, "[ALICE] MELTDOWN"); // what to do with the idiot?
				HUDTextOut_CLR(skp, 10, lineY[4], RED, "[BETTY] MELTDOWN");
				HUDTextOut_CLR(skp, 10, lineY[5], RED, "[CLARA] MELTDOWN");
			}
			else if (StageState != reactor_state_t::Offline && StageState != reactor_state_t::UllageBurn)
			{
				// 1 degree is too much for TVC adjustment!
				// tpa can be ok or cross-feeding the neighbour or down:

				// ok XfA XfB XfC OUT
				double flowRate = this->GetPropellantFlowrate(this->GetPropellantHandleByIndex(0))/3.0;
				HUDTextOut_CLR(skp, 10, lineY[2], GOLD, "Reactor  Thrust:  LH2 flow:   Period: TPA1 TPA2 Pitch   Yaw   TVC");
				HUDTextOut_CLR(skp, 10, lineY[3], GOLD, "[ALICE] %5d N   %3.1f kg/s   %3.0f s    ok   ok   %+1.1f   %+1.1f    %d", (int)currentThrust, flowRate, reactorPeriod_s, NTRGimbal[NTR_Alice][0], NTRGimbal[NTR_Alice][1], (int)(GimbalSelect[NTR_Alice]));
				HUDTextOut_CLR(skp, 10, lineY[4], GOLD, "[BETTY] %5d N   %3.1f kg/s   %3.0f s    ok   ok   %+1.1f   %+1.1f    %d", (int)currentThrust, flowRate, reactorPeriod_s, NTRGimbal[NTR_Betty][0], NTRGimbal[NTR_Betty][1], (int)(GimbalSelect[NTR_Betty]));
				HUDTextOut_CLR(skp, 10, lineY[5], GOLD, "[CLARA] %5d N   %3.1f kg/s   %3.0f s    ok   ok   %+1.1f   %+1.1f    %d", (int)currentThrust, flowRate, reactorPeriod_s, NTRGimbal[NTR_Clara][0], NTRGimbal[NTR_Clara][1], (int)(GimbalSelect[NTR_Clara]));
				double currentMass = this->GetStackMass();
				dV_sinceBurnStart = PB_NTR_ISP*log(massAtBurnStart/currentMass );
				// what is the final mass assuming we stop burn now?
				if (currentThrust>0.0)
				{
					double timeToShutdown = -log(ZeroThrustLevel/currentThrust)*ShutdownPeriod_s;
					double fuelMassToShutdown = (3.0*ShutdownPeriod_s*currentThrust/PB_NTR_ISP)*(1.0-exp(-timeToShutdown/ShutdownPeriod_s));
					double dV_atShutdown = PB_NTR_ISP*log(massAtBurnStart/(currentMass-fuelMassToShutdown));

					if (StageState == reactor_state_t::Shutdown)
					{
						HUDTextOut_CLR(skp, 10, lineY[7], GOLD, "dV since burn start: %10.2f final dV: %10.2f", dV_sinceBurnStart, dV_atShutdown);
					}
					else
					{
						HUDTextOut_CLR(skp, 10, lineY[7], GOLD, "dV since burn start: %10.2f final dV if shutdown now: %10.2f", dV_sinceBurnStart, dV_atShutdown);
					};
					VECTOR3 GlobalVelocity;
					this->GetGlobalVel(GlobalVelocity);
					HUDTextOut_CLR(skp, 10, lineY[8], GREEN, "VX: %10.2f VY: %10.2f VZ: %10.2f", GlobalVelocity.x, GlobalVelocity.y, GlobalVelocity.z);
				}
				else
				{			
					HUDTextOut_CLR(skp, 10, lineY[7], GOLD, "dV since burn start: %10.2f", dV_sinceBurnStart);
				};
			}
			else
			{
				COLORREF XeColor = (XePoisoning<=MaxAllowedXe135Poisoning?GREEN:RED);
				HUDTextOut_CLR(skp, 10, lineY[3], XeColor, "[ALICE] Xe-135 %d", (int)(XePoisoning*100.0)); // concentration of xenon? or simply time to next burn?
				HUDTextOut_CLR(skp, 10, lineY[4], XeColor, "[BETTY] Xe-135 %d", (int)(XePoisoning*100.0));
				HUDTextOut_CLR(skp, 10, lineY[5], XeColor, "[CLARA] Xe-135 %d", (int)(XePoisoning*100.0));

			};
			VECTOR3 Torque;
			this->GetTorqueVector(Torque);
			HUDTextOut_CLR(skp, 10, lineY[14], GREEN, "Torque X: %+10.9f Y: %+10.9f Z: %+10.9f", Torque.x, Torque.y, Torque.z);
			char pitchChannel[2],yawChannel[2], rollChannel[2];
			pitchChannel[1] = yawChannel[1] = rollChannel[1] = '\0';
			pitchChannel[0] = (ActiveDAPChannels[PITCH]? '*' : '~');
			yawChannel[0] = (ActiveDAPChannels[YAW]? '*' : '~');
			rollChannel[0] = (ActiveDAPChannels[ROLL]? '*' : '~');
			if (MaintainCommandedRates)
			{
				HUDTextOut_CLR(skp, 10, lineY[15], GREEN, "[Alt-L] - rate DAP ON [PDOT%s %+2.3f/s YDOT%s %+2.3f/s RDOT%s %+2.3f/s] [Ctrl-123]-toggle channels", pitchChannel, CommandedRates.data[PITCH], yawChannel, CommandedRates.data[YAW], rollChannel, CommandedRates.data[ROLL]);
			}
			else
			{
				HUDTextOut_CLR(skp, 10, lineY[15], GREEN, "[Alt-L] - ATT FREE; DAP is off ...");
			};


			VECTOR3 SunwardPYR, SunwardEulers;
			OBJHANDLE hSun = GetCentralStar(solarSystem); //oapiGetGbodyByName("Sun");

			OBJHANDLE hPlanetRef = this->GetGravityRef();
			double SolarBeta=0.0;
			double EclipseFraction = 0.0;
			double EclipseLength = 0.0;
			if (hPlanetRef!=hSun)
			{
				// beta angle
				VECTOR3 p_rel_velocity, p_rel_pos, p_rel_sun_pos;
				oapiGetRelativeVel(this->GetHandle(), hPlanetRef, &p_rel_velocity);
				oapiGetRelativePos(this->GetHandle(), hPlanetRef, &p_rel_pos);
				oapiGetRelativePos(hSun, hPlanetRef, &p_rel_sun_pos);
				VECTOR3 orb_normal = crossp(p_rel_pos, p_rel_velocity); //which hand???
				// more or less like it. at least the zero-beta orbit has Sun in the orbital plane
				SolarBeta = HALFPI-acos(dotp(unit(orb_normal),unit(p_rel_sun_pos)));
				ELEMENTS orbelm;
				ORBITPARAM orbparam;
				GetElements  ( hPlanetRef, orbelm, &orbparam, 0.0, FRAME_EQU);
				if (orbelm.e<0.5) //??use KOST to predict eclipse time!
				{
					double R_planet = oapiGetSize(hPlanetRef);
					double h_SC = orbelm.a-oapiGetSize(hPlanetRef);
					EclipseFraction = CircularOrbitEclipseFraction(R_planet, 
						h_SC, 
						SolarBeta);
					EclipseLength = orbparam.T * EclipseFraction;
				};
			};



			// sign of sine??
			//double solarHeatLoad = CylinderProjectionArea(acos(dot_p_ab_sun), NTR_cyl_length, NTR_cyl_diameter)*insolation;

			if (hSun!=NULL)
			{
				SunwardPYR = CalcTargetPYR(this, hSun, SunwardEulers);
				//double RotAng;
				//VECTOR3 RotAxis;
				//VECTOR3 currAtt;
				//this->GetGlobalOrientation(currAtt);
				//FindAngleAndAxis_KALCMANU(currAtt, SunwardEulers, RotAxis, RotAng);
				if (ShallTrackSun)
				{
					HUDTextOut_CLR(skp, 10, lineY[16], GOLD, "Sunward attitude errors P:%+3.2f Y:%+3.2f [Alt-S]-no Sun track {TRACKING SUN} [Alt-P]-track point", SunwardPYR.data[PITCH]*DEG, SunwardPYR.data[YAW]*DEG);
					//HUDTextOut_CLR(skp, 10, lineY[16], GOLD, "Sunward attitude errors P:%+3.2f Y:%+3.2f A:%g UX:%g UY:%g UZ:%g [Alt-S]-no Sun track {TRACKING SUN} [Alt-P]-track point", SunwardPYR.data[PITCH]*DEG, SunwardPYR.data[YAW]*DEG, RotAng*DEG, RotAxis.x*DEG, RotAxis.y*DEG, RotAxis.z*DEG);
				}
				else if (ShallTrackPoint)
				{
					VECTOR3 TrackPYR;
					TrackPYR = CalcPitchYawRollAngles(this, _V(0,0,1), TrackPointEulers);

					HUDTextOut_CLR(skp, 10, lineY[16], GOLD, "Designated point errors P:%+3.2f Y:%+3.2f [Alt-S]-Sun track {POINT TRACKING} [Alt-P]-no track point", TrackPYR.data[PITCH]*DEG, TrackPYR.data[YAW]*DEG);
				}
				else
				{
					HUDTextOut_CLR(skp, 10, lineY[16], GOLD, "Sunward attitude errors P:%+3.2f Y:%+3.2f [Alt-S]-Sun track [Alt-P]-track point", SunwardPYR.data[PITCH]*DEG, SunwardPYR.data[YAW]*DEG);
				};

				// implement automatic sunward tracking with deadbands!
			};

			int hrs = (int)floor(NTR_Life_s/3600.0);
			int mins = (int)floor((NTR_Life_s-3600.0*hrs)/60.0);
			int seconds = (int)(NTR_Life_s-3600.0*hrs-60.0*mins);
			HUDTextOut_CLR(skp, 10, lineY[6], RED, "Remaining engine lifetime: %d h %d min %d s", hrs, mins, seconds);

			if (commanded_dV>0.0)
			{
				// TODO: time to planned cutoff!
				// 06:50:20
				// 06:58:01
				HUDTextOut_CLR(skp, 10, lineY[9], GOLD, "Commanded dV: %10.2f m/s Ullage burn: %.1f s Burn time: %.1f s ", commanded_dV, UllageBurnPeriod_s, estimatedBurnTime_s);
			};

			char solarStatus[5];
			solarStatus[4] = '\0';
			solarStatus[0] = '1';
			switch (AnimationStatus[0])
			{
			case simple_anim_status_t::CLOSED: solarStatus[1] = 'X'; break;
			case simple_anim_status_t::CLOSING: solarStatus[1] = 'x'; break;
			case simple_anim_status_t::OPEN: solarStatus[1] = 'O'; break;
			case simple_anim_status_t::OPENING: solarStatus[1] = 'o'; break;
			default: solarStatus[1] = '?'; break;
			};
			solarStatus[2] = '2';
			switch (AnimationStatus[1])
			{
			case simple_anim_status_t::CLOSED: solarStatus[3] = 'X'; break;
			case simple_anim_status_t::CLOSING: solarStatus[3] = 'x'; break;
			case simple_anim_status_t::OPEN: solarStatus[3] = 'O'; break;
			case simple_anim_status_t::OPENING: solarStatus[3] = 'o'; break;
			default: solarStatus[3] = '?'; break;
			};



			HUDTextOut_CLR(skp, 10, lineY[10], GOLD, "Solar array: %s %.1fW Battery: %.1f Wh Beta: %g Eclipse: ~%d min", solarStatus, SolarArrayPowerOutput, BatteryCharge, SolarBeta*DEG, (int)(EclipseLength/60.0));
			HUDTextOut_CLR(skp, 10, lineY[11], GOLD, "Shear at dock: [TBI] Vibration spectrum: [TBI]");
			HUDTextOut_CLR(skp, 10, lineY[12], GOLD, "FWD gamma: [TBI] beta: [TBI] neutron flux: [TBI]");

			if (!CommandList.empty())
			{
				command_t topCommand = CommandList.top();
				int numCommands = CommandList.size();
				HUDTextOut_CLR(skp, 10, lineY[17], GREY, "Commands: %d Next command at %f: %s", numCommands, topCommand.first, const_cast<char*>(topCommand.second.c_str()));
			};

			return true;
		};
	case DAP_CONFIG:
		{
			HUDTextOut_CLR(skp, 10, lineY[0], GOLD, "DAP CONFIG [Ctrl-Y]-switch to POWER CONTROL [Alt-C]-command entry");
			HUDTextOut_CLR(skp, 10, lineY[1], GOLD, "Sample command to change gain: TVC Pitch Prop 100.0");
			HUDTextOut_CLR(skp, 10, lineY[2], GREEN, "STEADY RATE: %2.3f/s SLOW RATE: %2.3f/s", DAP_SteadyRate, DAP_SlowRate);

			HUDTextOut_CLR(skp, 10, lineY[3], GOLD, "Controller Prop  Int  Deriv MinOut MaxOut");
			for (int i = 0; i<NUM_CONTROLLERS; i++)
			{
				HUDTextOut_CLR(skp, 10, lineY[i+4], GOLD, "%-9s %-+4.1f %-+2.2f %-+2.2f %-+2.3f %-+2.3f", Controllers[i].name, Controllers[i].proportionalGain, Controllers[i].integralGain, Controllers[i].derivativeGain, Controllers[i].minOutput, Controllers[i].maxOutput);
			};
			HUDTextOut_CLR(skp, 10, lineY[NUM_CONTROLLERS+4], GREEN, " ATT DB PITCH:%.2f   YAW:%.2f   ROLL:%.2f", Att_Deadbands.data[AXIS::PITCH], Att_Deadbands.data[AXIS::YAW], Att_Deadbands.data[AXIS::ROLL]);
			HUDTextOut_CLR(skp, 10, lineY[NUM_CONTROLLERS+5], GREEN, "RATE DB PITCH:%.4f YAW:%.4f ROLL:%.4f", Rate_Deadbands.data[AXIS::PITCH], Rate_Deadbands.data[AXIS::YAW], Rate_Deadbands.data[AXIS::ROLL]);

			if (stack_RCS.size()>0)
			{
				if (AutoRCSEnabled)
				{
					if (ShallRecomputeRCSStrings)
					{
						HUDTextOut_CLR(skp, 10, lineY[NUM_CONTROLLERS+6], 
							GOLD, 
							"RCS strings enabled RECOMPUTE RCS AUTO [level: %.1f] To disable Alt-C AUTO RCS INHIBIT", 
							AutoRCSTorqueOrThrust);
					}
					else
					{
						HUDTextOut_CLR(skp, 10, lineY[NUM_CONTROLLERS+6], 
							GOLD, 
							"RCS strings enabled RECOMPUTE RCS MANUAL [level: %.1f] To disable Alt-C AUTO RCS INHIBIT", 
							AutoRCSTorqueOrThrust);
					};
				}
				else
				{
					if (ShallRecomputeRCSStrings)
					{
						HUDTextOut_CLR(skp, 10, lineY[NUM_CONTROLLERS+6], GOLD, "RCS strings disabled RECOMPUTE RCS AUTO [level: %.1f] To enable Alt-C AUTO RCS ENABLE", AutoRCSTorqueOrThrust);
					}
					else
					{
						HUDTextOut_CLR(skp, 10, lineY[NUM_CONTROLLERS+6], GOLD, "RCS strings disabled RECOMPUTE RCS MANUAL [level: %.1f] To enable Alt-C AUTO RCS ENABLE", AutoRCSTorqueOrThrust);
					};
				};
				HUDTextOut_CLR(skp, 10, lineY[NUM_CONTROLLERS+7], GOLD, "To set Alt-C AUTO RCS SET [string]; to change level Alt-C AUTO RCS LEVEL [number]");
				char lineout[255];
				int count = 255;
				int i = 0;
				for (stack_RCS_config_t::iterator citer = stack_RCS.begin(); citer!=stack_RCS.end(); citer++)
				{
					//oapiWriteLog("1_");
					RCS2string_s(citer->first, citer->second, lineout, count);
					HUDTextOut_CLR(skp, 10, lineY[NUM_CONTROLLERS+8+i], GOLD, lineout);
					i++;
				};
			};
			return true;
		};
	case POWER_CONTROL:
		{
			HUDTextOut_CLR(skp, 10, lineY[0], GOLD, "POWER CONTROL [Ctrl-Y]-switch to ENGINE CONTROL [Alt-C]-command entry");
			const double SecondsInAYear=(365.25*86400.0);
			double yrs = floor(MissionTime_s/SecondsInAYear);
			double days = floor((MissionTime_s-yrs*SecondsInAYear)/86400.0);
			HUDTextOut_CLR(skp, 10, lineY[1], GOLD, "Mission time: %d years %d days", (int) yrs, (int) days);
			HUDTextOut_CLR(skp, 10, lineY[2], GOLD, "Solar array output: %.1f W", SolarArrayPowerOutput);
			HUDTextOut_CLR(skp, 10, lineY[3], GOLD, "Battery charge: %.1f W*h Status: %d DISCHARGE THRESHOLD: %d W # cycles: %d", 
				BatteryCharge, 
				BatteryStatus, 
				(int)BatteryActivationThreshold, 
				NumBatteryCycles);
			HUDTextOut_CLR(skp, 10, lineY[4], GOLD, "Net power output: %.1f W", NetPowerOutput);

			HUDTextOut_CLR(skp, 10, lineY[6], GREEN, "Total heat load: %.1f W", TotalHeatLoad);
			HUDTextOut_CLR(skp, 10, lineY[7], GREEN, "External surface T: %d K", (int)T_ExternalSurface);
			HUDTextOut_CLR(skp, 10, lineY[8], GREEN, "LH2 tank T: %.1f K", T_LH2_Tank);
			if (SystemHealthMap[NTR_Component_t::CRYOCOOLER])
			{
				HUDTextOut_CLR(skp, 10, lineY[9], GREEN, "Cryocooler power: %d W", (int)CryocoolerPowerInput);
			}
			else
			{
				HUDTextOut_CLR(skp, 10, lineY[9], RED, "Cryocooler OFF");
			};
			HUDTextOut_CLR(skp, 10, lineY[10], GREEN, "Boiloff at current power: %.2f kg/day", Boiloff_per_day);
			double UllageVolumePercent = 100.0*(LH2_tank_V-GetPropellantMass(hpr_LH2)/LH2_density(T_LH2_Tank))/LH2_tank_V;
			HUDTextOut_CLR(skp, 10, lineY[11], GREEN, "Ullage volume: %d%%", (int)UllageVolumePercent);
			// we should have no probs with heating RCS or batteries... the external temps are very comfy.

			if (SystemHealthMap[NTR_Component_t::CDH_UNIT])
			{
				HUDTextOut_CLR(skp, 10, lineY[12], GREEN, "Command & data handling unit: ON");
			}
			else
			{
				HUDTextOut_CLR(skp, 10, lineY[12], RED,   "Command & data handling unit: OFF");
			};

			char ds_s[256];
			char line_s[1024];
			line_s[0] = '\0';
			for (unsigned int k = 0; k<drawdownSequence.size(); k++)
			{
				sprintf_s(ds_s, 255, " %s", drawdownSequence[k].pVessel->GetName());
				strcat_s(line_s, 1023, ds_s);
			};

			HUDTextOut_CLR(skp, 10, lineY[13], GREEN, "DRAWDOWN SEQUENCE: %s", line_s);
			return true;
		}
		//comms and telemetry
		// docking
	default: return true;
	};

	return true;

}

// suntracker from MBR-ITS:
//void p3p4::SolpSun (void) {				// auto sun tracking
//
//	double adj = oapiGetTimeAcceleration(); // dead-band adjustment factor
//
//
//	double solp_angle;
//	double Sun_angle;
//	char cbuf[7]= "Sun";
//	OBJHANDLE hRef = oapiGetObjectByName(cbuf);
//	VECTOR3 rsunloc;
//	VECTOR3 sunglobal;
//	oapiGetGlobalPos (hRef, &sunglobal);	// get the global pos of Sun
//	Global2Local(sunglobal,rsunloc);		// convert to vessel frame coords
//
//	/////////////// z axis rotation ////////////
//
//	solp_angle = (PI*((P4rot_proc*2)+ 0.0));	// current SARJ rotation angle
//	Sun_angle = atan2(-rsunloc.y, -rsunloc.x);	// current angle of sun around SARJ rotation axis(- adjusted 180deg)
//
//	if (Sun_angle < 0) Sun_angle = Sun_angle + (2*PI);		// make sure angle is always between
//	if (Sun_angle > (2*PI)) Sun_angle = Sun_angle - (2*PI);	// 0 to 360 deg
//
//
//	{
//	if ((solp_angle - Sun_angle) > -0.001*adj && (solp_angle - Sun_angle) < 0.001*adj) // deadband
//	{
//		P4_rot_sw = 0;					// turn off manual rot
//		P4rot_proc = Sun_angle/(2*PI);	// lock on to sun
//		SetAnimation (P4rot_anim, P4rot_proc);	// apply to animation
//	}
//	else if ((solp_angle - Sun_angle) > 0) // logic for deciding which way to rotate to find sun
//	{
//		if (solp_angle - Sun_angle < PI) P4_rot_sw = 2;
//		if (solp_angle - Sun_angle > PI) P4_rot_sw = 1;
//	}
//
//	else if ((solp_angle - Sun_angle) < 0)
//	{
//		if (solp_angle - Sun_angle > -PI) P4_rot_sw = 1;
//		if (solp_angle - Sun_angle < -PI) P4_rot_sw = 2;
//	}
//	}
//
//		double Sun_angleb;
//		// rotate frame of reference by amount equal to required SARJ rotation, and get x coord
//		double rotsunlocx = ((cos(Sun_angle))*rsunloc.x)+((sin(Sun_angle))*rsunloc.y);
//		Sun_angleb = atan2(rotsunlocx, rsunloc.z);  // angle of sun around BGA axis
//		// ensure angle is between 0-360
//		if (Sun_angleb < 0) Sun_angleb = Sun_angleb + (2*PI);
//		if (Sun_angleb > (2*PI)) Sun_angleb = Sun_angleb - (2*PI);
//
//		/////////////// AB y axis rotation ////////////
//
//{
//		double Solp_angleb;
//
//		// current BGA rotation angle
//		Solp_angleb = 2*PI*(ABrot_proc);
//
//
//	{
//	if ((Solp_angleb - Sun_angleb) > -0.001*adj && (Solp_angleb - Sun_angleb) < 0.001*adj) //deadband
//	{
//		rot_sw = 0;								// turn off manual rotation
//		ABrot_proc = Sun_angleb/(2*PI);			// lock on to sun
//		SetAnimation (ABrot_anim, ABrot_proc);	// apply to animation
//	}
//
//	// logic for deciding which way to turn to find sun
//	else if ((Solp_angleb - Sun_angleb) > 0)
//	{
//		if (Solp_angleb - Sun_angleb < PI) rot_sw = 2;
//		if (Solp_angleb - Sun_angleb > PI) rot_sw = 1;
//	}
//
//	else if ((Solp_angleb - Sun_angleb) < 0)
//	{
//		if (Solp_angleb - Sun_angleb > -PI) rot_sw = 1;
//		if (Solp_angleb - Sun_angleb < -PI) rot_sw = 2;
//	}
//	}
//}
//
//		/////////////// CD y axis rotation ////////////
//
//{
//		double Solp_anglec;
//
//
//		// current BGA rotation angle
//		Solp_anglec = 2*PI*(CDrot_proc);
//
//
//	{
//	if ((Solp_anglec - Sun_angleb) > -0.0005*adj && (Solp_anglec - Sun_angleb) < 0.0005*adj) //deadband
//	{
//		rot_sw2 = 0;								// turn off manual rotation
//		CDrot_proc = Sun_angleb/(2*PI);			// lock on to sun
//		SetAnimation (CDrot_anim, CDrot_proc);	// apply to animation
//	}
//
//	// logic for deciding which way to turn to find sun
//	else if ((Solp_anglec - Sun_angleb) > 0)
//	{
//		if (Solp_anglec - Sun_angleb < PI) rot_sw2 = 2;
//		if (Solp_anglec - Sun_angleb > PI) rot_sw2 = 1;
//	}
//
//	else if ((Solp_anglec - Sun_angleb) < 0)
//	{
//		if (Solp_anglec - Sun_angleb > -PI) rot_sw2 = 1;
//		if (Solp_anglec - Sun_angleb < -PI) rot_sw2 = 2;
//	}
//	}
//}
//
//}
//


// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void NTR_test::clbkSetClassCaps (FILEHANDLE cfg)
{
	THRUSTER_HANDLE th_main, th_rcs[NUM_RCS_THRUSTERS], th_group[MAX_NUM_THRUSTERS_PER_GROUP]; 

	// physical vessel parameters
	SetSize (PB_SIZE);
	SetEmptyMass (PB_EMPTYMASS);
	SetPMI (PB_PMI_FULL);
	SetCrossSections (PB_CS);
	SetRotDrag (PB_RD);
	SetTouchdownPoints (PB_TDP[0], PB_TDP[1], PB_TDP[2]);
	SetClipRadius(NTR_CLIPRADIUS);
	EnableTransponder(true);
	SetTransponderChannel(NTR_Transponder);

	// docking port definitions
	DOCKHANDLE hDock = CreateDock (PB_DOCK_POS, PB_DOCK_DIR, PB_DOCK_ROT);
	EnableIDS(hDock, true);
	SetIDSChannel(hDock, NTR_IDS_Channel);


	// airfoil definitions
	CreateAirfoil3 (LIFT_VERTICAL,   PB_COP, vlift, NULL, PB_VLIFT_C, PB_VLIFT_S, PB_VLIFT_A);
	CreateAirfoil3 (LIFT_HORIZONTAL, PB_COP, hlift, NULL, PB_HLIFT_C, PB_HLIFT_S, PB_HLIFT_A);

	// control surface animations
	//UINT anim_Laileron = CreateAnimation (0.5);
	//UINT anim_Raileron = CreateAnimation (0.5);
	//UINT anim_elevator = CreateAnimation (0.5);
	//AddAnimationComponent (anim_Laileron, 0, 1, &trans_Laileron);
	//AddAnimationComponent (anim_Raileron, 0, 1, &trans_Raileron);
	//AddAnimationComponent (anim_elevator, 0, 1, &trans_Lelevator);
	//AddAnimationComponent (anim_elevator, 0, 1, &trans_Relevator);

	// aerodynamic control surface defintions
	//CreateControlSurface (AIRCTRL_ELEVATOR, 1.5, 0.7, _V( 0,0,-2.5), AIRCTRL_AXIS_XPOS, anim_elevator);
	//CreateControlSurface (AIRCTRL_AILERON, 1.5, 0.25, _V( 1,0,-2.5), AIRCTRL_AXIS_XPOS, anim_Laileron);
	//CreateControlSurface (AIRCTRL_AILERON, 1.5, 0.25, _V(-1,0,-2.5), AIRCTRL_AXIS_XNEG, anim_Raileron);

	// propellant resources
	hpr_LH2 = CreatePropellantResource (PB_LH2MASS);
	hpr_RCS = CreatePropellantResource (PB_RCSFUELMASS);

	// separate RCS and hydrogen tanks.

	// main _dummy_ engine
	th_main = CreateThruster (_V(0,0,-4.35), _V(0,0,1), 0, hpr_LH2, 1.0);
	//th_main = CreateThruster (_V(0,0,-4.35), _V(0,0,1), PB_MAXMAINTH, hpr, PB_ISP);
	CreateThrusterGroup (&th_main, 1, THGROUP_MAIN);

	th_NTR[NTR_Alice] = CreateThruster (V_Alice, _V(0,0,1), NTR_MAXTHRUST, hpr_LH2, PB_NTR_ISP);
	th_NTR[NTR_Betty] = CreateThruster (V_Betty, _V(0,0,1), NTR_MAXTHRUST, hpr_LH2, PB_NTR_ISP);
	th_NTR[NTR_Clara] = CreateThruster (V_Clara, _V(0,0,1), NTR_MAXTHRUST, hpr_LH2, PB_NTR_ISP);
	CreateThrusterGroup (&th_NTR[0], 3, THGROUP_USER);

	//PARTICLESTREAMSPEC contrail_main = {
	//0, 5.0, 16, 200, 0.15, 1.0, 5, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
	//PARTICLESTREAMSPEC::LVL_PSQRT, 0, 2,
	//PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	//};
	PARTICLESTREAMSPEC exhaust_main = {
		0, 2.0, 20, 200, 0.05, 0.1, NTR_EXHAUST_LENGTH, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};

	SURFHANDLE tex_H2_exhaust = 0;//oapiRegisterExhaustTexture ("Exhaust_NTR");  // AWFUL!!
	//SURFHANDLE tex_H2_exhaust = oapiLoadTexture ("Exhaust_NTR", true);  // AWFUL!!
	Exhaust_A = AddExhaust (th_NTR[NTR_Alice], NTR_EXHAUST_LENGTH, 1, V_Alice, _V(0,0,-1), tex_H2_exhaust);
	Exhaust_B = AddExhaust (th_NTR[NTR_Betty], NTR_EXHAUST_LENGTH, 1, V_Betty, _V(0,0,-1), tex_H2_exhaust);
	Exhaust_C = AddExhaust (th_NTR[NTR_Clara], NTR_EXHAUST_LENGTH, 1, V_Clara, _V(0,0,-1), tex_H2_exhaust);
	//AddExhaustStream (th_NTR[NTR_Alice], _V(0,0.3,-10), &contrail_main);
	AddExhaustStream (th_NTR[NTR_Alice], V_Alice, &exhaust_main);
	AddExhaustStream (th_NTR[NTR_Betty], V_Betty, &exhaust_main);
	AddExhaustStream (th_NTR[NTR_Clara], V_Clara, &exhaust_main);

	// hover engine
	//th_hover = CreateThruster (_V(0,-1.5,0), _V(0,1,0), PB_MAXHOVERTH, hpr, PB_ISP);
	//CreateThrusterGroup (&th_hover, 1, THGROUP_HOVER);
	//AddExhaust (th_hover, 8, 1, _V(0,-1.5,1), _V(0,-1,0));
	//AddExhaust (th_hover, 8, 1, _V(0,-1.5,-1), _V(0,-1,0));

	//PARTICLESTREAMSPEC contrail_hover = {
	//	0, 5.0, 8, 200, 0.15, 1.0, 5, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
	//	PARTICLESTREAMSPEC::LVL_PSQRT, 0, 2,
	//	PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	//};
	//PARTICLESTREAMSPEC exhaust_hover = {
	//	0, 2.0, 10, 200, 0.05, 0.05, 8, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
	//	PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
	//	PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	//};

	//AddExhaustStream (th_hover, _V(0,-3, 1), &contrail_hover);
	//AddExhaustStream (th_hover, _V(0,-3,-1), &contrail_hover);
	//AddExhaustStream (th_hover, _V(0,-2, 1), &exhaust_hover);
	//AddExhaustStream (th_hover, _V(0,-2,-1), &exhaust_hover);

	// RCS engines : FIXME
	for (int i = 0; i<NUM_RCS_THRUSTERS; i++)
	{
		th_rcs[ i] = CreateThruster (V_RCS_POS[i], V_RCS_DIR[i], PB_MAXRCSTH, hpr_RCS, PB_ISP); //+Y
		AddExhaust (th_rcs[i], RCS_EXHAUST_LENGTH, 0.2, V_RCS_POS[i], -V_RCS_DIR[i]);
	};

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[2];
	th_group[2] = th_rcs[5];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[3];
	th_group[2] = th_rcs[4];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[4];
	th_group[2] = th_rcs[3];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[5];
	th_group[2] = th_rcs[2];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[4];
	th_group[2] = th_rcs[2];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_UP);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[5];
	th_group[2] = th_rcs[3];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_DOWN);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_LEFT);

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_RIGHT);

	// fix: 4 +z needed, located on aft quads
	// 2 -z on forward quads

	th_group[0] = th_rcs[12];
	th_group[1] = th_rcs[13];
	th_group[2] = th_rcs[14];
	th_group[3] = th_rcs[15];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_FORWARD);

	th_group[0] = th_rcs[16];
	th_group[1] = th_rcs[17];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BACK);

	// camera parameters
	SetCameraOffset (PB_DOCK_POS); // camshaking needed

	// associate a mesh for the visual
	MainMeshID = AddMesh ("NTR_test");
	//SetMeshVisibilityMode(meshID, MESHVIS_ALWAYS);
	SetMeshVisibilityMode(MainMeshID, MESHVIS_EXTERNAL);

	// enable only in space!
	for (int i = 0; i<NUM_BEACONS; i++)
	{
		beacons[i].shape = BEACONSHAPE_COMPACT;
		beacons[i].size = 0.2;
		beacons[i].falloff = 0.0;
		beacons[i].active = true; // in space
	};
	beacons[BEACON_PLUSX].pos = &beaconPlusXpos; //pos
	beacons[BEACON_PLUSX].col = &beaconColourGreen;
	beacons[BEACON_PLUSX].period = 0.0; //period
	beacons[BEACON_PLUSX].duration = 1.0; //duration
	beacons[BEACON_PLUSX].tofs = 0.0; //time offset

	beacons[BEACON_MINUSX].pos = &beaconMinusXpos; //pos
	beacons[BEACON_MINUSX].col = &beaconColourRed;
	beacons[BEACON_MINUSX].period = 0.0; //period
	beacons[BEACON_MINUSX].duration = 1.0; //duration
	beacons[BEACON_MINUSX].tofs = 0.0; //time offset

	beacons[BEACON_PLUSY].pos = &beaconPlusYpos; //pos
	beacons[BEACON_PLUSY].col = &beaconColourBlue;
	beacons[BEACON_PLUSY].period = 2.0; //period
	beacons[BEACON_PLUSY].duration = 1.0; //duration
	beacons[BEACON_PLUSY].tofs = 1.0; //time offset

	beacons[BEACON_MINUSY].pos = &beaconMinusYpos; //pos
	beacons[BEACON_MINUSY].col = &beaconColourBlue;
	beacons[BEACON_MINUSY].period = 2.0; //period
	beacons[BEACON_MINUSY].duration = 1.0; //duration
	beacons[BEACON_MINUSY].tofs = 0.0; //time offset

	for (int i = 0; i < NUM_BEACONS; i++)
	{
		AddBeacon(&beacons[i]);
	};

}

// ==============================================================
// Airfoil lift/drag functions
// ==============================================================

void NTR_test::vlift (VESSEL *v, double aoa, double M, double Re,
					  void *context, double *cl, double *cm, double *cd)
{
	static const double clp[] = {  // lift coefficient from -pi to pi in 10deg steps
		-0.1,-0.5,-0.4,-0.1,0,0,0,0,0,0,0,0,0,0,-0.2,-0.6,-0.6,-0.4,0.2,0.5,0.9,0.8,0.2,0,0,0,0,0,0,0,0,0,0.1,0.4,0.5,0.3,-0.1,-0.5
	};
	static const double aoa_step = 10.0*RAD;
	double a, fidx, saoa = sin(aoa);
	a = modf((aoa+PI)/aoa_step, &fidx);
	int idx = (int)(fidx+0.5);
	*cl = clp[idx]*(1.0-a) + clp[idx+1]*a;     // linear interpolation
	*cm = 0.0; //-0.03*sin(aoa-0.1);
	*cd = 0.03 + 0.4*saoa*saoa;                // profile drag
	*cd += oapiGetInducedDrag (*cl, 1.0, 0.5); // induced drag
	*cd += oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);  // wave drag
}

void NTR_test::hlift (VESSEL *v, double aoa, double M, double Re,
					  void *context, double *cl, double *cm, double *cd)
{
	static const double clp[] = {  // lift coefficient from -pi to pi in 45deg steps
		0,0.4,0,-0.4,0,0.4,0,-0.4,0,0.4
	};
	static const double aoa_step = 45.0*RAD;
	double a, fidx;
	a = modf((aoa+PI)/aoa_step, &fidx);
	int idx = (int)(fidx+0.5);
	*cl = clp[idx]*(1.0-a) + clp[idx+1]*a;     // linear interpolation
	*cm = 0.0;
	*cd = 0.03;
	*cd += oapiGetInducedDrag (*cl, 1.5, 0.6); // induced drag
	*cd += oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);  // wave drag
}

bool NTR_test::SetRCSString(char *str)
{
	if (str == NULL || stack_RCS.size() == 0)
	{
		oapiWriteLog("rs1");
		return true;
	};
	hash_map<string, unsigned int> stackNames;
	for (unsigned int i = 0; i<stackInfo.size(); i++)
	{
		stackNames[stackInfo[i].name] = i;
	};
	char seps[]   = ":";
	char * token;
	char * next_token;	
	string s;
	stack_RCS_item_t item;
	stack_RCS_string_t RCS_string;
	token = strtok_s( str, seps, &next_token);
	int tokno = 0;
	int toknoMOD3;
	int ID;
	double ratio;
	int controlType = 0; //none
	while( token != NULL )
	{
		toknoMOD3 = tokno % 3;
		if (tokno==0)
		{
			// control type
			if (_strnicmp(token, "P+", 2)==0)
			{
				controlType = ROT_PITCHUP_RCS;
			} else if (_strnicmp(token, "P-", 2)==0)
			{
				controlType = ROT_PITCHDOWN_RCS;
			} else if (_strnicmp(token, "B-", 2)==0)
			{
				controlType = ROT_BANKLEFT_RCS;
			} else if (_strnicmp(token, "B+", 2)==0)
			{
				controlType = ROT_BANKRIGHT_RCS;
			} else if (_strnicmp(token, "Y-", 2)==0)
			{
				controlType = ROT_YAWLEFT_RCS;
			} else if (_strnicmp(token, "Y+", 2)==0)
			{
				controlType = ROT_YAWRIGHT_RCS;
			} else if (_strnicmp(token, "FW", 2)==0)
			{
				controlType = TRAN_FORWARD_RCS;
			} else if (_strnicmp(token, "BK", 2)==0)
			{
				controlType = TRAN_BACK_RCS;
			} else if (_strnicmp(token, "UP", 2)==0)
			{
				controlType = TRAN_UP_RCS;
			} else if (_strnicmp(token, "DN", 2)==0)
			{
				controlType = TRAN_DOWN_RCS;
			} else if (_strnicmp(token, "LF", 2)==0)
			{
				controlType = TRAN_LEFT_RCS;
			} else if (_strnicmp(token, "RT", 2)==0)
			{
				controlType = TRAN_RIGHT_RCS;
			} else
			{
				oapiWriteLog("rs2");
				oapiWriteLog(token);
				return true;
			};
		} else if (toknoMOD3==1) // vessel name
		{
			//check if present in the current stack?
			s.assign(token);
			if (stackNames.find(s)==stackNames.end()) 
			{				
				oapiWriteLog("rs3");
				return true;
			};
			item.vesselName.assign(token);
		} else if (toknoMOD3==2) // thruster id
		{
			//check if the id is in the valid range
			ID = atoi(token);
			if (ID<0 || ID>=(int)(stackInfo[stackNames[item.vesselName]].thrusters.size())) 
			{
				oapiWriteLog("rs4");
				return true;
			};
			item.thrusterID = ID;
		} else if (toknoMOD3==0) // ratio, and we can push the thruster into the string
		{
			ratio = atof(token);
			if (ratio<0.0 || ratio >1.0)
			{
				oapiWriteLog("rs5");
				return true;
			};
			item.thrustRatio = ratio; // set index to thrusters of the stack!
			// can use inefficient linear lookup!
			for (unsigned int j = 0; j<stackThrusters.size(); j++)
			{
				if (strcmp(stackThrusters[j].pV->GetName(), item.vesselName.c_str())==0)
				{
					if (stackThrusters[j].thrusterID == item.thrusterID)
					{
						item.index = j;
						break;
					};
				};
			};
			RCS_string.push_back(item);
		};
		token = strtok_s( NULL, seps, &next_token);
		tokno++;
	};
	stack_RCS[controlType] = RCS_string;

	return true;
}

void NTR_test::LoadCommands(const char * cmdFileName)
{
	FILE* stream;
	char line[1024];
	char cmd[1024];
	float MJD;
	//CommandList = ; //?
	//GetCurrentDirectory(1023, line);
	//oapiWriteLog(const_cast<char *>(cmdFileName));
	if( (stream = _fsopen( cmdFileName, "rb", _SH_DENYNO )) != NULL )
	{
		while (!feof(stream))
		{
			if( fgets( line, 1023, stream ) == NULL)
			{
				break;
			};
			if (_strnicmp(line, ";", 1)==0) continue; // skip a bit with comments
			//separate into MJD and the rest of the line
			sscanf_s(line, "%g %[^\r]", &MJD, &cmd, 1023);
			CommandList.push(pair<double, string>((double)MJD, cmd));
			//oapiWriteLog(cmd);
			//sprintf_s(line, 1023, "%f", MJD);
			//oapiWriteLog(line);
		};
		fclose( stream );
	}
	else
	{
		oapiWriteLog("failed to open commands file" );
	};
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new NTR_test (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (NTR_test*)vessel;
}
