// various things useful for manipulating the craft in AP's.
#ifndef __AUTOPILOT_H__
#define __AUTOPILOT_H__

#define STRICT
#include "orbitersdk.h"
#include <vector>
#include <string>
#include <hash_set>
#include <queue>
#include <hash_map>
#include "jacobi.h"
#include "ap.h"
#include "optimization.h"
#include "rotations.h" //quaternions and stuff
#include <iso646.h>
#include "lp_lib.h" // linear programming.
#include "numeric_declarations.h"

using namespace std;
using namespace stdext;
using namespace alglib;
using namespace alglib_impl;
//stuck with a left-handed coord system...


#define TRAN_FORWARD_MAIN    (1)
#define TRAN_BACK_RETRO      (2)
#define TRAN_UP_HOVER        (3)
#define TRAN_FORWARD_RCS     (4)
#define TRAN_RCS_FIRST		 (TRAN_FORWARD_RCS)
#define TRAN_BACK_RCS        (5)
#define TRAN_UP_RCS          (6)
#define TRAN_DOWN_RCS        (7)
#define TRAN_LEFT_RCS        (8)
#define TRAN_RIGHT_RCS       (9)
#define TRAN_RCS_LAST		 (TRAN_RIGHT_RCS)
#define ROT_PITCHUP_RCS      (10)
#define ROT_RCS_FIRST        (ROT_PITCHUP_RCS)
#define ROT_PITCHDOWN_RCS    (11)
#define ROT_YAWLEFT_RCS      (12)
#define ROT_YAWRIGHT_RCS     (13)
#define ROT_BANKLEFT_RCS     (14)
#define ROT_BANKRIGHT_RCS    (15)
#define ROT_RCS_LAST         (ROT_BANKRIGHT_RCS)
// not compensated!
#define ROT_ELEVATOR_FCS     (16)
#define ROT_RUDDER_FCS       (17)
#define ROT_AILERON_FCS      (18)
#define ROT_FLAP_FCS         (19)
#define ROT_ELEVATORTRIM_FCS (20)
#define ROT_RUDDERTRIM_FCS   (21)

//static const char* controlStringNames[] ={"(none)      ",
//"Main engine ",
//"Retro       ",
//"Hover       ",
//"TRCS forward",
//"TRCS retro  ",
//"TRCS up     ",
//"TRCS down   ",
//"TRCS left   ",
//"TRCS right  ",
//"RRCS P up   ",
//"RRCS P down ",
//"RRCS Y left ",
//"RRCS Y right",
//"RRCS B left ",
//"RRCS B right",
//"FCS elevator",
//"FCS rudder  ",
//"FCS aileron ",
//"FCS flap    ",
//"FCS elv.trim",
//"FCS rud.trim"};

static const char* controlStringNames[] ={"(none)      ",
"Main engine ",
"Retro       ",
"Hover       ",
"FW",
"BK",
"UP",
"DN",
"LF",
"RT",
"P+",
"P-",
"Y-",
"Y+",
"B-",
"B+",
"FCS elevator",
"FCS rudder  ",
"FCS aileron ",
"FCS flap    ",
"FCS elv.trim",
"FCS rud.trim"};
// commands - rotate and translate. The main engine is simply a very powerful translation.
// stop all the thrusters!

// the design requirement is to be able to manage a stack of different vessels (e.g. ISS with Progresses, HTV, Dragons 
// and Shuttles, 
//Burch's DSV etc.) We don't assume all docks are collinear. Each vessel's fuel has constraints on it = 
// when to use and how much to use
// find the approximately equivalent thrusters and create a combined group... to offset 
// we don't allow non DAG docking configurations (e.g. cycles)


typedef struct {
	string vesselName; // will not be activated unless the vessel is part of the stack. Thus, we can define strings for various configurations.
	int thrusterID; // zero-based - for a vehicle
	int index; // a temporary index into the stack's thrusters' vector
	double thrustRatio; // 0 to 1 - defined for compensating extra torque/linear force
} stack_RCS_item_t;

typedef vector<stack_RCS_item_t> stack_RCS_string_t;
typedef hash_map<int, stack_RCS_string_t> stack_RCS_config_t;

typedef struct {
	stack_RCS_string_t * RCS_string;
	double startMJD;
	double level; //multiplying thrust etc. by the number across the board...
	double duration; // seconds
} AP_command_t;

enum rhs_t {tX, tY, tZ, fX, fY, fZ};

// there is no bloody symmetry between rcs strings due to off-COG placement of thrusters.

typedef struct {
	double mass;
	double maxmass;
	double efficiency; // 0 to 1.
	double minimum_mass; //constraint on all algorithms, may change!
	double cost; // we'll minimize weighted fuel usage in a bloody linear formula, 
	//can add quadratic or exponential terms later
} fuel_tank_t;

typedef struct {
	VESSEL* pV;
	int tankID;
	double mass;
	double maxmass;
	double efficiency; // 0 to 1.
	double minimum_mass; //constraint on all algorithms, may change!
	double cost; // we'll minimize weighted fuel usage in a bloody linear formula, 
	//can add quadratic or exponential terms later
} stack_fuel_tank_t;

// in vessel coord frame
typedef struct {
	VECTOR3 pos;
	VECTOR3 dir;
	VECTOR3 rot;
} joint_params_t;

// in vessel coord frame!
typedef struct {
	VECTOR3 pos;
	VECTOR3 dir;
	double maxThrust_vacuum;
	double maxThrust_SL;
	double Isp_vacuum;
	double Isp_SL;
	double maxFlowRate_vacuum; //kg/s - do we need this now??? sanity check!
	double maxFlowRate_SL;
	int propellantID; // internal index into the vector of propellant tank feeding the thruster!
	hash_set<int> groupID; // controlType defines go here.
} thruster_t;

typedef struct {
	VESSEL * pV;
	int thrusterID;
	VECTOR3 root_pos;
	VECTOR3 root_dir;
	double maxThrust_vacuum;
	double maxThrust_SL;
	double Isp_vacuum;
	double Isp_SL;
	double maxFlowRate_vacuum; //kg/s - do we need this now??? sanity check!
	double maxFlowRate_SL;
	int propellantID; // internal index into the vector of propellant tank feeding the thruster!
	//heck, don't need to look up anything
	double propellantCost;
	hash_set<int> groupID; // controlType defines go here, and I don't like linear search.
} stack_thruster_t;

typedef struct {
	int controlType; //from 1 to 15 - from TRAN_FORWARD_MAIN to ROT_BANKRIGHT_RCS
	stack_RCS_string_t* RCS_string;
	vector<stack_thruster_t>* thruster_array;
	VECTOR3 COG; // slapping forehead...
} RCS_string_optimization_package_t;

// torque = sum(crossp(root_pos,thrust*root_dir));

// translational: torque = 0; force = (1N;0;0)
// rotational: 

// one tank may feed several thrusters, but a thruster can have only one feeder tank (monoprop only)

enum joint_type_t {NO_CONNECTION, DOCK_CONNECTION, ATTACH_CONNECTION};

typedef struct {
	string name;
	VESSEL* pV;
	VESSEL* parentVessel;
	joint_type_t jointType;
	int parentJointParamsID;//zero-based
	int ownJointParamsID;//zero-based
	// let's collect data first we'll process it later
	joint_params_t parentJointParams;
	joint_params_t ownJointParams;
	VECTOR3 PMI;
	double mass;
	double GravityGradientDamping;
	double radius; // by GetSize, m
	vector<thruster_t> thrusters;
	vector<fuel_tank_t> fuel_tanks;
} stackVessel_t;

inline stackVessel_t FillRootVesselInfo(const OBJHANDLE hVessel)
{
	stackVessel_t sv;
	VESSEL* v_vessel = oapiGetVesselInterface(hVessel);

	if (v_vessel == NULL) 
	{
		oapiWriteLog("Invalid objhandle in FillRootVesselInfo");
		throw "Invalid objhandle in FillRootVesselInfo";
	};
	sv.pV = v_vessel;
	sv.parentVessel = NULL;
	sv.GravityGradientDamping = v_vessel->GetGravityGradientDamping();
	sv.mass = v_vessel->GetMass();
	sv.name.assign(v_vessel->GetName());
	v_vessel->GetPMI(sv.PMI);
	sv.radius = v_vessel->GetSize();
	sv.jointType = NO_CONNECTION;
	sv.parentJointParamsID = -1;
	sv.ownJointParamsID = -1;

	//oapiWriteLog("pp1");

	sv.fuel_tanks.resize(v_vessel->GetPropellantCount());

	hash_map<PROPELLANT_HANDLE, int> fuelhandles;
	for (unsigned int j = 0; j<sv.fuel_tanks.size(); j++)
	{
		PROPELLANT_HANDLE ph = v_vessel->GetPropellantHandleByIndex(j);
		fuelhandles[ph] = j;
		sv.fuel_tanks[j].efficiency = v_vessel->GetPropellantEfficiency(ph); 
		//if (sv.fuel_tanks[j].efficiency<=0.0) oapiWriteLog("negeff");
		sv.fuel_tanks[j].cost = 1.0;//for now
		sv.fuel_tanks[j].mass = v_vessel->GetPropellantMass(ph); 
		sv.fuel_tanks[j].maxmass = v_vessel->GetPropellantMaxMass(ph); 
		sv.fuel_tanks[j].minimum_mass = 0.0; // use all fuel available!
	};

	//oapiWriteLog("pp2");

	// lookup table - thruster -> type;
	hash_multimap<THRUSTER_HANDLE, int> group_assignments;
	typedef hash_multimap<THRUSTER_HANDLE, int>::const_iterator th_pair_citerator_t;
	typedef pair<THRUSTER_HANDLE, int> th_pair_t;

	//oapiWriteLog("tx1");

	for (unsigned int ii=THGROUP_MAIN;ii<=THGROUP_ATT_BACK;ii++)
	{
		//oapiWriteLog("tkk");
		THGROUP_HANDLE thg = v_vessel->GetThrusterGroupHandle((THGROUP_TYPE)ii);
		if (thg==NULL) continue; // silly me.
		for (unsigned int jj = 0; jj<v_vessel->GetGroupThrusterCount(thg);jj++)
		{
			THRUSTER_HANDLE thv = v_vessel->GetGroupThruster(thg, jj);
			switch((THGROUP_TYPE)ii)
			{
			case THGROUP_MAIN:
				group_assignments.insert(th_pair_t(thv, TRAN_FORWARD_MAIN));
				break;
			case THGROUP_RETRO:
				group_assignments.insert(th_pair_t(thv, TRAN_BACK_RETRO));
				break;
			case THGROUP_HOVER:
				group_assignments.insert(th_pair_t(thv, TRAN_UP_HOVER));
				break;
			case THGROUP_ATT_PITCHUP:
				group_assignments.insert(th_pair_t(thv, ROT_PITCHUP_RCS));
				break;
			case THGROUP_ATT_PITCHDOWN:
				group_assignments.insert(th_pair_t(thv, ROT_PITCHDOWN_RCS));
				break;
			case THGROUP_ATT_YAWLEFT:
				group_assignments.insert(th_pair_t(thv, ROT_YAWLEFT_RCS));
				break;
			case THGROUP_ATT_YAWRIGHT:
				group_assignments.insert(th_pair_t(thv, ROT_YAWRIGHT_RCS));
				break;
			case THGROUP_ATT_BANKLEFT:
				group_assignments.insert(th_pair_t(thv, ROT_BANKLEFT_RCS));
				break;
			case THGROUP_ATT_BANKRIGHT:
				group_assignments.insert(th_pair_t(thv, ROT_BANKRIGHT_RCS));
				break;
			case THGROUP_ATT_RIGHT:
				group_assignments.insert(th_pair_t(thv, TRAN_RIGHT_RCS));
				break;
			case THGROUP_ATT_LEFT:
				group_assignments.insert(th_pair_t(thv, TRAN_LEFT_RCS));
				break;
			case THGROUP_ATT_UP:
				group_assignments.insert(th_pair_t(thv, TRAN_UP_RCS));
				break;
			case THGROUP_ATT_DOWN:
				group_assignments.insert(th_pair_t(thv, TRAN_DOWN_RCS));
				break;
			case THGROUP_ATT_FORWARD:
				group_assignments.insert(th_pair_t(thv, TRAN_FORWARD_RCS));
				break;
			case THGROUP_ATT_BACK:
				group_assignments.insert(th_pair_t(thv, TRAN_BACK_RCS));
				break;
			default:
				break;
				// doing nothing - it ain't of concern to us yet
				//oapiWriteLog("fallthrough here");
				//throw("fallthrough here");
			};
		};
	};
	//oapiWriteLog("tx2zz");

	sv.thrusters.resize(v_vessel->GetThrusterCount());
	for (unsigned int k = 0; k<sv.thrusters.size(); k++)
	{
		THRUSTER_HANDLE th = v_vessel->GetThrusterHandleByIndex(k);
		sv.thrusters[k].groupID.clear();
		th_pair_citerator_t	citer;
		for (citer = group_assignments.lower_bound(th); citer!=group_assignments.end() && citer!=group_assignments.upper_bound(th); citer++)
		{
			sv.thrusters[k].groupID.insert(citer->second);
		};
		PROPELLANT_HANDLE ph_r = v_vessel->GetThrusterResource(th);
		sv.thrusters[k].propellantID = (ph_r == NULL ? -1 : fuelhandles.find(ph_r)->second); 
		v_vessel->GetThrusterDir(th, sv.thrusters[k].dir);
		v_vessel->GetThrusterRef(th, sv.thrusters[k].pos);
		sv.thrusters[k].Isp_SL = v_vessel->GetThrusterIsp(th, ATMP);
		sv.thrusters[k].Isp_vacuum = v_vessel->GetThrusterIsp0(th);
		sv.thrusters[k].maxThrust_SL = v_vessel->GetThrusterMax(th, ATMP);
		sv.thrusters[k].maxThrust_vacuum = v_vessel->GetThrusterMax0(th);
		if (sv.thrusters[k].propellantID==-1)
		{
			sv.thrusters[k].maxFlowRate_SL = 0.0;
			sv.thrusters[k].maxFlowRate_vacuum = 0.0;
		}
		else
		{
			double eff = sv.fuel_tanks[sv.thrusters[k].propellantID].efficiency;
			sv.thrusters[k].maxFlowRate_SL = (sv.thrusters[k].Isp_SL>0.0 ? sv.thrusters[k].maxThrust_SL/(sv.thrusters[k].Isp_SL*eff): 0.0);
			sv.thrusters[k].maxFlowRate_vacuum = (sv.thrusters[k].Isp_vacuum>0.0 ? sv.thrusters[k].maxThrust_vacuum/(sv.thrusters[k].Isp_vacuum*eff): 0.0);
		};
		//char xl[256];
		//sprintf_s(xl, 255, "%.4f %.4f %.4f", eff, sv.thrusters[k].Isp_vacuum, sv.thrusters[k].Isp_SL);
		//oapiWriteLog(xl);
	};

	//oapiWriteLog("pp4");

	return sv;
}

inline stackVessel_t FillDockedVesselInfo(const OBJHANDLE hParent, const int parentDockID)
{
	if (hParent == NULL) 
	{
		oapiWriteLog("Invalid objhandle in FillVesselInfo");
		throw "Invalid objhandle in FillVesselInfo";
	};
	stackVessel_t sv;
	VESSEL* parentVessel = oapiGetVesselInterface(hParent);
	if (parentVessel == NULL) 
	{
		oapiWriteLog("Invalid objhandle in FillVesselInfo");
		throw "Invalid objhandle in FillVesselInfo";
	};
	DOCKHANDLE dh_parent = parentVessel->GetDockHandle(parentDockID);
	if (dh_parent == NULL)
	{
		oapiWriteLog("Invalid parent dock id in FillVesselInfo");
		throw "Invalid parent dock id in FillVesselInfo";
	};
	OBJHANDLE hChild = parentVessel->GetDockStatus(dh_parent);
	if (hChild == NULL)
	{
		oapiWriteLog("Nothing docked to parent in FillVesselInfo");
		throw "Nothing docked to parent in FillVesselInfo";
	};
	VESSEL* childVessel = oapiGetVesselInterface(hChild);
	if (childVessel == NULL)
	{
		oapiWriteLog("Weird things in FillVesselInfo");
		throw "Weird things in FillVesselInfo";
	};
	sv = FillRootVesselInfo(hChild);
	sv.parentVessel = parentVessel;

	sv.jointType = DOCK_CONNECTION;
	sv.parentJointParamsID = parentDockID;
	parentVessel->GetDockParams(dh_parent, sv.parentJointParams.pos, sv.parentJointParams.dir, sv.parentJointParams.rot);
	sv.ownJointParamsID = -1;
	DOCKHANDLE dh_child;
	for (unsigned int i = 0; i<childVessel->DockCount(); i++)
	{
		dh_child = childVessel->GetDockHandle(i);
		if (childVessel->GetDockStatus(dh_child) == hParent)
		{
			sv.ownJointParamsID = i;
			break;
		};
	};
	if (sv.ownJointParamsID == -1)
	{
		sv.pV = NULL;
		//oapiWriteLog("Strange docking anomaly in FillVesselInfo");
		//throw "Strange docking anomaly in FillVesselInfo";
		return sv;
	};
	childVessel->GetDockParams(dh_child, sv.ownJointParams.pos, sv.ownJointParams.dir, sv.ownJointParams.rot);
	return sv;
}

inline stackVessel_t FillAttachedVesselInfo(const OBJHANDLE hParent, const bool toParent, const int parentAttachmentID)
{
	stackVessel_t sv;

	VESSEL* parentVessel = oapiGetVesselInterface(hParent);
	if (parentVessel == NULL) 
	{
		oapiWriteLog("Invalid objhandle in FillAttachedVesselInfo");
		throw "Invalid objhandle in FillAttachedVesselInfo";
	};

	ATTACHMENTHANDLE ah_parent = parentVessel->GetAttachmentHandle(toParent, parentAttachmentID);
	if (ah_parent == NULL)
	{
		oapiWriteLog("Invalid attach id in FillAttachedVesselInfo");
		throw "Invalid attach id in FillAttachedInfo";
	};

	OBJHANDLE hChild = parentVessel->GetAttachmentStatus(ah_parent);
	if (hChild == NULL)
	{
		oapiWriteLog("Nothing attached to parent in FillAttachedVesselInfo");
		throw "Nothing attached to parent in FillAttachedVesselInfo";
	};

	VESSEL* childVessel = oapiGetVesselInterface(hChild);
	if (childVessel == NULL)
	{
		oapiWriteLog("Weird things in FillAttachedVesselInfo");
		throw "Weird things in FillAttachedVesselInfo";
	};

	sv = FillRootVesselInfo(hChild);
	sv.parentVessel = parentVessel;
	sv.jointType = ATTACH_CONNECTION;
	sv.parentJointParamsID = parentAttachmentID;

	parentVessel->GetAttachmentParams(ah_parent, sv.parentJointParams.pos, sv.parentJointParams.dir, sv.parentJointParams.rot);

	sv.ownJointParamsID = -1;
	ATTACHMENTHANDLE ah_child;
	for (unsigned int i = 0; i<childVessel->AttachmentCount(not toParent); i++)
	{
		ah_child = childVessel->GetAttachmentHandle(not toParent, i);
		if (childVessel->GetAttachmentStatus(ah_child) == hParent)
		{
			sv.ownJointParamsID = i;
			break;
		};
	};

	if (sv.ownJointParamsID == -1)
	{
		oapiWriteLog("Strange attachment anomaly in FillAttachedVesselInfo");
		throw "Strange docking anomaly in FillAttachedVesselInfo";
	};
	childVessel->GetAttachmentParams(ah_child, sv.ownJointParams.pos, sv.ownJointParams.dir, sv.ownJointParams.rot);

	return sv;
}


inline void FillStackInfo(const VESSEL* refVessel, vector<stackVessel_t>& stackInfo)
{
	hash_set<OBJHANDLE> visitedVessels;
	stackInfo.clear();
	if (refVessel==NULL) return;
	deque<VESSEL*> vQueue;
	vQueue.push_back(const_cast<VESSEL*>(refVessel));
	VESSEL * currVessel;
	OBJHANDLE currHandle; 
	bool isRoot  = true;

	while (not vQueue.empty())
	{
		currVessel = vQueue.front();
		currHandle = currVessel->GetHandle();
		visitedVessels.insert(currHandle);
		//oapiWriteLog("qq1");

		if (isRoot)
		{
			stackInfo.push_back(FillRootVesselInfo(currHandle));
			isRoot = false;

		};
		//oapiWriteLog("qq2");

		int dock_count = currVessel->DockCount();
		for (int dock_ind = 0; dock_ind<dock_count; dock_ind++)
		{
			DOCKHANDLE dh = currVessel->GetDockHandle(dock_ind);
			if (dh==NULL) continue; //silly
			OBJHANDLE neighbour = currVessel->GetDockStatus(dh);
			if (neighbour==NULL) continue;
			if (visitedVessels.find(neighbour)!=visitedVessels.end()) continue;
			stackVessel_t sv = FillDockedVesselInfo(currHandle, dock_ind);
			if (sv.pV==NULL) continue;
			stackInfo.push_back(sv);
			vQueue.push_back(sv.pV);
		};
		//oapiWriteLog("qq3");

		//// now with attachments
		int attach_count;

		attach_count = currVessel->AttachmentCount(true);
		for (int attach_ind = 0; attach_ind<attach_count; attach_ind++)
		{
			ATTACHMENTHANDLE ah = currVessel->GetAttachmentHandle(true, attach_ind);
			if (ah==NULL) continue;
			OBJHANDLE neighbour = currVessel->GetAttachmentStatus(ah);
			if (neighbour==NULL) continue;
			VESSEL * nV = oapiGetVesselInterface(neighbour);
			if (nV==NULL) continue;
			if (visitedVessels.find(neighbour)!=visitedVessels.end()) continue;
			stackVessel_t sv = FillAttachedVesselInfo(currHandle, true, attach_ind);
			stackInfo.push_back(sv);
			vQueue.push_back(sv.pV);
		};
		//oapiWriteLog("qq4");

		attach_count = currVessel->AttachmentCount(false);
		for (int attach_ind = 0; attach_ind<attach_count; attach_ind++)
		{
			ATTACHMENTHANDLE ah = currVessel->GetAttachmentHandle(false, attach_ind);
			if (ah==NULL) continue;
			OBJHANDLE neighbour = currVessel->GetAttachmentStatus(ah);
			if (neighbour==NULL) continue;
			VESSEL * nV = oapiGetVesselInterface(neighbour);
			if (nV==NULL) continue;
			if (visitedVessels.find(neighbour)!=visitedVessels.end()) continue;
			stackVessel_t sv = FillAttachedVesselInfo(currHandle, false, attach_ind);
			stackInfo.push_back(sv);
			vQueue.push_back(sv.pV);
		};
		//oapiWriteLog("qq5");

		// remove
		vQueue.pop_front();
	};

	// tail!	
}

inline void TransformStackThrusterInfoIntoRootFrame(const vector<stackVessel_t>& stackInfo, 
													vector<stack_thruster_t>& thrusters, 
													vector<stack_fuel_tank_t>& tanks, 
													VECTOR3& stack_COG, 
													VECTOR3& stack_PMI) // could be the tensor with non-zero off-diag entries
{
	// the path of least resistance

	thrusters.clear();
	tanks.clear(); // solve it later.

	unsigned int stackSize = stackInfo.size();
	if (stackSize==0) return;

	VESSEL* rootVessel = stackInfo[0].pV;
	MATRIX3 rootRot;
	rootVessel->GetRotationMatrix(rootRot);
	stack_COG = _V(0,0,0);
	double stackMass = 0.0;
	hash_map<int, int> tankMap;

	for (unsigned int i = 0; i<stackSize; i++)
	{
		stackMass += stackInfo[i].pV->GetMass();
	};

	// find the ones with the largest torque? and minimum fuel cost.

	unsigned int tankID = 0;

	for (unsigned int i = 0; i<stackSize; i++)
	{
		//let's do the redundant code:
		VESSEL* xVessel = stackInfo[i].pV;
		VECTOR3 global_COG, xloc_COG;
		xVessel->GetGlobalPos(global_COG);
		rootVessel->Global2Local(global_COG, xloc_COG);
		stack_COG = stack_COG + (xVessel->GetMass()/stackMass)*xloc_COG;

		tankMap.clear();

		for (unsigned int j = 0; j<stackInfo[i].fuel_tanks.size();j++)
		{
			stack_fuel_tank_t sft;
			sft.cost = stackInfo[i].fuel_tanks[j].cost;
			sft.efficiency = stackInfo[i].fuel_tanks[j].efficiency;
			sft.mass = stackInfo[i].fuel_tanks[j].mass;
			sft.maxmass = stackInfo[i].fuel_tanks[j].maxmass;
			sft.minimum_mass = stackInfo[i].fuel_tanks[j].minimum_mass;
			sft.pV = xVessel;
			sft.tankID = j;
			tanks.push_back(sft);
			tankMap[j] = tankID;
			tankID++;
		};

		for (unsigned int j = 0; j<stackInfo[i].thrusters.size();j++)
		{
			stack_thruster_t st;
			st.pV=xVessel;
			if (stackInfo[i].thrusters[j].groupID.size()>0)
				st.groupID = hash_set<int>(stackInfo[i].thrusters[j].groupID.begin(), stackInfo[i].thrusters[j].groupID.end());
			VECTOR3 global_pos;
			xVessel->Local2Global(stackInfo[i].thrusters[j].pos, global_pos);
			rootVessel->Global2Local(global_pos, st.root_pos);
			MATRIX3 xRot;
			xVessel->GetRotationMatrix(xRot);
			VECTOR3 global_dir = mul(xRot, stackInfo[i].thrusters[j].dir);
			st.root_dir = tmul(rootRot, global_dir);
			st.Isp_SL = stackInfo[i].thrusters[j].Isp_SL;
			st.Isp_vacuum = stackInfo[i].thrusters[j].Isp_vacuum;
			st.maxFlowRate_SL = stackInfo[i].thrusters[j].maxFlowRate_SL;
			st.maxFlowRate_vacuum = stackInfo[i].thrusters[j].maxFlowRate_vacuum;
			st.maxThrust_SL = stackInfo[i].thrusters[j].maxThrust_SL;
			st.maxThrust_vacuum = stackInfo[i].thrusters[j].maxThrust_vacuum;
			st.thrusterID = j;
			st.propellantID = tankMap[stackInfo[i].thrusters[j].propellantID];
			st.propellantCost = tanks[st.propellantID].cost; // no idea how it works out.
			thrusters.push_back(st);
		};
	};
	// I'll try swapping items... x and y perhaps...

	// deduce vessel's rotation matrix from the root. Then rotate the thruster direction back to root.
	// the offset should be calculated as well, we need the thruster position in the root coords.
}

inline VECTOR3 ComputeRCSTorque(const real_1d_array& thrustRatio_and_LMs, RCS_string_optimization_package_t* Params)
{
	unsigned int numActiveThrusters = Params->RCS_string->size();
	VECTOR3 torque = _V(0,0,0);
	for (unsigned int i = 0; i<numActiveThrusters; i++)
	{
		stack_thruster_t& st = Params->thruster_array->at(Params->RCS_string->at(i).index);
		torque = torque+crossp(st.root_pos-Params->COG, st.maxThrust_vacuum*thrustRatio_and_LMs[i]*st.root_dir);
	};
	return torque;
}

inline VECTOR3 ComputeRCSTorque(vector<stack_thruster_t>* thruster_array, stack_RCS_string_t* RCS_string, VECTOR3 COG)
{
	unsigned int numActiveThrusters = RCS_string->size();
	VECTOR3 torque = _V(0,0,0);
	for (unsigned int i = 0; i<numActiveThrusters; i++)
	{
		stack_thruster_t& st = thruster_array->at(RCS_string->at(i).index);
		torque = torque+crossp(st.root_pos-COG, st.maxThrust_vacuum*RCS_string->at(i).thrustRatio*st.root_dir);
	};
	return torque;
}

inline void ComputeTorqueCoefficientArray(vector<stack_thruster_t>* thruster_array, 
										  stack_RCS_string_t* RCS_string, 
										  VECTOR3 COG, vector<VECTOR3>& torquearray)
{
	unsigned int numActiveThrusters = RCS_string->size();
	torquearray.clear();
	torquearray.resize(numActiveThrusters);
	for (unsigned int i = 0; i<numActiveThrusters; i++)
	{
		stack_thruster_t& st = thruster_array->at(RCS_string->at(i).index);
		VECTOR3 rp = st.root_pos-COG;
		VECTOR3 f = st.maxThrust_vacuum*st.root_dir; //??
		torquearray.at(i) = crossp(rp, f);
	};
}

inline void ComputeForceCoefficientArray(vector<stack_thruster_t>* thruster_array, 
										 stack_RCS_string_t* RCS_string, 
										 VECTOR3 COG, vector<VECTOR3>& forcearray)
{
	unsigned int numActiveThrusters = RCS_string->size();
	forcearray.clear();
	forcearray.resize(numActiveThrusters);
	for (unsigned int i = 0; i<numActiveThrusters; i++)
	{
		stack_thruster_t& st = thruster_array->at(RCS_string->at(i).index);
		forcearray.at(i) = st.maxThrust_vacuum*st.root_dir;
	};
}

inline VECTOR3 ComputeRCSForce(const real_1d_array& thrustRatio_and_LMs, RCS_string_optimization_package_t* Params)
{
	unsigned int numActiveThrusters = Params->RCS_string->size();
	VECTOR3 force = _V(0,0,0);
	for (unsigned int i = 0; i<numActiveThrusters; i++)
	{
		stack_thruster_t& st = Params->thruster_array->at(Params->RCS_string->at(i).index);
		force = force+st.maxThrust_vacuum*thrustRatio_and_LMs[i]*st.root_dir;
	};
	return force;
}

inline VECTOR3 ComputeRCSForce(vector<stack_thruster_t>* thruster_array, stack_RCS_string_t* RCS_string)
{
	unsigned int numActiveThrusters = RCS_string->size();
	VECTOR3 force = _V(0,0,0);
	for (unsigned int i = 0; i<numActiveThrusters; i++)
	{
		stack_thruster_t& st = thruster_array->at(RCS_string->at(i).index);
		force = force+st.maxThrust_vacuum*RCS_string->at(i).thrustRatio*st.root_dir;
	};
	return force;
}


inline double ComputeRCSFuelCost(const real_1d_array& thrustRatio_and_LMs, RCS_string_optimization_package_t* Params)
{
	unsigned int numActiveThrusters = Params->RCS_string->size();
	double fc = 0.0;
	for (unsigned int i = 0; i<numActiveThrusters; i++)
	{
		stack_thruster_t& st = Params->thruster_array->at(Params->RCS_string->at(i).index);
		fc = fc+st.maxFlowRate_vacuum*thrustRatio_and_LMs[i]*st.propellantCost;
	};
	return fc;
}

// per second burn!
inline double ComputeRCSFuelCost(vector<stack_thruster_t>* thruster_array, stack_RCS_string_t* RCS_string)
{
	unsigned int numActiveThrusters = RCS_string->size();
	double fc = 0.0;
	for (unsigned int i = 0; i<numActiveThrusters; i++)
	{
		stack_thruster_t& st = thruster_array->at(RCS_string->at(i).index);
		fc = fc+st.maxFlowRate_vacuum*RCS_string->at(i).thrustRatio*st.propellantCost;
	};
	return fc;
}

inline void ComputeFuelCostArray(vector<stack_thruster_t>* thruster_array, 
								 stack_RCS_string_t* RCS_string, vector<double>& fuelCostArray)
{
	unsigned int numActiveThrusters = RCS_string->size();
	fuelCostArray.clear();
	fuelCostArray.resize(numActiveThrusters);
	for (unsigned int i = 0; i<numActiveThrusters; i++)
	{
		stack_thruster_t& st = thruster_array->at(RCS_string->at(i).index);
		fuelCostArray.at(i)= st.maxFlowRate_vacuum*st.propellantCost;
	};
}


inline double ComputeRCSFuelExpenditure(const real_1d_array& thrustRatio_and_LMs, void* Params)
{
	RCS_string_optimization_package_t* info = (RCS_string_optimization_package_t*) Params;
	VECTOR3 Torque, Force; // use the Force, Luke!
	// 6 Lagrange multipliers.
	unsigned int numActiveThrusters = info->RCS_string->size();

	double L = ComputeRCSFuelCost(thrustRatio_and_LMs, info); //lagrangian

	Torque = ComputeRCSTorque(thrustRatio_and_LMs, info);
	Force = ComputeRCSForce(thrustRatio_and_LMs, info);
	double tX = thrustRatio_and_LMs[numActiveThrusters];
	double tY = thrustRatio_and_LMs[numActiveThrusters+1];
	double tZ = thrustRatio_and_LMs[numActiveThrusters+2];
	double fX = thrustRatio_and_LMs[numActiveThrusters+3];
	double fY = thrustRatio_and_LMs[numActiveThrusters+4];
	double fZ = thrustRatio_and_LMs[numActiveThrusters+5];
	switch (info->controlType)
	{
	case TRAN_FORWARD_MAIN:
		//z+
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z - 0.0) + fX*(Force.x - 0.0) + fY*(Force.y - 0.0) + fZ*(Force.z - 1.0);
		break;
	case TRAN_BACK_RETRO:
		//z-
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z - 0.0) + fX*(Force.x - 0.0) + fY*(Force.y - 0.0) + fZ*(Force.z + 1.0);
		break;
	case TRAN_UP_HOVER:
		//y+
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z - 0.0) + fX*(Force.x - 0.0) + fY*(Force.y - 1.0) + fZ*(Force.z - 0.0);
		break;
	case TRAN_FORWARD_RCS:
		//z+
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z - 0.0) + fX*(Force.x - 0.0) + fY*(Force.y - 0.0) + fZ*(Force.z - 1.0);
		break;
	case TRAN_BACK_RCS:
		//z-
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z - 0.0) + fX*(Force.x - 0.0) + fY*(Force.y - 0.0) + fZ*(Force.z + 1.0);
		break;
	case TRAN_UP_RCS:
		//y+
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z - 0.0) + fX*(Force.x - 0.0) + fY*(Force.y - 1.0) + fZ*(Force.z - 0.0);
		break;
	case TRAN_DOWN_RCS:
		//y-
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z - 0.0) + fX*(Force.x - 0.0) + fY*(Force.y + 1.0) + fZ*(Force.z - 0.0);
		break;
	case TRAN_LEFT_RCS:
		//x-
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z - 0.0) + fX*(Force.x + 1.0) + fY*(Force.y - 0.0) + fZ*(Force.z - 0.0);
		break;
	case TRAN_RIGHT_RCS:
		//x+
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z - 0.0) + fX*(Force.x - 1.0) + fY*(Force.y - 0.0) + fZ*(Force.z - 0.0);
		break;
	case ROT_PITCHUP_RCS:
		//y+ Torque.y = 1.0 others zero
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y - 1.0) + tZ*(Torque.z - 0.0) + fX*(Force.x - 0.0) + fY*(Force.y - 0.0) + fZ*(Force.z - 0.0);
		break;
	case ROT_PITCHDOWN_RCS:
		//y- wtf
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y + 1.0) + tZ*(Torque.z - 0.0) + fX*(Force.x - 0.0) + fY*(Force.y - 0.0) + fZ*(Force.z - 0.0);
		break;
	case ROT_YAWLEFT_RCS:
		//x-
		L = L + tX*(Torque.x + 1.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z - 0.0) + fX*(Force.x - 0.0) + fY*(Force.y - 0.0) + fZ*(Force.z - 0.0);
		break;
	case ROT_YAWRIGHT_RCS:
		//x+
		L = L + tX*(Torque.x - 1.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z - 0.0) + fX*(Force.x - 0.0) + fY*(Force.y - 0.0) + fZ*(Force.z - 0.0);
		break;
	case ROT_BANKLEFT_RCS:
		//z-
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z + 1.0) + fX*(Force.x - 0.0) + fY*(Force.y - 0.0) + fZ*(Force.z - 0.0);
		break;
	case ROT_BANKRIGHT_RCS:
		//z+
		L = L + tX*(Torque.x - 0.0) + tY*(Torque.y - 0.0) + tZ*(Torque.z - 1.0) + fX*(Force.x - 0.0) + fY*(Force.y - 0.0) + fZ*(Force.z - 0.0);
		break;
	default:
		break;
	};
	return L;
}

inline void ComputeOptimumRCSstring_LP(const int control,
									   vector<stack_thruster_t>* stackThrusters, 
									   stack_RCS_string_t* RCS_string,
									   const VECTOR3 COG, const double forceN, const bool ignoreSideEffects = false)
{
	// finally working
	unsigned int numActiveThrusters = RCS_string->size();
	vector<VECTOR3> forceCoeffArray, torqueCoeffArray;
	vector<double> fuelCostCoefficients;
	bool ShallSetTorque = false;
	ComputeForceCoefficientArray(stackThrusters, RCS_string, COG, forceCoeffArray);
	ComputeTorqueCoefficientArray(stackThrusters, RCS_string, COG, torqueCoeffArray);
	ComputeFuelCostArray(stackThrusters, RCS_string, fuelCostCoefficients);
	//for (unsigned int i = 0; i<numActiveThrusters; i++)
	//{
	//	char line[256];
	//	sprintf_s(line, 255, "force: %.3f %.3f %.3f torque: %.3f %.3f %.3f cost: %.3f", 
	//		forceCoeffArray[i].x, forceCoeffArray[i].y, forceCoeffArray[i].z,
	//		torqueCoeffArray[i].x, torqueCoeffArray[i].y, torqueCoeffArray[i].z,
	//		fuelCostCoefficients[i]);
	//	oapiWriteLog(line);
	//};
	lprec * ThrusterProblem;
	REAL * row = NULL;
	if (ignoreSideEffects)
	{
		ThrusterProblem = make_lp(3, numActiveThrusters); // 3 constraints, numActiveThrusters vars
	}
	else
	{
		ThrusterProblem = make_lp(6, numActiveThrusters); // 6 constraints (force/torque equalities), numActiveThrusters vars
	};

	if (ThrusterProblem == NULL)
	{
		oapiWriteLog("Cannot create LP record");
		throw "Cannot create LP record";
	};

	double RHS[6] = {0.0,0.0,0.0,0.0,0.0,0.0};
	if (ignoreSideEffects)
	{
	//	double RHS[3] = {0.0,0.0,0.0};


		switch (control)
		{
		case TRAN_FORWARD_MAIN:
			//z+
			RHS[fZ-3] = forceN;
			break;
		case TRAN_BACK_RETRO:
			//z-
			RHS[fZ-3] = -forceN;
			break;
		case TRAN_UP_HOVER:
			//y+
			RHS[fY-3] = forceN;
			break;
		case TRAN_FORWARD_RCS:
			//z+
			RHS[fZ-3] = forceN;
			break;
		case TRAN_BACK_RCS:
			//z-
			RHS[fZ-3] = -forceN;
			break;
		case TRAN_UP_RCS:
			//y+
			//RHS[fY] = forceN;
			RHS[fY-3] = -forceN;
			break;
		case TRAN_DOWN_RCS:
			//y-
			//RHS[fY] = -forceN;
			RHS[fY-3] = forceN;
			break;
		case TRAN_LEFT_RCS:
			//x-
			RHS[fX-3] = -forceN;
			break;
		case TRAN_RIGHT_RCS:
			//x+
			RHS[fX-3] = forceN;
			break;
		case ROT_PITCHUP_RCS:
			//y+ Torque.y = forceN others zero
			//RHS[tY] = forceN;
			RHS[tX] = -forceN;//WTF?
			ShallSetTorque = true;
			break;
		case ROT_PITCHDOWN_RCS:
			//y- wtf
			//RHS[tY] = -forceN;
			RHS[tX] = +forceN;
			ShallSetTorque = true;
			break;
		case ROT_YAWLEFT_RCS: //WTF the signs are off! the axis are off too
			//x-
			//RHS[tX] = -forceN;
			RHS[tY] = -forceN;
			ShallSetTorque = true;
			break;
		case ROT_YAWRIGHT_RCS:
			//x+
			//RHS[tX] = forceN;
			RHS[tY] = forceN;
			ShallSetTorque = true;
			break;
		case ROT_BANKLEFT_RCS:
			//z-
			//RHS[tZ] = -forceN;
			RHS[tZ] = forceN;
			ShallSetTorque = true;
			break;
		case ROT_BANKRIGHT_RCS:
			//z+
			//RHS[tZ] = forceN;
			RHS[tZ] = -forceN;
			ShallSetTorque = true;
			break;
		default:
			break;
		};
	}
	else
	{

		//p->y 8->1 4->6 1->2?

		// TODO: rebalancing after spending mass; saving pulse length in the file

		// CTD at shutdown - where do I have a dangling pointer?

		switch (control)
		{
		case TRAN_FORWARD_MAIN:
			//z+
			RHS[fZ] = forceN;
			break;
		case TRAN_BACK_RETRO:
			//z-
			RHS[fZ] = -forceN;
			break;
		case TRAN_UP_HOVER:
			//y+
			RHS[fY] = forceN;
			break;
		case TRAN_FORWARD_RCS:
			//z+
			RHS[fZ] = forceN;
			break;
		case TRAN_BACK_RCS:
			//z-
			RHS[fZ] = -forceN;
			break;
		case TRAN_UP_RCS:
			//y+
			//RHS[fY] = forceN;
			RHS[fY] = -forceN;
			break;
		case TRAN_DOWN_RCS:
			//y-
			//RHS[fY] = -forceN;
			RHS[fY] = forceN;
			break;
		case TRAN_LEFT_RCS:
			//x-
			RHS[fX] = -forceN;
			break;
		case TRAN_RIGHT_RCS:
			//x+
			RHS[fX] = forceN;
			break;
		case ROT_PITCHUP_RCS:
			//y+ Torque.y = forceN others zero
			//RHS[tY] = forceN;
			RHS[tX] = -forceN;//WTF?
			break;
		case ROT_PITCHDOWN_RCS:
			//y- wtf
			//RHS[tY] = -forceN;
			RHS[tX] = +forceN;
			break;
		case ROT_YAWLEFT_RCS: //WTF the signs are off! the axis are off too
			//x-
			//RHS[tX] = -forceN;
			RHS[tY] = -forceN;
			break;
		case ROT_YAWRIGHT_RCS:
			//x+
			//RHS[tX] = forceN;
			RHS[tY] = forceN;
			break;
		case ROT_BANKLEFT_RCS:
			//z-
			//RHS[tZ] = -forceN;
			RHS[tZ] = forceN;
			break;
		case ROT_BANKRIGHT_RCS:
			//z+
			//RHS[tZ] = forceN;
			RHS[tZ] = -forceN;
			break;
		default:
			break;
		};
	};

	row = (REAL *) malloc((numActiveThrusters+1) * sizeof(*row));

	if (row == NULL)
	{
		oapiWriteLog("Cannot allocate memory for value row");
		throw "Cannot allocate memory for value row";
	};

	set_add_rowmode(ThrusterProblem , TRUE);
	for (unsigned int j = 0; j<numActiveThrusters; j++)
	{
		char cn[256];
		sprintf_s(cn, 255, "%s_%d", RCS_string->at(j).vesselName.c_str(), RCS_string->at(j).thrusterID);
		for(char* pp = cn;(*pp)!=NULL;pp++) { if ((*pp)=='-' || (*pp)==' ' || (*pp)=='+') (*pp)='_'; };

		set_col_name(ThrusterProblem, j+1, cn);
	};

	if (ignoreSideEffects)
	{
		for (unsigned int i = 0; i<3; i++) // row by row
		{
			for (unsigned int j = 0; j<numActiveThrusters; j++)
			{
				if (ShallSetTorque)
				{
					switch (i)
					{
					case tX:
						row[j+1] = torqueCoeffArray[j].x;
						break;
					case tY:
						row[j+1] = torqueCoeffArray[j].y;
						break;
					case tZ:
						row[j+1] = torqueCoeffArray[j].z;
						break;
					default:
						break;
					};
				}
				else
				{
					switch (i)
					{
					case fX-3:
						row[j+1] = forceCoeffArray[j].x;
						break;
					case fY-3:
						row[j+1] = forceCoeffArray[j].y;
						break;
					case fZ-3:
						row[j+1] = forceCoeffArray[j].z;
						break;
					default:
						break;
					};
				};
			};
			/* add the row to lpsolve */ //? or plus 1?
			if(!add_constraint(ThrusterProblem, row, EQ, RHS[i]))
				//if(!add_constraintex(ThrusterProblem, numActiveThrusters, row, NULL, EQ, RHS[i]))
				//if(!add_constraintex(ThrusterProblem, numActiveThrusters+1, row, NULL, EQ, RHS[i]))
			{
				oapiWriteLog("Cannot add constraint");
				throw "Cannot add constraint";
			};
		};
	}
	else
	{
		for (unsigned int i = 0; i<6; i++) // row by row
		{
			for (unsigned int j = 0; j<numActiveThrusters; j++)
			{
				switch (i)
				{
				case tX:
					row[j+1] = torqueCoeffArray[j].x;
					break;
				case tY:
					row[j+1] = torqueCoeffArray[j].y;
					break;
				case tZ:
					row[j+1] = torqueCoeffArray[j].z;
					break;
				case fX:
					row[j+1] = forceCoeffArray[j].x;
					break;
				case fY:
					row[j+1] = forceCoeffArray[j].y;
					break;
				case fZ:
					row[j+1] = forceCoeffArray[j].z;
					break;
				default:
					break;
				};
			};
			/* add the row to lpsolve */ //? or plus 1?
			if(!add_constraint(ThrusterProblem, row, EQ, RHS[i]))
				//if(!add_constraintex(ThrusterProblem, numActiveThrusters, row, NULL, EQ, RHS[i]))
				//if(!add_constraintex(ThrusterProblem, numActiveThrusters+1, row, NULL, EQ, RHS[i]))
			{
				oapiWriteLog("Cannot add constraint");
				throw "Cannot add constraint";
			};
		};
	};
	for (unsigned int j = 0; j<numActiveThrusters; j++)
	{
		set_lowbo(ThrusterProblem, j+1, 0.0);
		set_upbo(ThrusterProblem, j+1, 1.0);
	};
	set_add_rowmode(ThrusterProblem , FALSE);
	//objective function - fuel costs.
	for (unsigned int j = 0; j<numActiveThrusters; j++)
	{
		row[j+1] = fuelCostCoefficients[j];
	};
	if(!set_obj_fnex(ThrusterProblem, numActiveThrusters, row, NULL))
	{
		oapiWriteLog("Cannot set objective function");
		throw "Cannot set objective function";
	};
	set_minim(ThrusterProblem);
	//test
	//write_lp(ThrusterProblem, "model.lp"); 
	int ret = solve(ThrusterProblem);
	//char argh[64];
	//sprintf_s(argh, 63, "retcode: %d", ret);
	//oapiWriteLog(argh);
	//	if (ret!=OPTIMAL)
	//	{
	//oapiWriteLog("Cannot solve");
	//throw "Cannot solve";
	//		for (unsigned int i = 0; i<numActiveThrusters; i++)
	//		{
	//			RCS_string->at(i).thrustRatio = row [i+1];
	//		};
	//
	//	};
	get_variables(ThrusterProblem, row);

	// TODO: rescale to maximum thrust? how?


	for (unsigned int i = 0; i<numActiveThrusters; i++)
	{
		RCS_string->at(i).thrustRatio = min(max(row [i],0.0),1.0); // SHIT!
	};
	// weed zero thrusters
	stack_RCS_string_t::iterator iter = RCS_string->begin();

	while (iter!=RCS_string->end())
	{
		//if (fabs(iter->thrustRatio)<numeric_limits<double>::epsilon())
		if (abs_eq(iter->thrustRatio, 0.0))
		{
			iter = RCS_string->erase(iter);
		}
		else
		{
			iter++;
		};
	};

	free(row);
	delete_lp(ThrusterProblem);

};


bool RCS2string_s(int controlType, const stack_RCS_string_t& RCS_string, char * line, int count)
{
	bool shallSkip = false;
	char rcs_s[255];
	if (RCS_string.size()==0) return false;
	switch (controlType)
	{
	case TRAN_FORWARD_RCS:
		sprintf_s(line, count, "FW");
		break;
	case TRAN_BACK_RCS:
		sprintf_s(line, count, "BK");
		break;
	case TRAN_UP_RCS:
		sprintf_s(line, count, "UP");
		break;
	case TRAN_DOWN_RCS:
		sprintf_s(line, count, "DN");
		break;
	case TRAN_LEFT_RCS:
		sprintf_s(line, count, "LF");
		break;
	case TRAN_RIGHT_RCS:
		sprintf_s(line, count, "RT");
		break;
	case ROT_PITCHUP_RCS:
		sprintf_s(line, count, "P+");
		break;
	case ROT_PITCHDOWN_RCS:
		sprintf_s(line, count, "P-");
		break;
	case ROT_YAWLEFT_RCS:
		sprintf_s(line, count, "Y-");
		break;
	case ROT_YAWRIGHT_RCS:
		sprintf_s(line, count, "Y+");
		break;
	case ROT_BANKLEFT_RCS:
		sprintf_s(line, count, "B-");
		break;
	case ROT_BANKRIGHT_RCS:
		sprintf_s(line, count, "B+");
		break;
	default:
		return false;
		break;
	};
	if (shallSkip) return false;
	for (unsigned int k = 0; k<RCS_string.size(); k++)
	{
		sprintf_s(rcs_s, 255, ":%s:%d:%.3f",
			RCS_string[k].vesselName.c_str(),
			RCS_string[k].thrusterID,
			RCS_string[k].thrustRatio);
		strcat_s(line, count, rcs_s);
	};
	return true;
}

// first set the whole shebang to be the active string, then weed out zero thrust items.

// optimization - select thrusters to do the job cleanly without side effects - i.e. with the resulting torque
// strictly along the principal axis.

// the first vessel is the root one... check stuff by global to local conversions...

// this is a raw collection. We have to transform all thrusters to the root's coord frame, 
// and run through the optimizer. (a linear program methinks)

// the conversion defines the frame, which is chained back to the root... lefthandedness 
//is freaking me out...

#endif //!__AUTOPILOT_H__