#ifndef __SOLARSYS_H__
#define __SOLARSYS_H__

#include "orbitersdk.h"
#include <vector>
#include <hash_map>
#include <list>
#include <map>
#include <string>

using namespace std;
using namespace stdext;

//code from TransX via Ajaja

struct GBODY{
	OBJHANDLE bodyhandle;
	GBODY *parent, *next, *previous; // next and previous bodies orbiting same body
	list <GBODY*> satellites;
	double soisize2;
	double gravbodyratio2;
	double mass;
	bool is_a_star;
} ;

struct SolSystem
{
	GBODY *sun;
	map<OBJHANDLE, GBODY*> bodyMap;

	inline OBJHANDLE getmajor(OBJHANDLE handle)
	{
		if(bodyMap[handle])
			if(bodyMap[handle]->parent)
				return bodyMap[handle]->parent->bodyhandle;
		return NULL;
	}

	inline OBJHANDLE getfirstmoon(OBJHANDLE handle)
	{
		GBODY* body = bodyMap[handle];
		if(body)
			if(body->satellites.size() > 0)
				return body->satellites.front()->bodyhandle;
		return NULL;
	}

	inline OBJHANDLE getnextpeer(OBJHANDLE handle)
	{
		if(bodyMap[handle])
			if(bodyMap[handle]->next)
				return bodyMap[handle]->next->bodyhandle;
		return NULL;
	}

	inline VECTOR3 getweightedvector(OBJHANDLE body, void (*func)(OBJHANDLE, VECTOR3*))
	{
		OBJHANDLE moon = getfirstmoon(body);
		double totalmass = oapiGetMass(body), bodymass;
		VECTOR3 barycentre = {0,0,0}, bodypos;
		func(body, &bodypos);
		barycentre.x += bodypos.x * totalmass;
		barycentre.y += bodypos.y * totalmass;
		barycentre.z += bodypos.z * totalmass;

		while(moon)
		{
			func(moon, &bodypos);
			bodymass = oapiGetMass(moon);
			barycentre.x += bodypos.x * bodymass;
			barycentre.y += bodypos.y * bodymass;
			barycentre.z += bodypos.z * bodymass;
			totalmass += bodymass;

			moon = getnextpeer(moon);
		}

		barycentre.x /= totalmass;
		barycentre.y /= totalmass;
		barycentre.z /= totalmass;

		return barycentre;
	}

	inline VECTOR3 getbarycentrevel(OBJHANDLE body)
	{
		return getweightedvector(body, &oapiGetGlobalVel);
	}

	inline VECTOR3 getbarycentre(OBJHANDLE body)
	{
		return getweightedvector(body, &oapiGetGlobalPos);
	}


	inline SolSystem() // should we arrange a map to the SPICE body names? no!
	{
		int totalbodies=oapiGetGbodyCount();
		list<GBODY*> templist; // used for easy accesss to the bodies rather than traversing the tree
		for (int i = 0; i < totalbodies; i++)
		{
			GBODY *body = new GBODY;
			body->next = body->previous = NULL;

			body->bodyhandle = oapiGetGbodyByIndex(i);
			body->mass = oapiGetMass(body->bodyhandle);
			bodyMap[body->bodyhandle] = body;
			int objType = oapiGetObjectType(body->bodyhandle);

			if(i == 0)
			{
				if (objType!=OBJTP_STAR)
				{
					oapiWriteLog("first body is not a star");
					throw "first body is not a star";
				};
				// The central star
				body->soisize2 = 1e100;//Infinite SOI
				body->parent = body->next = body->previous = NULL;//Sun is its own parent!
				body->is_a_star = true;
				sun = body;
			}
			else
			{
				// gravbodyratio2 and soisize
				// Find the parent body (what it is orbiting)
				list<GBODY*>::iterator it = templist.begin();
				GBODY* currparent;
				double currdistance2;
				while(it != templist.end())
				{
					char itname[30], name[30];
					oapiGetObjectName((*it)->bodyhandle, itname, 30);
					oapiGetObjectName(body->bodyhandle, name, 30);

					VECTOR3 pos;
					oapiGetRelativePos(body->bodyhandle, (*it)->bodyhandle, &pos);
					double distance2 = dotp(pos, pos);
					if(distance2 < (*it)->soisize2)
					{
						currparent = *it;
						currdistance2 = distance2;
					}

					it++;
				}
				body->parent = currparent;
				if(body->parent->satellites.size() > 0)
				{
					body->previous = body->parent->satellites.back();
					body->previous->next = body;
				}
				body->parent->satellites.push_back(body);
				body->gravbodyratio2 = pow(body->mass / body->parent->mass, 0.8);
				body->soisize2 = currdistance2 * body->gravbodyratio2;//Internal soi size
				body->is_a_star = (objType==OBJTP_STAR);

			}
			templist.push_back(body); // add it to the temp list
		}
	}
inline ~SolSystem()
{
	int totalbodies=oapiGetGbodyCount();
	for (int i = 0; i < totalbodies; i++)
	{
			delete bodyMap[oapiGetGbodyByIndex(i)];
	};
}

	inline void calc_Lagrange_Coeff(const char * body, const char *bodyref, 
		double &coeff_L1, double &coeff_L2, double &coeff_L3, double & coeff_L)
	{
		OBJHANDLE h,h1,h2;
		h1=oapiGetGbodyByName(const_cast<char *>(bodyref));
		h2=oapiGetGbodyByName(const_cast<char *>(body));
		double m1=oapiGetMass(h1);
		double m2=oapiGetMass(h2);


		if ((h=getfirstmoon(h2))!=NULL)
		{
			m2+=bodyMap[h]->mass;
			while ((h=getnextpeer(h))!=NULL) m2+=bodyMap[h]->mass;
		}

		double a=m2/(m1+m2);
		coeff_L1=1.0+Newton_L1(-pow(a/3.0,1.0/3.0),a,1e-17);
		coeff_L2=1.0+Newton_L2(pow(a/3.0,1.0/3.0),a,1e-17);
		coeff_L3=1.0+Newton_L3(-2.0+5.0/12.0*a,a,1e-17);
		coeff_L = m2/(m1+m2);
	}

	inline double Newton_L1(const double x1, const double a, const double e)
	{
		double result;
		double x;
		double d;
		int i=0;

		x = x1;
		d = fdf_L1(x,a);
		while((fabs(d)>e)&&(i<1000))
		{
			x = x-d;
			d = fdf_L1(x,a);
			i++;
		}
		result = x;
		return result;
	}

	inline double Newton_L2(const double x1, const double a, const double e)
	{
		double result;
		double x;
		double d;
		int i=0;

		x = x1;
		d = fdf_L2(x,a);
		while((fabs(d)>e)&&(i<1000))
		{
			x = x-d;
			d = fdf_L2(x,a);
			i++;
		}
		result = x;
		return result;
	}

	inline double Newton_L3(double x1, double a, double e)
	{
		double result;
		double x;
		double d;
		int i=0;

		x = x1;
		d = fdf_L3(x,a);
		while((fabs(d)>e)&&(i<1000))
		{
			x = x-d;
			d = fdf_L3(x,a);
			i++;
		}
		result = x;
		return result;
	}

	inline double fdf_L1(double u, double a)
	{
		double t1 = u*u;
		double t2 = t1*u;
		double t4 = t1*t1;
		double t8 = 2.0*a*u;
		double t9 = a*t1;
		double t10 = a*t2;
		return((-3.0*t2-3.0*t4-t4*u-a-t8-t9+2.0*t10+a*t4)/(-9.0*t1-12.0*t2-5.0*t4-2.0*a-t8+6.0*t9+4.0*t10));
	}

	inline double fdf_L2(double u, double a)
	{
		double t1 = u*u;
		double t2 = t1*u;
		double t4 = t1*t1;
		double t8 = 2.0*a*u;
		double t9 = a*t1;
		double t10 = a*t2;
		return((-3.0*t2-3.0*t4-t4*u+a+t8+t9+2.0*t10+a*t4)/(-9.0*t1-12.0*t2-5.0*t4+2.0*a+t8+6.0*t9+4.0*t10));
	}

	inline double fdf_L3(double u, double a)
	{
		double t1 = u*u;    
		double t3 = t1*u;
		double t5 = t1*t1;
		double t9 = 2.0*a*u;
		double t10 = a*t1;
		double t11 = a*t3;
		return((-2.0*t1-3.0*t3-3.0*t5-t5*u-a-t9+t10+2.0*t11+a*t5)/(-4.0*u-9.0*t1-12.0*t3-5.0*t5-2.0*a+t9+6.0*t10+4.0*t11));
	}

};

inline OBJHANDLE GetCentralStar(SolSystem* solSys)
{
	//API problems, not elegant! CTDs in Ups And.
	OBJHANDLE hstar = solSys->sun->bodyhandle;
	//OBJHANDLE hstar = oapiGetGbodyByName("Sun");
	return hstar;
	//OBJHANDLE hBody = oapiGetGbodyByIndex(0);
	//if (hBody==NULL) throw "nobody in the system :-)";
	//CELBODY* Body = NULL;
	//if ((Body=oapiGetCelbodyInterface(hBody))==NULL) throw "0th object not a CELBODY";
	//CELBODY2* Body2 = NULL;
	//OBJHANDLE hParent = NULL;
	//while ((hParent=Body->GetParent())!=NULL)
	//{
	//	if ((Body=oapiGetCelbodyInterface(hParent))==NULL) throw "Parent object not a CELBODY";
	//};
	//return Body->GetHandle();
}


#endif //!__SOLARSYS_H__